@@ -1,12 +1,16 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver / Accounting
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -27,10 +31,15 @@
 #include "radius_client.h"
 #include "eloop.h"
 #include "accounting.h"
-#include "driver.h"
 #include "ieee802_1x.h"
+#include "driver.h"
 
 
+/* Default interval in seconds for polling TX/RX octets from the driver if
+ * STA is not using interim accounting. This detects wrap arounds for
+ * input/output octets and updates Acct-{Input,Output}-Gigawords. */
+#define ACCT_DEFAULT_UPDATE_INTERVAL 300
+
 static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 					  int status_type)
 {
@@ -40,25 +49,24 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 	size_t len;
 
 	msg = radius_msg_new(RADIUS_CODE_ACCOUNTING_REQUEST,
-			     radius_client_get_id(hapd));
+			     radius_client_get_id(hapd->radius));
 	if (msg == NULL) {
 		printf("Could not create net RADIUS packet\n");
 		return NULL;
 	}
 
 	if (sta) {
-		radius_msg_make_authenticator(msg, (u8 *) sta, sizeof(sta));
+		radius_msg_make_authenticator(msg, (u8 *) sta, sizeof(*sta));
 
 		snprintf(buf, sizeof(buf), "%08X-%08X",
-			 hapd->radius->acct_session_id_hi,
-			 sta->acct_session_id_lo);
+			 sta->acct_session_id_hi, sta->acct_session_id_lo);
 		if (!radius_msg_add_attr(msg, RADIUS_ATTR_ACCT_SESSION_ID,
-					 buf, strlen(buf))) {
+					 (u8 *) buf, strlen(buf))) {
 			printf("Could not add Acct-Session-Id\n");
 			goto fail;
 		}
 	} else {
-		radius_msg_make_authenticator(msg, (u8 *) hapd, sizeof(hapd));
+		radius_msg_make_authenticator(msg, (u8 *) hapd, sizeof(*hapd));
 	}
 
 	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_ACCT_STATUS_TYPE,
@@ -80,8 +88,8 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 		if (!val) {
 			snprintf(buf, sizeof(buf), RADIUS_ADDR_FORMAT,
 				 MAC2STR(sta->addr));
-			val = buf;
-			len = strlen(val);
+			val = (u8 *) buf;
+			len = strlen(buf);
 		}
 
 		if (!radius_msg_add_attr(msg, RADIUS_ATTR_USER_NAME, val,
@@ -99,7 +107,7 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 
 	if (hapd->conf->nas_identifier &&
 	    !radius_msg_add_attr(msg, RADIUS_ATTR_NAS_IDENTIFIER,
-				 hapd->conf->nas_identifier,
+				 (u8 *) hapd->conf->nas_identifier,
 				 strlen(hapd->conf->nas_identifier))) {
 		printf("Could not add NAS-Identifier\n");
 		goto fail;
@@ -114,7 +122,7 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 	snprintf(buf, sizeof(buf), RADIUS_802_1X_ADDR_FORMAT ":%s",
 		 MAC2STR(hapd->own_addr), hapd->conf->ssid);
 	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CALLED_STATION_ID,
-				 buf, strlen(buf))) {
+				 (u8 *) buf, strlen(buf))) {
 		printf("Could not add Called-Station-Id\n");
 		goto fail;
 	}
@@ -123,7 +131,7 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 		snprintf(buf, sizeof(buf), RADIUS_802_1X_ADDR_FORMAT,
 			 MAC2STR(sta->addr));
 		if (!radius_msg_add_attr(msg, RADIUS_ATTR_CALLING_STATION_ID,
-					 buf, strlen(buf))) {
+					 (u8 *) buf, strlen(buf))) {
 			printf("Could not add Calling-Station-Id\n");
 			goto fail;
 		}
@@ -137,7 +145,7 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 
 		snprintf(buf, sizeof(buf), "CONNECT 11Mbps 802.11b");
 		if (!radius_msg_add_attr(msg, RADIUS_ATTR_CONNECT_INFO,
-					 buf, strlen(buf))) {
+					 (u8 *) buf, strlen(buf))) {
 			printf("Could not add Connect-Info\n");
 			goto fail;
 		}
@@ -159,40 +167,77 @@ static struct radius_msg * accounting_msg(hostapd *hapd, struct sta_info *sta,
 }
 
 
+static int accounting_sta_update_stats(struct hostapd_data *hapd,
+				       struct sta_info *sta,
+				       struct hostap_sta_driver_data *data)
+{
+	if (hostapd_read_sta_data(hapd, data, sta->addr))
+		return -1;
+
+	if (sta->last_rx_bytes > data->rx_bytes)
+		sta->acct_input_gigawords++;
+	if (sta->last_tx_bytes > data->tx_bytes)
+		sta->acct_output_gigawords++;
+	sta->last_rx_bytes = data->rx_bytes;
+	sta->last_tx_bytes = data->tx_bytes;
+
+	hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_RADIUS,
+		       HOSTAPD_LEVEL_DEBUG, "updated TX/RX stats: "
+		       "Acct-Input-Octets=%lu Acct-Input-Gigawords=%u "
+		       "Acct-Output-Octets=%lu Acct-Output-Gigawords=%u",
+		       sta->last_rx_bytes, sta->acct_input_gigawords,
+		       sta->last_tx_bytes, sta->acct_output_gigawords);
+
+	return 0;
+}
+
+
 static void accounting_interim_update(void *eloop_ctx, void *timeout_ctx)
 {
 	hostapd *hapd = eloop_ctx;
 	struct sta_info *sta = timeout_ctx;
-
-	accounting_sta_interim(hapd, sta);
+	int interval;
 
 	if (sta->acct_interim_interval) {
-		eloop_register_timeout(sta->acct_interim_interval, 0,
-				       accounting_interim_update, hapd, sta);
+		accounting_sta_interim(hapd, sta);
+		interval = sta->acct_interim_interval;
+	} else {
+		struct hostap_sta_driver_data data;
+		accounting_sta_update_stats(hapd, sta, &data);
+		interval = ACCT_DEFAULT_UPDATE_INTERVAL;
 	}
+
+	eloop_register_timeout(interval, 0, accounting_interim_update,
+			       hapd, sta);
 }
 
 
 void accounting_sta_start(hostapd *hapd, struct sta_info *sta)
 {
 	struct radius_msg *msg;
+	int interval;
 	
 	if (sta->acct_session_started)
 		return;
 
 	time(&sta->acct_session_start);
+	sta->last_rx_bytes = sta->last_tx_bytes = 0;
+	sta->acct_input_gigawords = sta->acct_output_gigawords = 0;
+	hostapd_sta_clear_stats(hapd, sta->addr);
 
 	if (!hapd->conf->acct_server)
 		return;
 
-	if (sta->acct_interim_interval) {
-		eloop_register_timeout(sta->acct_interim_interval, 0,
-				       accounting_interim_update, hapd, sta);
-	}
+	if (sta->acct_interim_interval)
+		interval = sta->acct_interim_interval;
+	else
+		interval = ACCT_DEFAULT_UPDATE_INTERVAL;
+	eloop_register_timeout(interval, 0, accounting_interim_update,
+			       hapd, sta);
 
 	msg = accounting_msg(hapd, sta, RADIUS_ACCT_STATUS_TYPE_START);
 	if (msg)
-		radius_client_send(hapd, msg, RADIUS_ACCT, sta->addr);
+		radius_client_send(hapd->radius, msg, RADIUS_ACCT, sta->addr);
 
 	sta->acct_session_started = 1;
 }
@@ -203,6 +248,7 @@ void accounting_sta_report(hostapd *hapd, struct sta_info *sta, int stop)
 	struct radius_msg *msg;
 	int cause = sta->acct_terminate_cause;
 	struct hostap_sta_driver_data data;
+	u32 gigawords;
 
 	if (!hapd->conf->acct_server)
 		return;
@@ -221,8 +267,7 @@ void accounting_sta_report(hostapd *hapd, struct sta_info *sta, int stop)
 		goto fail;
 	}
 
-	if (hostapd_read_sta_driver_data(hapd->driver.data,
-					 &data, sta->addr) == 0) {
+	if (accounting_sta_update_stats(hapd, sta, &data) == 0) {
 		if (!radius_msg_add_attr_int32(msg,
 					       RADIUS_ATTR_ACCT_INPUT_PACKETS,
 					       data.rx_packets)) {
@@ -241,12 +286,40 @@ void accounting_sta_report(hostapd *hapd, struct sta_info *sta, int stop)
 			printf("Could not add Acct-Input-Octets\n");
 			goto fail;
 		}
+		gigawords = sta->acct_input_gigawords;
+#if __WORDSIZE == 64
+		gigawords += data.rx_bytes >> 32;
+#endif
+		if (gigawords &&
+		    !radius_msg_add_attr_int32(
+			    msg, RADIUS_ATTR_ACCT_INPUT_GIGAWORDS,
+			    gigawords)) {
+			printf("Could not add Acct-Input-Gigawords\n");
+			goto fail;
+		}
 		if (!radius_msg_add_attr_int32(msg,
 					       RADIUS_ATTR_ACCT_OUTPUT_OCTETS,
 					       data.tx_bytes)) {
 			printf("Could not add Acct-Output-Octets\n");
 			goto fail;
 		}
+		gigawords = sta->acct_output_gigawords;
+#if __WORDSIZE == 64
+		gigawords += data.tx_bytes >> 32;
+#endif
+		if (gigawords &&
+		    !radius_msg_add_attr_int32(
+			    msg, RADIUS_ATTR_ACCT_OUTPUT_GIGAWORDS,
+			    gigawords)) {
+			printf("Could not add Acct-Output-Gigawords\n");
+			goto fail;
+		}
+	}
+
+	if (!radius_msg_add_attr_int32(msg, RADIUS_ATTR_EVENT_TIMESTAMP,
+				       time(NULL))) {
+		printf("Could not add Event-Timestamp\n");
+		goto fail;
 	}
 
 	if (eloop_terminated())
@@ -259,7 +332,7 @@ void accounting_sta_report(hostapd *hapd, struct sta_info *sta, int stop)
 		goto fail;
 	}
 
-	radius_client_send(hapd, msg,
+	radius_client_send(hapd->radius, msg,
 			   stop ? RADIUS_ACCT : RADIUS_ACCT_INTERIM,
 			   sta->addr);
 	return;
@@ -287,10 +360,19 @@ void accounting_sta_stop(hostapd *hapd, struct sta_info *sta)
 }
 
 
+void accounting_sta_get_id(struct hostapd_data *hapd, struct sta_info *sta)
+{
+	sta->acct_session_id_lo = hapd->acct_session_id_lo++;
+	if (hapd->acct_session_id_lo == 0) {
+		hapd->acct_session_id_hi++;
+	}
+	sta->acct_session_id_hi = hapd->acct_session_id_hi;
+}
+
+
 /* Process the RADIUS frames from Accounting Server */
 static RadiusRxResult
-accounting_receive(hostapd *hapd,
-		   struct radius_msg *msg, struct radius_msg *req,
+accounting_receive(struct radius_msg *msg, struct radius_msg *req,
 		   u8 *shared_secret, size_t shared_secret_len, void *data)
 {
 	if (msg->hdr->code != RADIUS_CODE_ACCOUNTING_RESPONSE) {
@@ -302,7 +384,7 @@ accounting_receive(hostapd *hapd,
 	{
 		printf("Incoming RADIUS packet did not have correct "
 		       "Authenticator - dropped\n");
-		return RADIUS_RX_UNKNOWN;
+		return RADIUS_RX_INVALID_AUTHENTICATOR;
 	}
 
 	return RADIUS_RX_PROCESSED;
@@ -333,7 +415,7 @@ static void accounting_report_state(struct hostapd_data *hapd, int on)
 		return;
 	}
 
-	radius_client_send(hapd, msg, RADIUS_ACCT, NULL);
+	radius_client_send(hapd->radius, msg, RADIUS_ACCT, NULL);
 }
 
 
@@ -341,10 +423,10 @@ int accounting_init(hostapd *hapd)
 {
 	/* Acct-Session-Id should be unique over reboots. If reliable clock is
 	 * not available, this could be replaced with reboot counter, etc. */
-	hapd->radius->acct_session_id_hi = time(NULL);
+	hapd->acct_session_id_hi = time(NULL);
 
-	if (radius_client_register(hapd, RADIUS_ACCT, accounting_receive,
-				   NULL))
+	if (radius_client_register(hapd->radius, RADIUS_ACCT,
+				   accounting_receive, hapd))
 		return -1;
 
 	accounting_report_state(hapd, 1);
