@@ -1,12 +1,16 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver / IEEE 802.1X Authenticator - EAPOL state machine
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -22,13 +26,11 @@
 #include "eloop.h"
 #include "wpa.h"
 #include "sta_info.h"
+#include "eap.h"
 
-/* TODO:
- * implement state machines: Controlled Directions and Key Receive
- */
+static struct eapol_callbacks eapol_cb;
 
-
-/* EAPOL state machines are described in IEEE Std 802.1X-2001, Chap. 8.5 */
+/* EAPOL state machines are described in IEEE Std 802.1X-REV-d11, Chap. 8.2 */
 
 #define setPortAuthorized() \
 ieee802_1x_set_sta_authorized(sm->hapd, sm->sta, 1)
@@ -36,18 +38,13 @@ ieee802_1x_set_sta_authorized(sm->hapd, sm->sta, 1)
 ieee802_1x_set_sta_authorized(sm->hapd, sm->sta, 0)
 
 /* procedures */
-#define txCannedFail(x) ieee802_1x_tx_canned_eap(sm->hapd, sm->sta, (x), 0)
-#define txCannedSuccess(x) ieee802_1x_tx_canned_eap(sm->hapd, sm->sta, (x), 1)
-/* TODO: IEEE 802.1aa/D4 replaces txReqId(x) with txInitialMsg(x); value of
- * initialEAPMsg should be used to select which type of EAP packet is sent;
- * Currently, hostapd only supports EAP Request/Identity, so this can be
- * hardcoded. */
-#define txInitialMsg(x) ieee802_1x_request_identity(sm->hapd, sm->sta, (x))
-#define txReq(x) ieee802_1x_tx_req(sm->hapd, sm->sta, (x))
-#define sendRespToServer ieee802_1x_send_resp_to_server(sm->hapd, sm->sta)
-/* TODO: check if abortAuth would be needed for something */
-#define abortAuth do { } while (0)
-#define txKey(x) ieee802_1x_tx_key(sm->hapd, sm->sta, (x))
+#define txCannedFail() ieee802_1x_tx_canned_eap(sm->hapd, sm->sta, 0)
+#define txCannedSuccess() ieee802_1x_tx_canned_eap(sm->hapd, sm->sta, 1)
+#define txReq() ieee802_1x_tx_req(sm->hapd, sm->sta)
+#define sendRespToServer() ieee802_1x_send_resp_to_server(sm->hapd, sm->sta)
+#define abortAuth() ieee802_1x_abort_auth(sm->hapd, sm->sta)
+#define txKey() ieee802_1x_tx_key(sm->hapd, sm->sta)
+#define processKey() do { } while (0)
 
 
 /* Definitions for clarifying state machine implementation */
@@ -83,14 +80,11 @@ static void eapol_port_timers_tick(void *eloop_ctx, void *timeout_ctx)
 		state->quietWhile--;
 	if (state->reAuthWhen > 0)
 		state->reAuthWhen--;
-	if (state->txWhen > 0)
-		state->txWhen--;
 
 	if (state->hapd->conf->debug >= HOSTAPD_DEBUG_MSGDUMPS)
 		printf("IEEE 802.1X: " MACSTR " Port Timers TICK "
-		       "(timers: %d %d %d %d)\n", MAC2STR(state->addr),
-		       state->aWhile, state->quietWhile, state->reAuthWhen,
-		       state->txWhen);
+		       "(timers: %d %d %d)\n", MAC2STR(state->addr),
+		       state->aWhile, state->quietWhile, state->reAuthWhen);
 
 	eapol_sm_step(state);
 
@@ -104,8 +98,9 @@ static void eapol_port_timers_tick(void *eloop_ctx, void *timeout_ctx)
 SM_STATE(AUTH_PAE, INITIALIZE)
 {
 	SM_ENTRY(AUTH_PAE, INITIALIZE, auth_pae);
-	sm->currentId = 0;
 	sm->auth_pae.portMode = Auto;
+
+	sm->currentId = 255;
 }
 
 
@@ -113,52 +108,52 @@ SM_STATE(AUTH_PAE, DISCONNECTED)
 {
 	int from_initialize = sm->auth_pae.state == AUTH_PAE_INITIALIZE;
 
-	if (sm->auth_pae.state == AUTH_PAE_CONNECTING &&
-	    sm->auth_pae.eapLogoff)
-		sm->auth_pae.authEapLogoffsWhileConnecting++;
+	if (sm->auth_pae.eapolLogoff) {
+		if (sm->auth_pae.state == AUTH_PAE_CONNECTING)
+			sm->auth_pae.authEapLogoffsWhileConnecting++;
+		else if (sm->auth_pae.state == AUTH_PAE_AUTHENTICATED)
+			sm->auth_pae.authAuthEapLogoffWhileAuthenticated++;
+	}
 
 	SM_ENTRY(AUTH_PAE, DISCONNECTED, auth_pae);
 
-	sm->portStatus = Unauthorized;
+	sm->authPortStatus = Unauthorized;
 	setPortUnauthorized();
-	sm->auth_pae.eapLogoff = FALSE;
 	sm->auth_pae.reAuthCount = 0;
-	/* IEEE 802.1X state machine uses txCannedFail() always in this state.
-	 * However, sending EAP packet with failure code seems to cause WinXP
-	 * Supplicant to deauthenticate, which will set portEnabled = FALSE and
-	 * state machines end back to INITIALIZE and then back here to send
-	 * canned failure, and so on.. Avoid this by not sending failure packet
-	 * when DISCONNECTED state is entered from INITIALIZE state. */
+	sm->auth_pae.eapolLogoff = FALSE;
 	if (!from_initialize) {
-		txCannedFail(sm->currentId);
-		sm->currentId++;
 		if (sm->flags & EAPOL_SM_PREAUTH)
 			rsn_preauth_finished(sm->hapd, sm->sta, 0);
 	}
 }
 
 
+SM_STATE(AUTH_PAE, RESTART)
+{
+	if (sm->auth_pae.state == AUTH_PAE_AUTHENTICATED) {
+		if (sm->reAuthenticate)
+			sm->auth_pae.authAuthReauthsWhileAuthenticated++;
+		if (sm->auth_pae.eapolStart)
+			sm->auth_pae.authAuthEapStartsWhileAuthenticated++;
+		if (sm->auth_pae.eapolLogoff)
+			sm->auth_pae.authAuthEapLogoffWhileAuthenticated++;
+	}
+
+	SM_ENTRY(AUTH_PAE, RESTART, auth_pae);
+
+	sm->auth_pae.eapRestart = TRUE;
+	ieee802_1x_request_identity(sm->hapd, sm->sta);
+}
+
+
 SM_STATE(AUTH_PAE, CONNECTING)
 {
 	if (sm->auth_pae.state != AUTH_PAE_CONNECTING)
 		sm->auth_pae.authEntersConnecting++;
 
-	if (sm->auth_pae.state == AUTH_PAE_AUTHENTICATED) {
-		if (sm->reAuthenticate)
-			sm->auth_pae.authAuthReauthsWhileAuthenticated++;
-		if (sm->auth_pae.eapStart)
-			sm->auth_pae.authAuthEapStartsWhileAuthenticated++;
-		if (sm->auth_pae.eapLogoff)
-			sm->auth_pae.authAuthEapLogoffWhileAuthenticated++;
-	}
-
 	SM_ENTRY(AUTH_PAE, CONNECTING, auth_pae);
 
-	sm->auth_pae.eapStart = FALSE;
 	sm->reAuthenticate = FALSE;
-	sm->txWhen = sm->auth_pae.txPeriod;
-	sm->auth_pae.rxInitialRsp = FALSE;
-	txInitialMsg(sm->currentId);
 	sm->auth_pae.reAuthCount++;
 }
 
@@ -170,11 +165,10 @@ SM_STATE(AUTH_PAE, HELD)
 
 	SM_ENTRY(AUTH_PAE, HELD, auth_pae);
 
-	sm->portStatus = Unauthorized;
+	sm->authPortStatus = Unauthorized;
 	setPortUnauthorized();
 	sm->quietWhile = sm->auth_pae.quietPeriod;
-	sm->auth_pae.eapLogoff = FALSE;
-	sm->currentId++;
+	sm->auth_pae.eapolLogoff = FALSE;
 
 	hostapd_logger(sm->hapd, sm->addr, HOSTAPD_MODULE_IEEE8021X,
 		       HOSTAPD_LEVEL_WARNING, "authentication failed");
@@ -190,10 +184,9 @@ SM_STATE(AUTH_PAE, AUTHENTICATED)
 							
 	SM_ENTRY(AUTH_PAE, AUTHENTICATED, auth_pae);
 
-	sm->portStatus = Authorized;
+	sm->authPortStatus = Authorized;
 	setPortAuthorized();
 	sm->auth_pae.reAuthCount = 0;
-	sm->currentId++;
 	hostapd_logger(sm->hapd, sm->addr, HOSTAPD_MODULE_IEEE8021X,
 		       HOSTAPD_LEVEL_INFO, "authenticated");
 	if (sm->flags & EAPOL_SM_PREAUTH)
@@ -203,12 +196,14 @@ SM_STATE(AUTH_PAE, AUTHENTICATED)
 
 SM_STATE(AUTH_PAE, AUTHENTICATING)
 {
-	if (sm->auth_pae.state == AUTH_PAE_CONNECTING &&
-	    sm->auth_pae.rxInitialRsp)
+	if (sm->auth_pae.state == AUTH_PAE_CONNECTING && sm->rx_identity) {
 		sm->auth_pae.authEntersAuthenticating++;
+		sm->rx_identity = FALSE;
+	}
 
 	SM_ENTRY(AUTH_PAE, AUTHENTICATING, auth_pae);
 
+	sm->auth_pae.eapolStart = FALSE;
 	sm->authSuccess = FALSE;
 	sm->authFail = FALSE;
 	sm->authTimeout = FALSE;
@@ -223,11 +218,9 @@ SM_STATE(AUTH_PAE, ABORTING)
 	if (sm->auth_pae.state == AUTH_PAE_AUTHENTICATING) {
 		if (sm->authTimeout)
 			sm->auth_pae.authAuthTimeoutsWhileAuthenticating++;
-		if (sm->reAuthenticate)
-			sm->auth_pae.authAuthReauthsWhileAuthenticating++;
-		if (sm->auth_pae.eapStart)
+		if (sm->auth_pae.eapolStart)
 			sm->auth_pae.authAuthEapStartsWhileAuthenticating++;
-		if (sm->auth_pae.eapLogoff)
+		if (sm->auth_pae.eapolLogoff)
 			sm->auth_pae.authAuthEapLogoffWhileAuthenticating++;
 	}
 
@@ -236,7 +229,6 @@ SM_STATE(AUTH_PAE, ABORTING)
 	sm->authAbort = TRUE;
 	sm->keyRun = FALSE;
 	sm->keyDone = FALSE;
-	sm->currentId++;
 }
 
 
@@ -244,12 +236,11 @@ SM_STATE(AUTH_PAE, FORCE_AUTH)
 {
 	SM_ENTRY(AUTH_PAE, FORCE_AUTH, auth_pae);
 
-	sm->portStatus = Authorized;
+	sm->authPortStatus = Authorized;
 	setPortAuthorized();
 	sm->auth_pae.portMode = ForceAuthorized;
-	sm->auth_pae.eapStart = FALSE;
-	txCannedSuccess(sm->currentId);
-	sm->currentId++;
+	sm->auth_pae.eapolStart = FALSE;
+	txCannedSuccess();
 }
 
 
@@ -257,12 +248,11 @@ SM_STATE(AUTH_PAE, FORCE_UNAUTH)
 {
 	SM_ENTRY(AUTH_PAE, FORCE_UNAUTH, auth_pae);
 
-	sm->portStatus = Unauthorized;
+	sm->authPortStatus = Unauthorized;
 	setPortUnauthorized();
 	sm->auth_pae.portMode = ForceUnauthorized;
-	sm->auth_pae.eapStart = FALSE;
-	txCannedFail(sm->currentId);
-	sm->currentId++;
+	sm->auth_pae.eapolStart = FALSE;
+	txCannedFail();
 }
 
 
@@ -286,30 +276,30 @@ SM_STEP(AUTH_PAE)
 			SM_ENTER(AUTH_PAE, DISCONNECTED);
 			break;
 		case AUTH_PAE_DISCONNECTED:
-			SM_ENTER(AUTH_PAE, CONNECTING);
+			SM_ENTER(AUTH_PAE, RESTART);
+			break;
+		case AUTH_PAE_RESTART:
+			if (!sm->auth_pae.eapRestart)
+				SM_ENTER(AUTH_PAE, CONNECTING);
 			break;
 		case AUTH_PAE_HELD:
 			if (sm->quietWhile == 0)
-				SM_ENTER(AUTH_PAE, CONNECTING);
+				SM_ENTER(AUTH_PAE, RESTART);
 			break;
 		case AUTH_PAE_CONNECTING:
-			if (sm->auth_pae.eapLogoff ||
+			if (sm->auth_pae.eapolLogoff ||
 			    sm->auth_pae.reAuthCount > sm->auth_pae.reAuthMax)
 				SM_ENTER(AUTH_PAE, DISCONNECTED);
-			else if (sm->auth_pae.rxInitialRsp &&
-				 sm->auth_pae.reAuthCount <=
-				 sm->auth_pae.reAuthMax)
+			else if ((sm->be_auth.eapReq &&
+				  sm->auth_pae.reAuthCount <=
+				  sm->auth_pae.reAuthMax) ||
+				 sm->eapSuccess || sm->eapFail)
 				SM_ENTER(AUTH_PAE, AUTHENTICATING);
-			else if ((sm->txWhen == 0 || sm->auth_pae.eapStart ||
-				  sm->reAuthenticate) &&
-				 sm->auth_pae.reAuthCount <=
-				 sm->auth_pae.reAuthMax)
-				SM_ENTER(AUTH_PAE, CONNECTING);
 			break;
 		case AUTH_PAE_AUTHENTICATED:
-			if (sm->auth_pae.eapStart || sm->reAuthenticate)
-				SM_ENTER(AUTH_PAE, CONNECTING);
-			else if (sm->auth_pae.eapLogoff || !sm->portValid)
+			if (sm->auth_pae.eapolStart || sm->reAuthenticate)
+				SM_ENTER(AUTH_PAE, RESTART);
+			else if (sm->auth_pae.eapolLogoff || !sm->portValid)
 				SM_ENTER(AUTH_PAE, DISCONNECTED);
 			break;
 		case AUTH_PAE_AUTHENTICATING:
@@ -318,23 +308,22 @@ SM_STEP(AUTH_PAE)
 			else if (sm->authFail ||
 				 (sm->keyDone && !sm->portValid))
 				SM_ENTER(AUTH_PAE, HELD);
-			else if (sm->reAuthenticate || sm->auth_pae.eapStart ||
-				 sm->auth_pae.eapLogoff ||
-				 sm->authTimeout)
+			else if (sm->auth_pae.eapolStart ||
+				 sm->auth_pae.eapolLogoff || sm->authTimeout)
 				SM_ENTER(AUTH_PAE, ABORTING);
 			break;
 		case AUTH_PAE_ABORTING:
-			if (sm->auth_pae.eapLogoff && !sm->authAbort)
+			if (sm->auth_pae.eapolLogoff && !sm->authAbort)
 				SM_ENTER(AUTH_PAE, DISCONNECTED);
-			else if (!sm->auth_pae.eapLogoff && !sm->authAbort)
-				SM_ENTER(AUTH_PAE, CONNECTING);
+			else if (!sm->auth_pae.eapolLogoff && !sm->authAbort)
+				SM_ENTER(AUTH_PAE, RESTART);
 			break;
 		case AUTH_PAE_FORCE_AUTH:
-			if (sm->auth_pae.eapStart)
+			if (sm->auth_pae.eapolStart)
 				SM_ENTER(AUTH_PAE, FORCE_AUTH);
 			break;
 		case AUTH_PAE_FORCE_UNAUTH:
-			if (sm->auth_pae.eapStart)
+			if (sm->auth_pae.eapolStart)
 				SM_ENTER(AUTH_PAE, FORCE_UNAUTH);
 			break;
 		}
@@ -349,7 +338,8 @@ SM_STATE(BE_AUTH, INITIALIZE)
 {
 	SM_ENTRY(BE_AUTH, INITIALIZE, be_auth);
 
-	abortAuth;
+	abortAuth();
+	sm->be_auth.eapNoReq = FALSE;
 	sm->authAbort = FALSE;
 }
 
@@ -358,11 +348,9 @@ SM_STATE(BE_AUTH, REQUEST)
 {
 	SM_ENTRY(BE_AUTH, REQUEST, be_auth);
 
-	sm->currentId = sm->be_auth.idFromServer;
-	txReq(sm->currentId);
+	txReq();
+	sm->be_auth.eapReq = FALSE;
 	sm->be_auth.backendOtherRequestsToSupplicant++;
-	sm->aWhile = sm->be_auth.suppTimeout;
-	sm->be_auth.reqCount++;
 }
 
 
@@ -370,12 +358,12 @@ SM_STATE(BE_AUTH, RESPONSE)
 {
 	SM_ENTRY(BE_AUTH, RESPONSE, be_auth);
 
-	sm->be_auth.aReq = sm->be_auth.aSuccess = FALSE;
 	sm->authTimeout = FALSE;
-	sm->be_auth.rxResp = sm->be_auth.aFail = FALSE;
+	sm->eapolEap = FALSE;
+	sm->be_auth.eapNoReq = FALSE;
 	sm->aWhile = sm->be_auth.serverTimeout;
-	sm->be_auth.reqCount = 0;
-	sendRespToServer;
+	sm->be_auth.eapResp = TRUE;
+	sendRespToServer();
 	sm->be_auth.backendResponses++;
 }
 
@@ -384,8 +372,7 @@ SM_STATE(BE_AUTH, SUCCESS)
 {
 	SM_ENTRY(BE_AUTH, SUCCESS, be_auth);
 
-	sm->currentId = sm->be_auth.idFromServer;
-	txReq(sm->currentId);
+	txReq();
 	sm->authSuccess = TRUE;
 	sm->keyRun = TRUE;
 }
@@ -395,11 +382,14 @@ SM_STATE(BE_AUTH, FAIL)
 {
 	SM_ENTRY(BE_AUTH, FAIL, be_auth);
 
-	sm->currentId = sm->be_auth.idFromServer;
+	/* Note: IEEE 802.1X-REV-d11 has unconditional txReq() here.
+	 * txCannelFail() is used as a workaround for the case where
+	 * authentication server does not include EAP-Message with
+	 * Access-Reject. */
 	if (sm->last_eap_radius == NULL)
-		txCannedFail(sm->currentId);
+		txCannedFail();
 	else
-		txReq(sm->currentId);
+		txReq();
 	sm->authFail = TRUE;
 }
 
@@ -408,8 +398,6 @@ SM_STATE(BE_AUTH, TIMEOUT)
 {
 	SM_ENTRY(BE_AUTH, TIMEOUT, be_auth);
 
-	if (sm->portStatus == Unauthorized)
-		txCannedFail(sm->currentId);
 	sm->authTimeout = TRUE;
 }
 
@@ -419,7 +407,14 @@ SM_STATE(BE_AUTH, IDLE)
 	SM_ENTRY(BE_AUTH, IDLE, be_auth);
 
 	sm->authStart = FALSE;
-	sm->be_auth.reqCount = 0;
+}
+
+
+SM_STATE(BE_AUTH, IGNORE)
+{
+	SM_ENTRY(BE_AUTH, IGNORE, be_auth);
+
+	sm->be_auth.eapNoReq = FALSE;
 }
 
 
@@ -435,27 +430,25 @@ SM_STEP(BE_AUTH)
 		SM_ENTER(BE_AUTH, IDLE);
 		break;
 	case BE_AUTH_REQUEST:
-		if (sm->aWhile == 0 &&
-		    sm->be_auth.reqCount != sm->be_auth.maxReq)
-			SM_ENTER(BE_AUTH, REQUEST);
-		else if (sm->be_auth.rxResp)
+		if (sm->eapolEap)
 			SM_ENTER(BE_AUTH, RESPONSE);
-		else if (sm->aWhile == 0 &&
-			 sm->be_auth.reqCount >= sm->be_auth.maxReq)
+		else if (sm->be_auth.eapReq)
+			SM_ENTER(BE_AUTH, REQUEST);
+		else if (sm->eapTimeout)
 			SM_ENTER(BE_AUTH, TIMEOUT);
 		break;
 	case BE_AUTH_RESPONSE:
-		if (sm->be_auth.aReq) {
+		if (sm->be_auth.eapNoReq)
+			SM_ENTER(BE_AUTH, IGNORE);
+		if (sm->be_auth.eapReq) {
 			sm->be_auth.backendAccessChallenges++;
 			SM_ENTER(BE_AUTH, REQUEST);
 		} else if (sm->aWhile == 0)
 			SM_ENTER(BE_AUTH, TIMEOUT);
-		else if (sm->be_auth.aFail) {
+		else if (sm->eapFail) {
 			sm->be_auth.backendAuthFails++;
 			SM_ENTER(BE_AUTH, FAIL);
-		} else if (sm->be_auth.aSuccess /* &&
-			((sm->keyTxEnabled && !sm->keyAvailable) ||
-			!sm->keyTxEnabled) */) {
+		} else if (sm->eapSuccess) {
 			sm->be_auth.backendAuthSuccesses++;
 			SM_ENTER(BE_AUTH, SUCCESS);
 		}
@@ -470,8 +463,20 @@ SM_STEP(BE_AUTH)
 		SM_ENTER(BE_AUTH, IDLE);
 		break;
 	case BE_AUTH_IDLE:
-		if (sm->authStart)
+		if (sm->eapFail && sm->authStart)
+			SM_ENTER(BE_AUTH, FAIL);
+		else if (sm->be_auth.eapReq && sm->authStart)
+			SM_ENTER(BE_AUTH, REQUEST);
+		else if (sm->eapSuccess && sm->authStart)
+			SM_ENTER(BE_AUTH, SUCCESS);
+		break;
+	case BE_AUTH_IGNORE:
+		if (sm->eapolEap)
 			SM_ENTER(BE_AUTH, RESPONSE);
+		else if (sm->be_auth.eapReq)
+			SM_ENTER(BE_AUTH, REQUEST);
+		else if (sm->eapTimeout)
+			SM_ENTER(BE_AUTH, TIMEOUT);
 		break;
 	}
 }
@@ -500,7 +505,7 @@ SM_STATE(REAUTH_TIMER, REAUTHENTICATE)
 SM_STEP(REAUTH_TIMER)
 {
 	if (sm->portControl != Auto || sm->initialize ||
-	    sm->portStatus == Unauthorized ||
+	    sm->authPortStatus == Unauthorized ||
 	    !sm->reauth_timer.reAuthEnabled) {
 		SM_ENTER(REAUTH_TIMER, INITIALIZE);
 		return;
@@ -531,7 +536,7 @@ SM_STATE(AUTH_KEY_TX, KEY_TRANSMIT)
 {
 	SM_ENTRY(AUTH_KEY_TX, KEY_TRANSMIT, auth_key_tx);
 
-	txKey(sm->currentId);
+	txKey();
 	sm->keyAvailable = FALSE;
 	sm->keyDone = TRUE;
 }
@@ -561,6 +566,85 @@ SM_STEP(AUTH_KEY_TX)
 
 
 
+/* Key Receive state machine */
+
+SM_STATE(KEY_RX, NO_KEY_RECEIVE)
+{
+	SM_ENTRY(KEY_RX, NO_KEY_RECEIVE, key_rx);
+}
+
+
+SM_STATE(KEY_RX, KEY_RECEIVE)
+{
+	SM_ENTRY(KEY_RX, KEY_RECEIVE, key_rx);
+
+	processKey();
+	sm->key_rx.rxKey = FALSE;
+}
+
+
+SM_STEP(KEY_RX)
+{
+	if (sm->initialize || !sm->portEnabled) {
+		SM_ENTER(KEY_RX, NO_KEY_RECEIVE);
+		return;
+	}
+
+	switch (sm->key_rx.state) {
+	case KEY_RX_NO_KEY_RECEIVE:
+		if (sm->key_rx.rxKey)
+			SM_ENTER(KEY_RX, KEY_RECEIVE);
+		break;
+	case KEY_RX_KEY_RECEIVE:
+		if (sm->key_rx.rxKey)
+			SM_ENTER(KEY_RX, KEY_RECEIVE);
+		break;
+	}
+}
+
+
+
+/* Controlled Directions state machine */
+
+SM_STATE(CTRL_DIR, FORCE_BOTH)
+{
+	SM_ENTRY(CTRL_DIR, FORCE_BOTH, ctrl_dir);
+	sm->ctrl_dir.operControlledDirections = Both;
+}
+
+
+SM_STATE(CTRL_DIR, IN_OR_BOTH)
+{
+	SM_ENTRY(CTRL_DIR, IN_OR_BOTH, ctrl_dir);
+	sm->ctrl_dir.operControlledDirections =
+		sm->ctrl_dir.adminControlledDirections;
+}
+
+
+SM_STEP(CTRL_DIR)
+{
+	if (sm->initialize) {
+		SM_ENTER(CTRL_DIR, IN_OR_BOTH);
+		return;
+	}
+
+	switch (sm->ctrl_dir.state) {
+	case CTRL_DIR_FORCE_BOTH:
+		if (sm->portEnabled && sm->ctrl_dir.operEdge)
+			SM_ENTER(CTRL_DIR, IN_OR_BOTH);
+		break;
+	case CTRL_DIR_IN_OR_BOTH:
+		if (sm->ctrl_dir.operControlledDirections !=
+		    sm->ctrl_dir.adminControlledDirections)
+			SM_ENTER(CTRL_DIR, IN_OR_BOTH);
+		if (!sm->portEnabled || !sm->ctrl_dir.operEdge)
+			SM_ENTER(CTRL_DIR, FORCE_BOTH);
+		break;
+	}
+}
+
+
+
 struct eapol_state_machine *
 eapol_sm_alloc(hostapd *hapd, struct sta_info *sta)
 {
@@ -583,23 +667,25 @@ eapol_sm_alloc(hostapd *hapd, struct sta_info *sta)
 	/* Set default values for state machine constants */
 	sm->auth_pae.state = AUTH_PAE_INITIALIZE;
 	sm->auth_pae.quietPeriod = AUTH_PAE_DEFAULT_quietPeriod;
-	sm->auth_pae.initialEAPMsg = AUTH_PAE_DEFAULT_initialEAPMsg;
 	sm->auth_pae.reAuthMax = AUTH_PAE_DEFAULT_reAuthMax;
-	sm->auth_pae.txPeriod = AUTH_PAE_DEFAULT_txPeriod;
 
 	sm->be_auth.state = BE_AUTH_INITIALIZE;
-	sm->be_auth.suppTimeout = BE_AUTH_DEFAULT_suppTimeout;
 	sm->be_auth.serverTimeout = BE_AUTH_DEFAULT_serverTimeout;
-	sm->be_auth.maxReq = BE_AUTH_DEFAULT_maxReq;
 
 	sm->reauth_timer.state = REAUTH_TIMER_INITIALIZE;
-	sm->reauth_timer.reAuthPeriod = REAUTH_TIMER_DEFAULT_reAuthPeriod;
-	sm->reauth_timer.reAuthEnabled = REAUTH_TIMER_DEFAULT_reAuthEnabled;
+	sm->reauth_timer.reAuthPeriod = hapd->conf->eap_reauth_period;
+	sm->reauth_timer.reAuthEnabled = hapd->conf->eap_reauth_period > 0 ?
+		TRUE : FALSE;
+
+	sm->auth_key_tx.state = AUTH_KEY_TX_NO_KEY_TRANSMIT;
+
+	sm->key_rx.state = KEY_RX_NO_KEY_RECEIVE;
+
+	sm->ctrl_dir.state = CTRL_DIR_IN_OR_BOTH;
 
 	sm->portEnabled = FALSE;
 	sm->portControl = Auto;
 
-	/* IEEE 802.1aa/D4 */
 	sm->keyAvailable = FALSE;
 	if (!hapd->conf->wpa &&
 	    (hapd->default_wep_key || hapd->conf->individual_wep_key_len > 0))
@@ -611,6 +697,18 @@ eapol_sm_alloc(hostapd *hapd, struct sta_info *sta)
 	else
 		sm->portValid = TRUE;
 
+	if (hapd->conf->eap_authenticator) {
+		struct eap_config eap_conf;
+		memset(&eap_conf, 0, sizeof(eap_conf));
+		eap_conf.ssl_ctx = hapd->ssl_ctx;
+		eap_conf.eap_sim_db_priv = hapd->eap_sim_db_priv;
+		sm->eap = eap_sm_init(sm, &eapol_cb, &eap_conf);
+		if (sm->eap == NULL) {
+			eapol_sm_free(sm);
+			return NULL;
+		}
+	}
+
 	eapol_sm_initialize(sm);
 
 	return sm;
@@ -623,7 +721,8 @@ void eapol_sm_free(struct eapol_state_machine *sm)
 		return;
 
 	eloop_cancel_timeout(eapol_port_timers_tick, sm->hapd, sm);
-
+	if (sm->eap)
+		eap_sm_deinit(sm->eap);
 	free(sm);
 }
 
@@ -641,19 +740,23 @@ static int eapol_sm_sta_entry_alive(struct hostapd_data *hapd, u8 *addr)
 void eapol_sm_step(struct eapol_state_machine *sm)
 {
 	struct hostapd_data *hapd = sm->hapd;
-	u8 addr[6];
-	int prev_auth_pae, prev_be_auth, prev_reauth_timer, prev_auth_key_tx;
+	u8 addr[ETH_ALEN];
+	int prev_auth_pae, prev_be_auth, prev_reauth_timer, prev_auth_key_tx,
+		prev_key_rx, prev_ctrl_dir;
 
 	/* FIX: could re-run eapol_sm_step from registered timeout (after
 	 * 0 sec) to make sure that other possible timeouts/events are
 	 * processed */
 
-	memcpy(addr, sm->sta->addr, 6);
+	memcpy(addr, sm->sta->addr, ETH_ALEN);
+restart:
 	do {
 		prev_auth_pae = sm->auth_pae.state;
 		prev_be_auth = sm->be_auth.state;
 		prev_reauth_timer = sm->reauth_timer.state;
 		prev_auth_key_tx = sm->auth_key_tx.state;
+		prev_key_rx = sm->key_rx.state;
+		prev_ctrl_dir = sm->ctrl_dir.state;
 
 		SM_STEP_RUN(AUTH_PAE);
 		if (!eapol_sm_sta_entry_alive(hapd, addr))
@@ -667,10 +770,23 @@ void eapol_sm_step(struct eapol_state_machine *sm)
 		SM_STEP_RUN(AUTH_KEY_TX);
 		if (!eapol_sm_sta_entry_alive(hapd, addr))
 			break;
+		SM_STEP_RUN(KEY_RX);
+		if (!eapol_sm_sta_entry_alive(hapd, addr))
+			break;
+		SM_STEP_RUN(CTRL_DIR);
+		if (!eapol_sm_sta_entry_alive(hapd, addr))
+			break;
 	} while (prev_auth_pae != sm->auth_pae.state ||
 		 prev_be_auth != sm->be_auth.state ||
 		 prev_reauth_timer != sm->reauth_timer.state ||
-		 prev_auth_key_tx != sm->auth_key_tx.state);
+		 prev_auth_key_tx != sm->auth_key_tx.state ||
+		 prev_key_rx != sm->key_rx.state ||
+		 prev_ctrl_dir != sm->ctrl_dir.state);
+
+	if (eapol_sm_sta_entry_alive(hapd, addr) && sm->eap) {
+		if (eap_sm_step(sm->eap))
+			goto restart;
+	}
 
 	if (eapol_sm_sta_entry_alive(hapd, addr))
 		wpa_sm_notify(sm->hapd, sm->sta);
@@ -714,6 +830,16 @@ static inline const char * port_state_txt(PortState ps)
 }
 
 
+static inline const char * ctrl_dir_txt(ControlledDirection dir)
+{
+	switch (dir) {
+	case Both: return "Both";
+	case In: return "In";
+	default: return "Unknown";
+	}
+}
+
+
 static inline const char * auth_pae_state_txt(int s)
 {
 	switch (s) {
@@ -726,15 +852,7 @@ static inline const char * auth_pae_state_txt(int s)
 	case AUTH_PAE_HELD: return "HELD";
 	case AUTH_PAE_FORCE_AUTH: return "FORCE_AUTH";
 	case AUTH_PAE_FORCE_UNAUTH: return "FORCE_UNAUTH";
-	default: return "Unknown";
-	}
-}
-
-
-static inline const char * auth_pae_eaptype_txt(EAPMsgType t)
-{
-	switch (t) {
-	case EAPRequestIdentity: return "EAP Request/Identity";
+	case AUTH_PAE_RESTART: return "RESTART";
 	default: return "Unknown";
 	}
 }
@@ -750,6 +868,7 @@ static inline const char * be_auth_state_txt(int s)
 	case BE_AUTH_TIMEOUT: return "TIMEOUT";
 	case BE_AUTH_IDLE: return "IDLE";
 	case BE_AUTH_INITIALIZE: return "INITIALIZE";
+	case BE_AUTH_IGNORE: return "IGNORE";
 	default: return "Unknown";
 	}
 }
@@ -775,63 +894,78 @@ static inline const char * auth_key_tx_state_txt(int s)
 }
 
 
+static inline const char * key_rx_state_txt(int s)
+{
+	switch (s) {
+	case KEY_RX_NO_KEY_RECEIVE: return "NO_KEY_RECEIVE";
+	case KEY_RX_KEY_RECEIVE: return "KEY_RECEIVE";
+	default: return "Unknown";
+	}
+}
+
+
+static inline const char * ctrl_dir_state_txt(int s)
+{
+	switch (s) {
+	case CTRL_DIR_FORCE_BOTH: return "FORCE_BOTH";
+	case CTRL_DIR_IN_OR_BOTH: return "IN_OR_BOTH";
+	default: return "Unknown";
+	}
+}
+
+
 void eapol_sm_dump_state(FILE *f, const char *prefix,
 			 struct eapol_state_machine *sm)
 {
 	fprintf(f, "%sEAPOL state machine:\n", prefix);
-	fprintf(f, "%s  aWhile=%d quietWhile=%d reAuthWhen=%d "
-		"txWhen=%d\n", prefix,
-		sm->aWhile, sm->quietWhile, sm->reAuthWhen, sm->txWhen);
+	fprintf(f, "%s  aWhile=%d quietWhile=%d reAuthWhen=%d\n", prefix,
+		sm->aWhile, sm->quietWhile, sm->reAuthWhen);
 #define _SB(b) ((b) ? "TRUE" : "FALSE")
-	fprintf(f, "%s  authAbort=%s authFail=%s authStart=%s "
-		"authTimeout=%s\n", prefix,
-		_SB(sm->authAbort), _SB(sm->authFail), _SB(sm->authStart),
-		_SB(sm->authTimeout));
-	fprintf(f, "%s  authSuccess=%s currentId=%d initialize=%s "
-		"reAuthenticate=%s\n", prefix, _SB(sm->authSuccess),
-		sm->currentId, _SB(sm->initialize), _SB(sm->reAuthenticate));
-	fprintf(f, "%s  portControl=%s portEnabled=%s portStatus=%s\n", prefix,
-		port_type_txt(sm->portControl), _SB(sm->portEnabled),
-		port_state_txt(sm->portStatus));
-	fprintf(f, "%s  keyAvailable=%s keyTxEnabled=%s portValid=%s\n",
-		prefix, _SB(sm->keyAvailable), _SB(sm->keyTxEnabled),
-		_SB(sm->portValid));
-	fprintf(f, "%s  keyRun=%s keyDone=%s\n",
-		prefix, _SB(sm->keyRun), _SB(sm->keyDone));
+	fprintf(f,
+		"%s  authAbort=%s authFail=%s authPortStatus=%s authStart=%s\n"
+		"%s  authTimeout=%s authSuccess=%s eapFail=%s eapolEap=%s\n"
+		"%s  eapSuccess=%s eapTimeout=%s initialize=%s "
+		"keyAvailable=%s\n"
+		"%s  keyDone=%s keyRun=%s keyTxEnabled=%s portControl=%s\n"
+		"%s  portEnabled=%s portValid=%s reAuthenticate=%s\n",
+		prefix, _SB(sm->authAbort), _SB(sm->authFail),
+		port_state_txt(sm->authPortStatus), _SB(sm->authStart),
+		prefix, _SB(sm->authTimeout), _SB(sm->authSuccess),
+		_SB(sm->eapFail), _SB(sm->eapolEap),
+		prefix, _SB(sm->eapSuccess), _SB(sm->eapTimeout),
+		_SB(sm->initialize), _SB(sm->keyAvailable),
+		prefix, _SB(sm->keyDone), _SB(sm->keyRun),
+		_SB(sm->keyTxEnabled), port_type_txt(sm->portControl),
+		prefix, _SB(sm->portEnabled), _SB(sm->portValid),
+		_SB(sm->reAuthenticate));
 
 	fprintf(f, "%s  Authenticator PAE:\n"
 		"%s    state=%s\n"
-		"%s    eapLogoff=%s eapStart=%s portMode=%s\n"
-		"%s    reAuthCount=%d rxInitialRsp=%s\n"
-		"%s    quietPeriod=%d initialEAPMsg=%s\n"
-		"%s    reAuthMax=%d txPeriod=%d\n"
+		"%s    eapolLogoff=%s eapolStart=%s eapRestart=%s\n"
+		"%s    portMode=%s reAuthCount=%d\n"
+		"%s    quietPeriod=%d reAuthMax=%d\n"
 		"%s    authEntersConnecting=%d\n"
 		"%s    authEapLogoffsWhileConnecting=%d\n"
 		"%s    authEntersAuthenticating=%d\n"
 		"%s    authAuthSuccessesWhileAuthenticating=%d\n"
 		"%s    authAuthTimeoutsWhileAuthenticating=%d\n"
 		"%s    authAuthFailWhileAuthenticating=%d\n"
-		"%s    authAuthReauthsWhileAuthenticating=%d\n"
 		"%s    authAuthEapStartsWhileAuthenticating=%d\n"
 		"%s    authAuthEapLogoffWhileAuthenticating=%d\n"
 		"%s    authAuthReauthsWhileAuthenticated=%d\n"
 		"%s    authAuthEapStartsWhileAuthenticated=%d\n"
 		"%s    authAuthEapLogoffWhileAuthenticated=%d\n",
 		prefix, prefix, auth_pae_state_txt(sm->auth_pae.state), prefix,
-		_SB(sm->auth_pae.eapLogoff), _SB(sm->auth_pae.eapStart),
-		port_type_txt(sm->auth_pae.portMode),
-		prefix, sm->auth_pae.reAuthCount,
-		_SB(sm->auth_pae.rxInitialRsp),
-		prefix, sm->auth_pae.quietPeriod,
-		auth_pae_eaptype_txt(sm->auth_pae.initialEAPMsg),
-		prefix, sm->auth_pae.reAuthMax, sm->auth_pae.txPeriod,
+		_SB(sm->auth_pae.eapolLogoff), _SB(sm->auth_pae.eapolStart),
+		_SB(sm->auth_pae.eapRestart), prefix,
+		port_type_txt(sm->auth_pae.portMode), sm->auth_pae.reAuthCount,
+		prefix, sm->auth_pae.quietPeriod, sm->auth_pae.reAuthMax,
 		prefix, sm->auth_pae.authEntersConnecting,
 		prefix, sm->auth_pae.authEapLogoffsWhileConnecting,
 		prefix, sm->auth_pae.authEntersAuthenticating,
 		prefix, sm->auth_pae.authAuthSuccessesWhileAuthenticating,
 		prefix, sm->auth_pae.authAuthTimeoutsWhileAuthenticating,
 		prefix, sm->auth_pae.authAuthFailWhileAuthenticating,
-		prefix, sm->auth_pae.authAuthReauthsWhileAuthenticating,
 		prefix, sm->auth_pae.authAuthEapStartsWhileAuthenticating,
 		prefix, sm->auth_pae.authAuthEapLogoffWhileAuthenticating,
 		prefix, sm->auth_pae.authAuthReauthsWhileAuthenticated,
@@ -840,27 +974,21 @@ void eapol_sm_dump_state(FILE *f, const char *prefix,
 
 	fprintf(f, "%s  Backend Authentication:\n"
 		"%s    state=%s\n"
-		"%s    reqCount=%d rxResp=%s aSuccess=%s aFail=%s aReq=%s\n"
-		"%s    idFromServer=%d\n"
-		"%s    suppTimeout=%d serverTimeout=%d maxReq=%d\n"
+		"%s    eapNoReq=%s eapReq=%s eapResp=%s\n"
+		"%s    serverTimeout=%d\n"
 		"%s    backendResponses=%d\n"
 		"%s    backendAccessChallenges=%d\n"
 		"%s    backendOtherRequestsToSupplicant=%d\n"
-		"%s    backendNonNakResponsesFromSupplicant=%d\n"
 		"%s    backendAuthSuccesses=%d\n"
 		"%s    backendAuthFails=%d\n",
 		prefix, prefix,
-		be_auth_state_txt(sm->be_auth.state), prefix,
-		sm->be_auth.reqCount,
-		_SB(sm->be_auth.rxResp), _SB(sm->be_auth.aSuccess),
-		_SB(sm->be_auth.aFail), _SB(sm->be_auth.aReq),
-		prefix, sm->be_auth.idFromServer, prefix,
-		sm->be_auth.suppTimeout, sm->be_auth.serverTimeout,
-		sm->be_auth.maxReq,
+		be_auth_state_txt(sm->be_auth.state),
+		prefix, _SB(sm->be_auth.eapNoReq), _SB(sm->be_auth.eapReq),
+		_SB(sm->be_auth.eapResp),
+		prefix, sm->be_auth.serverTimeout,
 		prefix, sm->be_auth.backendResponses,
 		prefix, sm->be_auth.backendAccessChallenges,
 		prefix, sm->be_auth.backendOtherRequestsToSupplicant,
-		prefix, sm->be_auth.backendNonNakResponsesFromSupplicant,
 		prefix, sm->be_auth.backendAuthSuccesses,
 		prefix, sm->be_auth.backendAuthFails);
 
@@ -874,6 +1002,182 @@ void eapol_sm_dump_state(FILE *f, const char *prefix,
 	fprintf(f, "%s  Authenticator Key Transmit:\n"
 		"%s    state=%s\n", prefix, prefix,
 		auth_key_tx_state_txt(sm->auth_key_tx.state));
+
+	fprintf(f, "%s  Key Receive:\n"
+		"%s    state=%s\n"
+		"%s    rxKey=%s\n", prefix, prefix,
+		key_rx_state_txt(sm->key_rx.state),
+		prefix, _SB(sm->key_rx.rxKey));
+
+	fprintf(f, "%s  Controlled Directions:\n"
+		"%s    state=%s\n"
+		"%s    adminControlledDirections=%s "
+		"operControlledDirections=%s\n"
+		"%s    operEdge=%s\n", prefix, prefix,
+		ctrl_dir_state_txt(sm->ctrl_dir.state),
+		prefix, ctrl_dir_txt(sm->ctrl_dir.adminControlledDirections),
+		ctrl_dir_txt(sm->ctrl_dir.operControlledDirections),
+		prefix, _SB(sm->ctrl_dir.operEdge));
 #undef _SB
 }
 #endif /* HOSTAPD_DUMP_STATE */
+
+
+static Boolean eapol_sm_get_bool(void *ctx, enum eapol_bool_var variable)
+{
+	struct eapol_state_machine *sm = ctx;
+	if (sm == NULL)
+		return FALSE;
+	switch (variable) {
+	case EAPOL_eapSuccess:
+		return sm->eapSuccess;
+	case EAPOL_eapRestart:
+		return sm->auth_pae.eapRestart;
+	case EAPOL_eapFail:
+		return sm->eapFail;
+	case EAPOL_eapResp:
+		return sm->be_auth.eapResp;
+	case EAPOL_eapReq:
+		return sm->be_auth.eapReq;
+	case EAPOL_eapNoReq:
+		return sm->be_auth.eapNoReq;
+	case EAPOL_portEnabled:
+		return sm->portEnabled;
+	case EAPOL_eapTimeout:
+		return sm->eapTimeout;
+	}
+	return FALSE;
+}
+
+
+static void eapol_sm_set_bool(void *ctx, enum eapol_bool_var variable,
+			      Boolean value)
+{
+	struct eapol_state_machine *sm = ctx;
+	if (sm == NULL)
+		return;
+	switch (variable) {
+	case EAPOL_eapSuccess:
+		sm->eapSuccess = value;
+		break;
+	case EAPOL_eapRestart:
+		sm->auth_pae.eapRestart = value;
+		break;
+	case EAPOL_eapFail:
+		sm->eapFail = value;
+		break;
+	case EAPOL_eapResp:
+		sm->be_auth.eapResp = value;
+		break;
+	case EAPOL_eapReq:
+		sm->be_auth.eapReq = value;
+		break;
+	case EAPOL_eapNoReq:
+		sm->be_auth.eapNoReq = value;
+		break;
+	case EAPOL_portEnabled:
+		sm->portEnabled = value;
+		break;
+	case EAPOL_eapTimeout:
+		sm->eapTimeout = value;
+		break;
+	}
+}
+
+
+static void eapol_sm_set_eapReqData(void *ctx, const u8 *eapReqData,
+				    size_t eapReqDataLen)
+{
+	struct eapol_state_machine *sm = ctx;
+	if (sm == NULL)
+		return;
+
+	free(sm->last_eap_radius);
+	sm->last_eap_radius = malloc(eapReqDataLen);
+	if (sm->last_eap_radius == NULL)
+		return;
+	memcpy(sm->last_eap_radius, eapReqData, eapReqDataLen);
+	sm->last_eap_radius_len = eapReqDataLen;
+}
+
+
+static void eapol_sm_set_eapKeyData(void *ctx, const u8 *eapKeyData,
+				    size_t eapKeyDataLen)
+{
+	struct eapol_state_machine *sm = ctx;
+	struct hostapd_data *hapd;
+
+	if (sm == NULL)
+		return;
+
+	hapd = sm->hapd;
+
+	if (eapKeyData && eapKeyDataLen >= 64) {
+		free(sm->eapol_key_sign);
+		free(sm->eapol_key_crypt);
+		sm->eapol_key_crypt = malloc(32);
+		if (sm->eapol_key_crypt) {
+			memcpy(sm->eapol_key_crypt, eapKeyData, 32);
+			sm->eapol_key_crypt_len = 32;
+		}
+		sm->eapol_key_sign = malloc(32);
+		if (sm->eapol_key_sign) {
+			memcpy(sm->eapol_key_sign, eapKeyData + 32, 32);
+			sm->eapol_key_sign_len = 32;
+		}
+		if (hapd->default_wep_key ||
+		    hapd->conf->individual_wep_key_len > 0 ||
+		    hapd->conf->wpa)
+			sm->keyAvailable = TRUE;
+	} else {
+		free(sm->eapol_key_sign);
+		free(sm->eapol_key_crypt);
+		sm->eapol_key_sign = NULL;
+		sm->eapol_key_crypt = NULL;
+		sm->eapol_key_sign_len = 0;
+		sm->eapol_key_crypt_len = 0;
+		sm->keyAvailable = FALSE;
+	}
+}
+
+
+static int eapol_sm_get_eap_user(void *ctx, const u8 *identity,
+				 size_t identity_len, int phase2,
+				 struct eap_user *user)
+{
+	struct eapol_state_machine *sm = ctx;
+	const struct hostapd_eap_user *eap_user;
+
+	eap_user = hostapd_get_eap_user(sm->hapd->conf, identity,
+					identity_len, phase2);
+	if (eap_user == NULL)
+		return -1;
+
+	memset(user, 0, sizeof(*user));
+	user->phase2 = phase2;
+	memcpy(user->methods, eap_user->methods,
+	       EAP_USER_MAX_METHODS > EAP_MAX_METHODS ?
+	       EAP_USER_MAX_METHODS : EAP_MAX_METHODS);
+
+	if (eap_user->password) {
+		user->password = malloc(eap_user->password_len);
+		if (user->password == NULL)
+			return -1;
+		memcpy(user->password, eap_user->password,
+		       eap_user->password_len);
+		user->password_len = eap_user->password_len;
+	}
+	user->force_version = eap_user->force_version;
+
+	return 0;
+}
+
+
+static struct eapol_callbacks eapol_cb =
+{
+	.get_bool = eapol_sm_get_bool,
+	.set_bool = eapol_sm_set_bool,
+	.set_eapReqData = eapol_sm_set_eapReqData,
+	.set_eapKeyData = eapol_sm_set_eapKeyData,
+	.get_eap_user = eapol_sm_get_eap_user,
+};
