@@ -1,6 +1,6 @@
 /*
  * Event loop
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -21,6 +21,10 @@
 #include <errno.h>
 #include <signal.h>
 
+#ifdef CONFIG_NATIVE_WINDOWS
+#include "common.h"
+#endif /* CONFIG_NATIVE_WINDOWS */
+
 #include "eloop.h"
 
 
@@ -57,6 +61,7 @@ struct eloop_data {
 	int signal_count;
 	struct eloop_signal *signals;
 	int signaled;
+	int pending_terminate;
 
 	int terminate;
 };
@@ -198,10 +203,33 @@ int eloop_cancel_timeout(void (*handler)(void *eloop_ctx, void *sock_ctx),
 }
 
 
+#ifndef CONFIG_NATIVE_WINDOWS
+static void eloop_handle_alarm(int sig)
+{
+	fprintf(stderr, "eloop: could not process SIGINT or SIGTERM in two "
+		"seconds. Looks like there\n"
+		"is a bug that ends up in a busy loop that "
+		"prevents clean shutdown.\n"
+		"Killing program forcefully.\n");
+	exit(1);
+}
+#endif /* CONFIG_NATIVE_WINDOWS */
+
+
 static void eloop_handle_signal(int sig)
 {
 	int i;
 
+#ifndef CONFIG_NATIVE_WINDOWS
+	if ((sig == SIGINT || sig == SIGTERM) && !eloop.pending_terminate) {
+		/* Use SIGALRM to break out from potential busy loops that
+		 * would not allow the program to be killed. */
+		eloop.pending_terminate = 1;
+		signal(SIGALRM, eloop_handle_alarm);
+		alarm(2);
+	}
+#endif /* CONFIG_NATIVE_WINDOWS */
+
 	eloop.signaled++;
 	for (i = 0; i < eloop.signal_count; i++) {
 		if (eloop.signals[i].sig == sig) {
@@ -220,6 +248,13 @@ static void eloop_process_pending_signals(void)
 		return;
 	eloop.signaled = 0;
 
+	if (eloop.pending_terminate) {
+#ifndef CONFIG_NATIVE_WINDOWS
+		alarm(0);
+#endif /* CONFIG_NATIVE_WINDOWS */
+		eloop.pending_terminate = 0;
+	}
+
 	for (i = 0; i < eloop.signal_count; i++) {
 		if (eloop.signals[i].signaled) {
 			eloop.signals[i].signaled = 0;
