@@ -1,12 +1,16 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -33,7 +37,12 @@
 #include "sta_info.h"
 #include "driver.h"
 #include "radius_client.h"
+#include "radius_server.h"
 #include "wpa.h"
+#include "ctrl_iface.h"
+#include "tls.h"
+#include "eap_sim_db.h"
+#include "version.h"
 
 
 struct hapd_interfaces {
@@ -44,6 +53,11 @@ struct hapd_interfaces {
 unsigned char rfc1042_header[6] = { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };
 
 
+extern int wpa_debug_level;
+extern int wpa_debug_show_keys;
+extern int wpa_debug_timestamp;
+
+
 void hostapd_logger(hostapd *hapd, u8 *addr, unsigned int module, int level,
 		    char *fmt, ...)
 {
@@ -83,6 +97,12 @@ void hostapd_logger(hostapd *hapd, u8 *addr, unsigned int module, int level,
 	case HOSTAPD_MODULE_WPA:
 		module_str = "WPA";
 		break;
+	case HOSTAPD_MODULE_DRIVER:
+		module_str = "DRIVER";
+		break;
+	case HOSTAPD_MODULE_IAPP:
+		module_str = "IAPP";
+		break;
 	default:
 		module_str = NULL;
 		break;
@@ -140,19 +160,26 @@ void hostapd_logger(hostapd *hapd, u8 *addr, unsigned int module, int level,
 
 static void hostapd_deauth_all_stas(hostapd *hapd)
 {
+#if 0
 	u8 addr[ETH_ALEN];
 
 	memset(addr, 0xff, ETH_ALEN);
-	ieee802_11_send_deauth(hapd, addr, WLAN_REASON_PREV_AUTH_NOT_VALID);
+	hostapd_sta_deauth(hapd, addr, WLAN_REASON_PREV_AUTH_NOT_VALID);
+#else
+	/* New Prism2.5/3 STA firmware versions seem to have issues with this
+	 * broadcast deauth frame. This gets the firmware in odd state where
+	 * nothing works correctly, so let's skip sending this for a while
+	 * until the issue has been resolved. */
+#endif
 }
 
 
 /* This function will be called whenever a station associates with the AP */
 void hostapd_new_assoc_sta(hostapd *hapd, struct sta_info *sta)
 {
-	/* IEEE 802.11f (IAPP) */
+	/* IEEE 802.11F (IAPP) */
 	if (hapd->conf->ieee802_11f)
-		iapp_new_station(hapd, sta);
+		iapp_new_station(hapd->iapp, sta);
 
 	/* Start accounting here, if IEEE 802.1X is not used. IEEE 802.1X code
 	 * will start accounting after the station has been authorized. */
@@ -191,7 +218,7 @@ static void handle_reload(int sig, void *eloop_ctx, void *signal_ctx)
 		/* TODO: update dynamic data based on changed configuration
 		 * items (e.g., open/close sockets, remove stations added to
 		 * deny list, etc.) */
-		radius_client_flush(hapd);
+		radius_client_flush(hapd->radius);
 		hostapd_config_free(hapd->conf);
 		hapd->conf = newconf;
 	}
@@ -205,6 +232,7 @@ static void hostapd_dump_state(hostapd *hapd)
 	time_t now;
 	struct sta_info *sta;
 	int i;
+	char *buf;
 
 	if (!hapd->conf->dump_log_name) {
 		printf("Dump file not defined - ignoring dump request\n");
@@ -259,6 +287,25 @@ static void hostapd_dump_state(hostapd *hapd)
 		ieee802_1x_dump_state(f, "  ", sta);
 	}
 
+	buf = malloc(4096);
+	if (buf) {
+		int count = radius_client_get_mib(hapd->radius, buf, 4096);
+		if (count < 0)
+			count = 0;
+		else if (count > 4095)
+			count = 4095;
+		buf[count] = '\0';
+		fprintf(f, "%s", buf);
+
+		count = radius_server_get_mib(hapd->radius_srv, buf, 4096);
+		if (count < 0)
+			count = 0;
+		else if (count > 4095)
+			count = 4095;
+		buf[count] = '\0';
+		fprintf(f, "%s", buf);
+		free(buf);
+	}
 	fclose(f);
 }
 #endif /* HOSTAPD_DUMP_STATE */
@@ -278,88 +325,41 @@ static void handle_dump_state(int sig, void *eloop_ctx, void *signal_ctx)
 }
 
 
-static void hostapd_set_broadcast_wep(hostapd *hapd)
+static void hostapd_cleanup(struct hostapd_data *hapd)
 {
-	if (hapd->conf->default_wep_key_len < 1)
-		return;
+	hostapd_ctrl_iface_deinit(hapd);
 
-	hapd->default_wep_key = malloc(hapd->conf->default_wep_key_len);
-	if (hapd->default_wep_key == NULL ||
-	    hostapd_get_rand(hapd->default_wep_key,
-			     hapd->conf->default_wep_key_len)) {
-		printf("Could not generate random WEP key.\n");
-		free(hapd->default_wep_key);
-		exit(1);
-	}
-}
-
-
-/* The rekeying function: generate a new broadcast WEP key, rotate
- * the key index, and direct Key Transmit State Machines of all of the
- * authenticators to send a new key to the authenticated stations.
- */
-static void hostapd_rotate_wep(void *eloop_ctx, void *timeout_ctx)
-{
-	struct sta_info *s;
-	hostapd *hapd = eloop_ctx;
-
-	if (hapd->default_wep_key)
-		free(hapd->default_wep_key);
-
-	if (hapd->default_wep_key_idx >= 3)
-		hapd->default_wep_key_idx =
-			hapd->conf->individual_wep_key_len > 0 ? 1 : 0;
-	else
-		hapd->default_wep_key_idx++;
-
-	hostapd_set_broadcast_wep(hapd);
-
-	for (s = hapd->sta_list; s != NULL; s = s->next)
-		if (s->eapol_sm)
-			s->eapol_sm->keyAvailable = TRUE;
-
-	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL)) {
-		hostapd_hexdump("New WEP key generated",
-				hapd->default_wep_key,
-				hapd->conf->default_wep_key_len);
-	}
-
-	/* TODO: Could setup key for RX here, but change default TX keyid only
-	 * after new broadcast key has been sent to all stations. */
-	if (hostapd_set_encryption(hapd, "WEP", NULL,
-    				   hapd->default_wep_key_idx,
-			 	   hapd->default_wep_key,
-			 	   hapd->conf->default_wep_key_len)) {
-		printf("Could not set WEP encryption.\n");
-	}
-
-	if (hapd->conf->wep_rekeying_period > 0)
-		eloop_register_timeout(hapd->conf->wep_rekeying_period, 0,
-				       hostapd_rotate_wep, hapd, NULL);
-}
-
-
-static void hostapd_cleanup(hostapd *hapd)
-{
 	free(hapd->default_wep_key);
-	if (hapd->conf->ieee802_11f)
-		iapp_deinit(hapd);
+	hapd->default_wep_key = NULL;
+	iapp_deinit(hapd->iapp);
 	accounting_deinit(hapd);
 	wpa_deinit(hapd);
 	ieee802_1x_deinit(hapd);
 	hostapd_acl_deinit(hapd);
-	radius_client_deinit(hapd);
+	radius_client_deinit(hapd->radius);
+	hapd->radius = NULL;
+	radius_server_deinit(hapd->radius_srv);
+	hapd->radius_srv = NULL;
 
-	if (hapd->sock >= 0)
-		close(hapd->sock);
-	if (hapd->ioctl_sock >= 0)
-		close(hapd->ioctl_sock);
 	hostapd_wireless_event_deinit(hapd);
 
+	if (hapd->driver)
+		hostapd_driver_deinit(hapd);
+
 	hostapd_config_free(hapd->conf);
 	hapd->conf = NULL;
 
 	free(hapd->config_fname);
+
+#ifdef EAP_TLS_FUNCS
+	if (hapd->ssl_ctx) {
+		tls_deinit(hapd->ssl_ctx);
+		hapd->ssl_ctx = NULL;
+	}
+#endif /* EAP_TLS_FUNCS */
+
+	if (hapd->eap_sim_db_priv)
+		eap_sim_db_deinit(hapd->eap_sim_db_priv);
 }
 
 
@@ -379,80 +379,88 @@ static int hostapd_flush_old_stations(hostapd *hapd)
 }
 
 
-static int hostapd_setup_encryption(hostapd *hapd)
+static int hostapd_setup_interface(struct hostapd_data *hapd)
 {
-	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL))
-		hostapd_hexdump("Default WEP key", hapd->default_wep_key,
-				hapd->conf->default_wep_key_len);
+	struct hostapd_config *conf = hapd->conf;
+	u8 ssid[HOSTAPD_SSID_LEN + 1];
+	int ssid_len, set_ssid;
+	int ret = 0;
 
-	hostapd_set_encryption(hapd, "none", NULL, 0, NULL, 0);
-
-	if (hostapd_set_encryption(hapd, "WEP", NULL,
-				   hapd->default_wep_key_idx,
-				   hapd->default_wep_key,
-				   hapd->conf->default_wep_key_len)) {
-		printf("Could not set WEP encryption.\n");
+	if (hostapd_driver_init(hapd)) {
+		printf("%s driver initialization failed.\n",
+			hapd->driver ? hapd->driver->name : "Unknown");
+		hapd->driver = NULL;
 		return -1;
 	}
 
-	/* Setup rekeying timer. */
-	if (hapd->conf->wep_rekeying_period > 0 &&
-	    (hapd->default_wep_key || hapd->conf->individual_wep_key_len > 0)
-	    && eloop_register_timeout(hapd->conf->wep_rekeying_period, 0,
-				      hostapd_rotate_wep, hapd, NULL)) {
-		printf("Couldn't set rekeying timer.\n");
+	/*
+	 * Fetch the SSID from the system and use it or,
+	 * if one was specified in the config file, verify they
+	 * match.
+	 */
+	ssid_len = hostapd_get_ssid(hapd, ssid, sizeof(ssid));
+	if (ssid_len < 0) {
+		printf("Could not read SSID from system\n");
 		return -1;
 	}
-
-	return 0;
-}
-
-
-static int hostapd_setup_interface(hostapd *hapd)
-{
-	hapd->ioctl_sock = socket(PF_INET, SOCK_DGRAM, 0);
-	if (hapd->ioctl_sock < 0) {
-		perror("socket[PF_INET,SOCK_DGRAM]");
-		return -1;
+	if (conf->ssid_set) {
+		/*
+		 * If SSID is specified in the config file and it differs
+		 * from what is being used then force installation of the
+		 * new SSID.
+		 */
+		set_ssid = (conf->ssid_len != ssid_len ||
+			    memcmp(conf->ssid, ssid, ssid_len) != 0);
+	} else {
+		/*
+		 * No SSID in the config file; just use the one we got
+		 * from the system.
+		 */
+		set_ssid = 0;
+		conf->ssid_len = ssid_len;
+		memcpy(conf->ssid, ssid, conf->ssid_len);
+		conf->ssid[conf->ssid_len] = '\0';
 	}
 
-	if (hostap_ioctl_prism2param(hapd, PRISM2_PARAM_HOSTAPD, 1)) {
-		printf("Could not enable hostapd mode for interface %s\n",
-		       hapd->conf->iface);
-		return -1;
-	}
-
-	if (hapd->conf->assoc_ap &&
-	    hostap_ioctl_prism2param(hapd, PRISM2_PARAM_HOSTAPD_STA, 1)) {
-		printf("Could not enable hostapd STA mode for interface %s\n",
-		       hapd->conf->iface);
-		return -1;
-	}
-
-	if (hostapd_init_sockets(hapd))
-		return -1;
-
-	printf("Using interface %sap with hwaddr " MACSTR " and ssid '%s'\n",
+	printf("Using interface %s with hwaddr " MACSTR " and ssid '%s'\n",
 	       hapd->conf->iface, MAC2STR(hapd->own_addr), hapd->conf->ssid);
 
+	if (hostapd_setup_wpa_psk(conf)) {
+		printf("WPA-PSK setup failed.\n");
+		return -1;
+	}
+
 	/* Set SSID for the kernel driver (to be used in beacon and probe
 	 * response frames) */
-	if (hostap_ioctl_setiwessid(hapd, hapd->conf->ssid,
-				    hapd->conf->ssid_len)) {
+	if (set_ssid && hostapd_set_ssid(hapd, (u8 *) conf->ssid,
+					 conf->ssid_len)) {
 		printf("Could not set SSID for kernel driver\n");
 		return -1;
 	}
 
-	if (radius_client_init(hapd)) {
+	hapd->radius = radius_client_init(hapd);
+	if (hapd->radius == NULL) {
 		printf("RADIUS client initialization failed.\n");
 		return -1;
 	}
-
+	if (conf->radius_server_clients) {
+		struct radius_server_conf srv;
+		memset(&srv, 0, sizeof(srv));
+		srv.client_file = conf->radius_server_clients;
+		srv.auth_port = conf->radius_server_auth_port;
+		srv.hostapd_conf = conf;
+		srv.eap_sim_db_priv = hapd->eap_sim_db_priv;
+		srv.ssl_ctx = hapd->ssl_ctx;
+		hapd->radius_srv = radius_server_init(&srv);
+		if (hapd->radius_srv == NULL) {
+			printf("RADIUS server initialization failed.\n");
+			return -1;
+		}
+	}
 	if (hostapd_acl_init(hapd)) {
 		printf("ACL initialization failed.\n");
 		return -1;
 	}
-
 	if (ieee802_1x_init(hapd)) {
 		printf("IEEE 802.1X initialization failed.\n");
 		return -1;
@@ -468,37 +476,131 @@ static int hostapd_setup_interface(hostapd *hapd)
 		return -1;
 	}
 
-	if (hapd->conf->ieee802_11f && iapp_init(hapd)) {
-		printf("IEEE 802.11f (IAPP) initialization failed.\n");
+	if (hapd->conf->ieee802_11f &&
+	    (hapd->iapp = iapp_init(hapd, hapd->conf->iapp_iface)) == NULL) {
+		printf("IEEE 802.11F (IAPP) initialization failed.\n");
 		return -1;
 	}
 
 	if (hostapd_wireless_event_init(hapd) < 0)
 		return -1;
 
-	if (hapd->default_wep_key && hostapd_setup_encryption(hapd))
-		return -1;
-
 	if (hostapd_flush_old_stations(hapd))
 		return -1;
 
-	return 0;
+	if (hostapd_ctrl_iface_init(hapd)) {
+		printf("Failed to setup control interface\n");
+		ret = -1;
+	}
+
+	return ret;
+}
+
+
+struct driver {
+	struct driver *next;
+	char *name;
+	const struct driver_ops *ops;
+};
+static struct driver *drivers = NULL;
+
+void driver_register(const char *name, const struct driver_ops *ops)
+{
+	struct driver *d;
+
+	d = malloc(sizeof(struct driver));
+	if (d == NULL) {
+		printf("Failed to register driver %s!\n", name);
+		return;
+	}
+	d->name = strdup(name);
+	if (d->name == NULL) {
+		printf("Failed to register driver %s!\n", name);
+		free(d);
+		return;
+	}
+	d->ops = ops;
+
+	d->next = drivers;
+	drivers = d;
+}
+
+
+void driver_unregister(const char *name)
+{
+	struct driver *p, **pp;
+
+	for (pp = &drivers; (p = *pp) != NULL; pp = &p->next) {
+		if (strcasecmp(p->name, name) == 0) {
+			*pp = p->next;
+			p->next = NULL;
+			free(p->name);
+			free(p);
+			break;
+		}
+	}
+}
+
+
+static void driver_unregister_all(void)
+{
+	struct driver *p, *pp;
+	p = drivers;
+	drivers = NULL;
+	while (p) {
+		pp = p;
+		p = p->next;
+		free(pp->name);
+		free(pp);
+	}
+}
+
+
+const struct driver_ops * driver_lookup(const char *name)
+{
+	struct driver *p;
+
+	if (strcmp(name, "default") == 0) {
+		p = drivers;
+		while (p && p->next)
+			p = p->next;
+		return p->ops;
+	}
+
+	for (p = drivers; p != NULL; p = p->next) {
+		if (strcasecmp(p->name, name) == 0)
+			return p->ops;
+	}
+
+	return NULL;
+}
+
+
+static void show_version(void)
+{
+	fprintf(stderr,
+		"hostapd v" VERSION_STR "\n"
+		"Host AP user space daemon for management functionality of "
+		"Host AP kernel driver\n"
+		"Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi> "
+		"and contributors\n");
 }
 
 
 static void usage(void)
 {
+	show_version();
 	fprintf(stderr,
-		"Host AP user space daemon for management functionality of "
-		"Host AP kernel driver\n"
-		"Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>\n"
 		"\n"
-		"usage: hostapd [-hdB] <configuration file>\n"
+		"usage: hostapd [-hdB] <configuration file(s)>\n"
 		"\n"
 		"options:\n"
 		"   -h   show this usage\n"
 		"   -d   show more debug messages (-dd for even more)\n"
-		"   -B   run daemon in the background\n");
+		"   -B   run daemon in the background\n"
+		"   -K   include key data in debug messages\n"
+		"   -t   include timestamps in some debug messages\n"
+		"   -v   show hostapd version\n");
 
 	exit(1);
 }
@@ -511,22 +613,19 @@ static hostapd * hostapd_init(const char *config_file)
 	hapd = malloc(sizeof(*hapd));
 	if (hapd == NULL) {
 		printf("Could not allocate memory for hostapd data\n");
-		exit(1);
+		goto fail;
 	}
 	memset(hapd, 0, sizeof(*hapd));
-	hapd->sock = hapd->ioctl_sock = -1;
-
 
 	hapd->config_fname = strdup(config_file);
 	if (hapd->config_fname == NULL) {
 		printf("Could not allocate memory for config_fname\n");
-		exit(1);
+		goto fail;
 	}
 
 	hapd->conf = hostapd_config_read(hapd->config_fname);
 	if (hapd->conf == NULL) {
-		free(hapd->config_fname);
-		exit(1);
+		goto fail;
 	}
 
 	if (hapd->conf->individual_wep_key_len > 0) {
@@ -534,21 +633,75 @@ static hostapd * hostapd_init(const char *config_file)
 		hapd->default_wep_key_idx = 1;
 	}
 
+#ifdef EAP_TLS_FUNCS
+	if (hapd->conf->eap_authenticator &&
+	    (hapd->conf->ca_cert || hapd->conf->server_cert)) {
+		hapd->ssl_ctx = tls_init();
+		if (hapd->ssl_ctx == NULL) {
+			printf("Failed to initialize TLS\n");
+			goto fail;
+		}
+		if (tls_global_ca_cert(hapd->ssl_ctx, hapd->conf->ca_cert)) {
+			printf("Failed to load CA certificate (%s)\n",
+				hapd->conf->ca_cert);
+			goto fail;
+		}
+		if (tls_global_client_cert(hapd->ssl_ctx,
+					   hapd->conf->server_cert)) {
+			printf("Failed to load server certificate (%s)\n",
+				hapd->conf->server_cert);
+			goto fail;
+		}
+		if (tls_global_private_key(hapd->ssl_ctx,
+					   hapd->conf->private_key,
+					   hapd->conf->private_key_passwd)) {
+			printf("Failed to load private key (%s)\n",
+			       hapd->conf->private_key);
+		}
+	}
+#endif /* EAP_TLS_FUNCS */
+
+	if (hapd->conf->eap_sim_db) {
+		hapd->eap_sim_db_priv =
+			eap_sim_db_init(hapd->conf->eap_sim_db);
+		if (hapd->eap_sim_db_priv == NULL) {
+			printf("Failed to initialize EAP-SIM database "
+			       "interface\n");
+			goto fail;
+		}
+	}
+
 	if (hapd->conf->assoc_ap)
 		hapd->assoc_ap_state = WAIT_BEACON;
 
+	/* FIX: need to fix this const vs. not */
+	hapd->driver = (struct driver_ops *) hapd->conf->driver;
+
 	return hapd;
+
+fail:
+	if (hapd) {
+		if (hapd->ssl_ctx)
+			tls_deinit(hapd->ssl_ctx);
+		if (hapd->conf)
+			hostapd_config_free(hapd->conf);
+		free(hapd->config_fname);
+		free(hapd);
+	}
+	return NULL;
 }
 
 
+void register_drivers(void);
+
 int main(int argc, char *argv[])
 {
 	struct hapd_interfaces interfaces;
-	int ret = 1, i;
+	int ret = 1, i, j;
 	int c, debug = 0, daemonize = 0;
 
 	for (;;) {
-		c = getopt(argc, argv, "Bdh");
+		c = getopt(argc, argv, "BdhKtv");
 		if (c < 0)
 			break;
 		switch (c) {
@@ -561,6 +714,16 @@ int main(int argc, char *argv[])
 		case 'B':
 			daemonize++;
 			break;
+		case 'K':
+			wpa_debug_show_keys++;
+			break;
+		case 't':
+			wpa_debug_timestamp++;
+			break;
+		case 'v':
+			show_version();
+			exit(1);
+			break;
 
 		default:
 			usage();
@@ -571,6 +734,8 @@ int main(int argc, char *argv[])
 	if (optind == argc)
 		usage();
 
+	register_drivers();		/* NB: generated by Makefile */
+
 	interfaces.count = argc - optind;
 
 	interfaces.hapd = malloc(interfaces.count * sizeof(hostapd *));
@@ -590,9 +755,15 @@ int main(int argc, char *argv[])
 		interfaces.hapd[i] = hostapd_init(argv[optind + i]);
 		if (!interfaces.hapd[i])
 			goto out;
-		hostapd_set_broadcast_wep(interfaces.hapd[i]);
+		for (j = 0; j < debug; j++) {
+			if (interfaces.hapd[i]->conf->logger_stdout_level > 0)
+				interfaces.hapd[i]->conf->
+					logger_stdout_level--;
+			interfaces.hapd[i]->conf->debug++;
+		}
 		if (hostapd_setup_interface(interfaces.hapd[i]))
 			goto out;
+		wpa_debug_level -= interfaces.hapd[0]->conf->debug;
 	}
 
 	if (daemonize && daemon(0, 0)) {
@@ -615,11 +786,6 @@ int main(int argc, char *argv[])
 	for (i = 0; i < interfaces.count; i++) {
 		if (!interfaces.hapd[i])
 			continue;
-		(void) hostapd_set_iface_flags(interfaces.hapd[i], 0);
-		(void) hostap_ioctl_prism2param(interfaces.hapd[i],
-						PRISM2_PARAM_HOSTAPD, 0);
-		(void) hostap_ioctl_prism2param(interfaces.hapd[i],
-						PRISM2_PARAM_HOSTAPD_STA, 0);
 
 		hostapd_cleanup(interfaces.hapd[i]);
 		free(interfaces.hapd[i]);
@@ -630,5 +796,7 @@ int main(int argc, char *argv[])
 
 	closelog();
 
+	driver_unregister_all();
+
 	return ret;
 }
