@@ -17,6 +17,12 @@
 #include "hostap_common.h"
 #include "config.h"
 
+struct ieee8023_hdr {
+	u8 dest[6];
+	u8 src[6];
+	u16 ethertype;
+} __attribute__ ((packed));
+
 
 struct ieee80211_hdr {
 	u16 frame_control;
@@ -46,16 +52,14 @@ extern unsigned char rfc1042_header[6];
 
 typedef struct hostapd_data hostapd;
 
-struct driver_info {
-	void *data; /* driver specific data - each driver can store data for
-		     * its own use in this pointer */
-	int (*send_mgmt_frame)(void *priv, const void *msg, size_t len,
-			       int flags);
-	int (*send_eapol)(void *driver, u8 *addr, u8 *data, size_t data_len,
-			  int encrypt);
-	void (*set_sta_authorized)(void *driver, u8 *addr, int authorized);
+struct hostap_sta_driver_data {
+	unsigned long rx_packets, tx_packets, rx_bytes, tx_bytes;
 };
 
+struct driver_ops;
+struct wpa_ctrl_dst;
+struct radius_server_data;
+
 struct hostapd_data {
 	struct hostapd_config *conf;
 	char *config_fname;
@@ -72,17 +76,15 @@ struct hostapd_data {
 	 */
 	struct sta_info *sta_aid[MAX_AID_TABLE_SIZE];
 
-	struct driver_info driver;
+	struct driver_ops *driver;
 
 	u8 *default_wep_key;
 	u8 default_wep_key_idx;
 
 	struct radius_client_data *radius;
+	u32 acct_session_id_hi, acct_session_id_lo;
 
-	u16 iapp_identifier; /* next IAPP identifier */
-	struct in_addr iapp_own, iapp_broadcast;
-	int iapp_udp_sock;
-	int iapp_packet_sock;
+	struct iapp_data *iapp;
 
 	enum { DO_NOT_ASSOC = 0, WAIT_BEACON, AUTHENTICATE, ASSOCIATE,
 	       ASSOCIATED } assoc_ap_state;
@@ -107,9 +109,14 @@ struct hostapd_data {
 	time_t michael_mic_failure;
 	int michael_mic_failures;
 	int tkip_countermeasures;
-};
 
-	struct driver_info *driver;
+	int ctrl_sock;
+	struct wpa_ctrl_dst *ctrl_dst;
+
+	void *ssl_ctx;
+	void *eap_sim_db_priv;
+	struct radius_server_data *radius_srv;
+};
 
 void hostapd_new_assoc_sta(hostapd *hapd, struct sta_info *sta);
 void hostapd_logger(hostapd *hapd, u8 *addr, unsigned int module, int level,
@@ -118,7 +125,7 @@ void hostapd_logger(hostapd *hapd, u8 *addr, unsigned int module, int level,
 
 #define HOSTAPD_DEBUG(level, args...) \
 do { \
-	if (hapd->conf->debug >= (level)) \
+	if (hapd->conf == NULL || hapd->conf->debug >= (level)) \
 		printf(args); \
 } while (0)
 
