@@ -1,17 +1,22 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver / IEEE 802.11 Management
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <sys/socket.h>
 #include <net/if.h>
 #include <arpa/inet.h>
 #include <time.h>
@@ -23,12 +28,12 @@
 #include "radius_client.h"
 #include "ieee802_11_auth.h"
 #include "sta_info.h"
-#include "driver.h"
 #include "eapol_sm.h"
 #include "rc4.h"
 #include "ieee802_1x.h"
 #include "wpa.h"
 #include "accounting.h"
+#include "driver.h"
 
 
 static u8 * hostapd_eid_supp_rates(hostapd *hapd, u8 *eid)
@@ -60,8 +65,10 @@ static u16 hostapd_own_capab_info(hostapd *hapd)
 
 static const u8 WPA_OUI_TYPE[] = { 0x00, 0x50, 0xf2, 1 };
 
-ParseRes ieee802_11_parse_elems(hostapd *hapd, u8 *start, size_t len,
-				struct ieee802_11_elems *elems)
+ParseRes ieee802_11_parse_elems(struct hostapd_data *hapd, u8 *start,
+				size_t len,
+				struct ieee802_11_elems *elems,
+				int show_errors)
 {
 	size_t left = len;
 	u8 *pos = start;
@@ -77,10 +84,15 @@ ParseRes ieee802_11_parse_elems(hostapd *hapd, u8 *start, size_t len,
 		left -= 2;
 
 		if (elen > left) {
-			HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-				      "IEEE 802.11 element parse failed "
-				      "(id=%d elen=%d left=%d)\n",
-				      id, elen, left);
+			if (show_errors) {
+				HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
+					      "IEEE 802.11 element parse "
+					      "failed (id=%d elen=%d "
+					      "left=%lu)\n",
+					      id, elen, (unsigned long) left);
+				if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL))
+					hostapd_hexdump("IEs", start, len);
+			}
 			return ParseFailed;
 		}
 
@@ -121,7 +133,7 @@ ParseRes ieee802_11_parse_elems(hostapd *hapd, u8 *start, size_t len,
 			if (elen > 4 && memcmp(pos, WPA_OUI_TYPE, 4) == 0) {
 				elems->wpa_ie = pos;
 				elems->wpa_ie_len = elen;
-			} else {
+			} else if (show_errors) {
 				HOSTAPD_DEBUG(HOSTAPD_DEBUG_EXCESSIVE,
 					      "IEEE 802.11 element parse "
 					      "ignored unknown generic element"
@@ -133,6 +145,8 @@ ParseRes ieee802_11_parse_elems(hostapd *hapd, u8 *start, size_t len,
 					      elen >= 3 ? pos[2] : 0,
 					      elen >= 4 ? pos[3] : 0);
 				unknown++;
+			} else {
+				unknown++;
 			}
 			break;
 		case WLAN_EID_RSN:
@@ -140,11 +154,13 @@ ParseRes ieee802_11_parse_elems(hostapd *hapd, u8 *start, size_t len,
 			elems->rsn_ie_len = elen;
 			break;
 		default:
+			unknown++;
+			if (!show_errors)
+				break;
 			HOSTAPD_DEBUG(HOSTAPD_DEBUG_EXCESSIVE,
 				      "IEEE 802.11 element parse ignored "
 				      "unknown element (id=%d elen=%d)\n",
 				      id, elen);
-			unknown++;
 			break;
 		}
 
@@ -171,40 +187,6 @@ static void ieee802_11_print_ssid(u8 *ssid, u8 len)
 }
 
 
-void ieee802_11_send_deauth(hostapd *hapd, u8 *addr, u8 reason)
-{
-	struct ieee80211_mgmt mgmt;
-
-	memset(&mgmt, 0, sizeof(mgmt));
-	mgmt.frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
-					  WLAN_FC_STYPE_DEAUTH);
-	memcpy(mgmt.da, addr, ETH_ALEN);
-	memcpy(mgmt.sa, hapd->own_addr, ETH_ALEN);
-	memcpy(mgmt.bssid, hapd->own_addr, ETH_ALEN);
-	mgmt.u.deauth.reason_code = host_to_le16(reason);
-	if (send(hapd->sock, &mgmt, IEEE80211_HDRLEN + sizeof(mgmt.u.deauth),
-		 0) < 0)
-		perror("ieee802_11_send_deauth: send");
-}
-
-
-void ieee802_11_send_disassoc(hostapd *hapd, u8 *addr, u8 reason)
-{
-	struct ieee80211_mgmt mgmt;
-
-	memset(&mgmt, 0, sizeof(mgmt));
-	mgmt.frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
-					  WLAN_FC_STYPE_DISASSOC);
-	memcpy(mgmt.da, addr, ETH_ALEN);
-	memcpy(mgmt.sa, hapd->own_addr, ETH_ALEN);
-	memcpy(mgmt.bssid, hapd->own_addr, ETH_ALEN);
-	mgmt.u.disassoc.reason_code = host_to_le16(reason);
-	if (send(hapd->sock, &mgmt, IEEE80211_HDRLEN +
-		 sizeof(mgmt.u.disassoc), 0) < 0)
-		perror("ieee802_11_send_disassoc: send");
-}
-
-
 static void ieee802_11_sta_authenticate(void *eloop_ctx, void *timeout_ctx)
 {
 	hostapd *hapd = eloop_ctx;
@@ -231,8 +213,8 @@ static void ieee802_11_sta_authenticate(void *eloop_ctx, void *timeout_ctx)
 	mgmt.u.auth.auth_alg = host_to_le16(WLAN_AUTH_OPEN);
 	mgmt.u.auth.auth_transaction = host_to_le16(1);
 	mgmt.u.auth.status_code = host_to_le16(0);
-	if (send(hapd->sock, &mgmt, IEEE80211_HDRLEN + sizeof(mgmt.u.auth),
-		 0) < 0)
+	if (hostapd_send_mgmt_frame(hapd, &mgmt, IEEE80211_HDRLEN +
+				    sizeof(mgmt.u.auth), 0) < 0)
 		perror("ieee802_11_sta_authenticate: send");
 
 	/* Try to authenticate again, if this attempt fails or times out. */
@@ -276,7 +258,7 @@ static void ieee802_11_sta_associate(void *eloop_ctx, void *timeout_ctx)
 
 	p = hostapd_eid_supp_rates(hapd, p);
 
-	if (send(hapd->sock, mgmt, p - (u8 *) mgmt, 0) < 0)
+	if (hostapd_send_mgmt_frame(hapd, mgmt, p - (u8 *) mgmt, 0) < 0)
 		perror("ieee802_11_sta_associate: send");
 
 	/* Try to authenticate again, if this attempt fails or times out. */
@@ -284,44 +266,6 @@ static void ieee802_11_sta_associate(void *eloop_ctx, void *timeout_ctx)
 }
 
 
-static struct sta_info * auth_get_sta(hostapd *hapd, u8 *addr)
-{
-	struct sta_info *sta;
-
-	sta = ap_get_sta(hapd, addr);
-	if (sta)
-		return sta;
-
-	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "  New STA\n");
-	if (hapd->num_sta >= MAX_STA_COUNT) {
-		/* FIX: might try to remove some old STAs first? */
-		printf("  no more room for new STAs (%d/%d)\n",
-		       hapd->num_sta, MAX_STA_COUNT);
-		return NULL;
-	}
-
-	sta = (struct sta_info *) malloc(sizeof(struct sta_info));
-	if (sta == NULL) {
-		printf("  malloc failed\n");
-		return NULL;
-	}
-	memset(sta, 0, sizeof(struct sta_info));
-	sta->radius_identifier = -1;
-	sta->acct_session_id_lo = hapd->radius->acct_session_id_lo++;
-	sta->acct_interim_interval = hapd->conf->radius_acct_interim_interval;
-
-	/* initialize STA info data */
-	eloop_register_timeout(AP_MAX_INACTIVITY, 0, ap_handle_timer,
-			       hapd, sta);
-	memcpy(sta->addr, addr, ETH_ALEN);
-	sta->next = hapd->sta_list;
-	hapd->sta_list = sta;
-	hapd->num_sta++;
-	ap_sta_hash_add(hapd, sta);
-	return sta;
-}
-
-
 static u16 auth_shared_key(hostapd *hapd, struct sta_info *sta,
 			   u16 auth_transaction, u8 *challenge, int iswep)
 {
@@ -419,7 +363,7 @@ static void send_auth_reply(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 		      "auth_transaction=%d resp=%d%s\n",
 		      MAC2STR(mgmt->sa), auth_alg, auth_transaction,
 		      resp, challenge ? " challenge" : "");
-	if (send(hapd->sock, reply, rlen, 0) < 0)
+	if (hostapd_send_mgmt_frame(hapd, reply, rlen, 0) < 0)
 		perror("send_auth_reply: send");
 }
 
@@ -435,7 +379,8 @@ static void handle_auth(hostapd *hapd, struct ieee80211_mgmt *mgmt, size_t len)
 	u32 session_timeout, acct_interim_interval;
 
 	if (len < IEEE80211_HDRLEN + sizeof(mgmt->u.auth)) {
-		printf("handle_auth - too short payload (len=%d)\n", len);
+		printf("handle_auth - too short payload (len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
@@ -496,6 +441,13 @@ static void handle_auth(hostapd *hapd, struct ieee80211_mgmt *mgmt, size_t len)
 		goto fail;
 	}
 
+	if (memcmp(mgmt->sa, hapd->own_addr, ETH_ALEN) == 0) {
+		printf("Station " MACSTR " not allowed to authenticate.\n",
+		       MAC2STR(mgmt->sa));
+		resp = WLAN_STATUS_UNSPECIFIED_FAILURE;
+		goto fail;
+	}
+
 	res = hostapd_allowed_address(hapd, mgmt->sa, (u8 *) mgmt, len,
 				      &session_timeout,
 				      &acct_interim_interval);
@@ -515,12 +467,13 @@ static void handle_auth(hostapd *hapd, struct ieee80211_mgmt *mgmt, size_t len)
 		return;
 	}
 
-	sta = auth_get_sta(hapd, mgmt->sa);
+	sta = ap_sta_add(hapd, mgmt->sa);
 	if (!sta) {
 		resp = WLAN_STATUS_UNSPECIFIED_FAILURE;
 		goto fail;
 	}
 	sta->flags &= ~WLAN_STA_PREAUTH;
+	ieee802_1x_notify_pre_auth(sta->eapol_sm, 0);
 
 	if (hapd->conf->radius_acct_interim_interval == 0 &&
 	    acct_interim_interval)
@@ -568,8 +521,8 @@ static void handle_assoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 
 	if (len < IEEE80211_HDRLEN + (reassoc ? sizeof(mgmt->u.reassoc_req) :
 				      sizeof(mgmt->u.assoc_req))) {
-		printf("handle_assoc(reassoc=%d) - too short payload (len=%d)"
-		       "\n", reassoc, len);
+		printf("handle_assoc(reassoc=%d) - too short payload (len=%lu)"
+		       "\n", reassoc, (unsigned long) len);
 		return;
 	}
 
@@ -618,8 +571,8 @@ static void handle_assoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	sta->capability = capab_info;
 
 	/* followed by SSID and Supported rates */
-	if (ieee802_11_parse_elems(hapd, pos, left, &elems) == ParseFailed ||
-	    !elems.ssid) {
+	if (ieee802_11_parse_elems(hapd, pos, left, &elems, 1) == ParseFailed
+	    || !elems.ssid) {
 		printf("STA " MACSTR " sent invalid association request\n",
 		       MAC2STR(sta->addr));
 		resp = WLAN_STATUS_UNSPECIFIED_FAILURE;
@@ -632,7 +585,8 @@ static void handle_assoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 		       "unknown SSID '", MAC2STR(sta->addr));
 		ieee802_11_print_ssid(elems.ssid, elems.ssid_len);
 		printf("'\n");
-		return;
+		resp = WLAN_STATUS_UNSPECIFIED_FAILURE;
+		goto fail;
 	}
 
 	if (elems.supp_rates) {
@@ -640,7 +594,8 @@ static void handle_assoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 			printf("STA " MACSTR ": Invalid supported rates "
 			       "element length %d\n", MAC2STR(sta->addr),
 			       elems.supp_rates_len);
-			return;
+			resp = WLAN_STATUS_UNSPECIFIED_FAILURE;
+			goto fail;
 		}
 
 		memset(sta->supported_rates, 0, sizeof(sta->supported_rates));
@@ -776,7 +731,7 @@ static void handle_assoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 		mgmt->frame_control |= host_to_le16(BIT(1));
 	}
 
-	if (send(hapd->sock, mgmt, send_len, 0) < 0)
+	if (hostapd_send_mgmt_frame(hapd, mgmt, send_len, 0) < 0)
 		perror("handle_assoc: send");
 }
 
@@ -785,7 +740,6 @@ static void handle_assoc_resp(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 			      size_t len)
 {
 	u16 status_code, aid;
-	struct prism2_hostapd_param param;
 
 	if (hapd->assoc_ap_state != ASSOCIATE) {
 		printf("Unexpected association response received from " MACSTR
@@ -794,8 +748,8 @@ static void handle_assoc_resp(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	}
 
 	if (len < IEEE80211_HDRLEN + sizeof(mgmt->u.assoc_resp)) {
-		printf("handle_assoc_resp - too short payload (len=%d)\n",
-		       len);
+		printf("handle_assoc_resp - too short payload (len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
@@ -826,10 +780,7 @@ static void handle_assoc_resp(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	hapd->assoc_ap_aid = aid;
 	hapd->assoc_ap_state = ASSOCIATED;
 
-	memset(&param, 0, sizeof(param));
-	param.cmd = PRISM2_HOSTAPD_SET_ASSOC_AP_ADDR;
-	memcpy(param.sta_addr, hapd->conf->assoc_ap_addr, ETH_ALEN);
-	if (hostapd_ioctl(hapd, &param, sizeof(param))) {
+	if (hostapd_set_assoc_ap(hapd, hapd->conf->assoc_ap_addr)) {
 		printf("Could not set associated AP address to kernel "
 		       "driver.\n");
 	}
@@ -842,7 +793,8 @@ static void handle_disassoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	struct sta_info *sta;
 
 	if (len < IEEE80211_HDRLEN + sizeof(mgmt->u.disassoc)) {
-		printf("handle_disassoc - too short payload (len=%d)\n", len);
+		printf("handle_disassoc - too short payload (len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
@@ -881,7 +833,15 @@ static void handle_disassoc(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	 * authenticated. */
 	accounting_sta_stop(hapd, sta);
 	ieee802_1x_free_station(sta);
-	remove_sta(hapd, sta);
+	hostapd_sta_remove(hapd, sta->addr);
+
+	if (sta->timeout_next == STA_NULLFUNC ||
+	    sta->timeout_next == STA_DISASSOC) {
+		sta->timeout_next = STA_DEAUTH;
+		eloop_cancel_timeout(ap_handle_timer, hapd, sta);
+		eloop_register_timeout(AP_DEAUTH_DELAY, 0, ap_handle_timer,
+				       hapd, sta);
+	}
 }
 
 
@@ -891,7 +851,8 @@ static void handle_deauth(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	struct sta_info *sta;
 
 	if (len < IEEE80211_HDRLEN + sizeof(mgmt->u.deauth)) {
-		printf("handle_deauth - too short payload (len=%d)\n", len);
+		printf("handle_deauth - too short payload (len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
@@ -935,13 +896,15 @@ static void handle_beacon(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	struct ieee802_11_elems elems;
 
 	if (len < IEEE80211_HDRLEN + sizeof(mgmt->u.beacon)) {
-		printf("handle_beacon - too short payload (len=%d)\n", len);
+		printf("handle_beacon - too short payload (len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
 	(void) ieee802_11_parse_elems(hapd, mgmt->u.beacon.variable,
 				      len - (IEEE80211_HDRLEN +
-					     sizeof(mgmt->u.beacon)), &elems);
+					     sizeof(mgmt->u.beacon)), &elems,
+				      0);
 
 	if (hapd->assoc_ap_state == WAIT_BEACON &&
 	    memcmp(mgmt->sa, hapd->conf->assoc_ap_addr, ETH_ALEN) == 0) {
@@ -969,7 +932,7 @@ static void handle_beacon(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 }
 
 
-void ieee802_11_mgmt(hostapd *hapd, char *buf, size_t len, u16 stype)
+void ieee802_11_mgmt(struct hostapd_data *hapd, u8 *buf, size_t len, u16 stype)
 {
 	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *) buf;
 
@@ -1045,7 +1008,8 @@ static void handle_auth_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	}
 
 	if (len < IEEE80211_HDRLEN + sizeof(mgmt->u.auth)) {
-		printf("handle_auth_cb - too short payload (len=%d)\n", len);
+		printf("handle_auth_cb - too short payload (len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
@@ -1074,7 +1038,6 @@ static void handle_assoc_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 			    size_t len, int reassoc, int ok)
 {
 	u16 status;
-	struct prism2_hostapd_param param;
 	struct sta_info *sta;
 	int new_assoc = 1;
 
@@ -1088,7 +1051,7 @@ static void handle_assoc_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	if (len < IEEE80211_HDRLEN + (reassoc ? sizeof(mgmt->u.reassoc_req) :
 				      sizeof(mgmt->u.assoc_req))) {
 		printf("handle_assoc_cb(reassoc=%d) - too short payload "
-		       "(len=%d)\n", reassoc, len);
+		       "(len=%lu)\n", reassoc, (unsigned long) len);
 		return;
 	}
 
@@ -1107,19 +1070,23 @@ static void handle_assoc_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	if (status != WLAN_STATUS_SUCCESS)
 		goto fail;
 
+	/* Stop previous accounting session, if one is started, and allocate
+	 * new session id for the new session. */
+	accounting_sta_stop(hapd, sta);
+	accounting_sta_get_id(hapd, sta);
+
 	hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_IEEE80211,
-		       HOSTAPD_LEVEL_INFO, "associated (aid %d)", sta->aid);
+		       HOSTAPD_LEVEL_INFO,
+		       "associated (aid %d, accounting session %08X-%08X)",
+		       sta->aid, sta->acct_session_id_hi,
+		       sta->acct_session_id_lo);
+
 	if (sta->flags & WLAN_STA_ASSOC)
 		new_assoc = 0;
 	sta->flags |= WLAN_STA_ASSOC;
 
-	memset(&param, 0, sizeof(param));
-	param.cmd = PRISM2_HOSTAPD_ADD_STA;
-	memcpy(param.sta_addr, sta->addr, ETH_ALEN);
-	param.u.add_sta.aid = sta->aid;
-	param.u.add_sta.capability = sta->capability;
-	param.u.add_sta.tx_supp_rates = sta->tx_supp_rates;
-	if (hostapd_ioctl(hapd, &param, sizeof(param))) {
+	if (hostapd_sta_add(hapd, sta->addr, sta->aid, sta->capability,
+			    sta->tx_supp_rates)) {
 		printf("Could not add station to kernel driver.\n");
 	}
 
@@ -1129,8 +1096,7 @@ static void handle_assoc_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 	else
 		wpa_sm_event(hapd, sta, WPA_REAUTH);
 
-	if (sta->eapol_sm)
-		sta->eapol_sm->portEnabled = TRUE;
+	ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
 
  fail:
 	/* Copy of the association request is not needed anymore */
@@ -1141,8 +1107,8 @@ static void handle_assoc_cb(hostapd *hapd, struct ieee80211_mgmt *mgmt,
 }
 
 
-void ieee802_11_mgmt_cb(hostapd *hapd, char *buf, size_t len, u16 stype,
-			int ok)
+void ieee802_11_mgmt_cb(struct hostapd_data *hapd, u8 *buf, size_t len,
+			u16 stype, int ok)
 {
 	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *) buf;
 
@@ -1185,24 +1151,45 @@ static void ieee80211_tkip_countermeasures_start(struct hostapd_data *hapd)
 	hostapd_logger(hapd, NULL, HOSTAPD_MODULE_IEEE80211,
 		       HOSTAPD_LEVEL_INFO, "TKIP countermeasures initiated");
 
+	if (hapd->wpa_auth)
+		hapd->wpa_auth->dot11RSNATKIPCounterMeasuresInvoked++;
 	hapd->tkip_countermeasures = 1;
 	wpa_gtk_rekey(hapd);
 	eloop_cancel_timeout(ieee80211_tkip_countermeasures_stop, hapd, NULL);
 	eloop_register_timeout(60, 0, ieee80211_tkip_countermeasures_stop,
 			       hapd, NULL);
 	for (sta = hapd->sta_list; sta != NULL; sta = sta->next) {
-		ieee802_11_send_deauth(hapd, sta->addr,
-				       WLAN_REASON_MICHAEL_MIC_FAILURE);
+		hostapd_sta_deauth(hapd, sta->addr,
+				   WLAN_REASON_MICHAEL_MIC_FAILURE);
 		sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC |
 				WLAN_STA_AUTHORIZED);
-		remove_sta(hapd, sta);
+		hostapd_sta_remove(hapd, sta->addr);
 	}
 }
 
 
-void ieee80211_michael_mic_failure(struct hostapd_data *hapd)
+void ieee80211_michael_mic_failure(struct hostapd_data *hapd, u8 *addr,
+				   int local)
 {
 	time_t now;
+
+	if (addr && local) {
+		struct sta_info *sta = ap_get_sta(hapd, addr);
+		if (sta != NULL) {
+			sta->dot11RSNAStatsTKIPLocalMICFailures++;
+			hostapd_logger(hapd, addr, HOSTAPD_MODULE_IEEE80211,
+				       HOSTAPD_LEVEL_INFO,
+				       "Michael MIC failure detected in "
+				       "received frame");
+		} else {
+			HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
+				      "MLME-MICHAELMICFAILURE.indication "
+				      "for not associated STA (" MACSTR
+				      ") ignored\n", MAC2STR(addr));
+			return;
+		}
+	}
+
 	time(&now);
 	if (now > hapd->michael_mic_failure + 60) {
 		hapd->michael_mic_failures = 1;
@@ -1213,3 +1200,18 @@ void ieee80211_michael_mic_failure(struct hostapd_data *hapd)
 	}
 	hapd->michael_mic_failure = now;
 }
+
+
+int ieee802_11_get_mib(struct hostapd_data *hapd, char *buf, size_t buflen)
+{
+	/* TODO */
+	return 0;
+}
+
+
+int ieee802_11_get_mib_sta(struct hostapd_data *hapd, struct sta_info *sta,
+			   char *buf, size_t buflen)
+{
+	/* TODO */
+	return 0;
+}
