@@ -1,12 +1,16 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver / IEEE 802.1X Authenticator
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -21,7 +25,6 @@
 #include <sys/time.h>
 #include <sys/socket.h>
 
-
 #include "hostapd.h"
 #include "ieee802_1x.h"
 #include "accounting.h"
@@ -32,22 +35,28 @@
 #include "rc4.h"
 #include "eloop.h"
 #include "sta_info.h"
-#include "driver.h"
 #include "wpa.h"
+#include "driver.h"
+#include "eap.h"
+
+
+static void ieee802_1x_new_auth_session(struct hostapd_data *hapd,
+					struct sta_info *sta);
 
 
 static void ieee802_1x_send(hostapd *hapd, struct sta_info *sta, u8 type,
 			    u8 *data, size_t datalen)
 {
-	char *buf;
+	u8 *buf;
 	struct ieee802_1x_hdr *xhdr;
 	size_t len;
 	int encrypt = 0;
 
 	len = sizeof(*xhdr) + datalen;
-	buf = (char *) malloc(len);
+	buf = malloc(len);
 	if (buf == NULL) {
-		printf("malloc() failed for ieee802_1x_send(len=%d)\n", len);
+		printf("malloc() failed for ieee802_1x_send(len=%lu)\n",
+		       (unsigned long) len);
 		return;
 	}
 
@@ -77,10 +86,7 @@ static void ieee802_1x_send(hostapd *hapd, struct sta_info *sta, u8 type,
 	if (sta->flags & WLAN_STA_PREAUTH) {
 		rsn_preauth_send(hapd, sta, buf, len);
 	} else {
-		if (hapd->driver.send_eapol) {
-			hapd->driver.send_eapol(hapd->driver.data, sta->addr,
-						buf, len, encrypt);
-		}
+		hostapd_send_eapol(hapd, sta->addr, buf, len, encrypt);
 	}
 
 	free(buf);
@@ -103,20 +109,43 @@ void ieee802_1x_set_sta_authorized(hostapd *hapd, struct sta_info *sta,
 			       HOSTAPD_LEVEL_DEBUG, "unauthorizing port");
 	}
 
-	if (hapd->driver.set_sta_authorized)
-		hapd->driver.set_sta_authorized(hapd->driver.data, sta->addr,
-						authorized);
+	hostapd_set_sta_authorized(hapd, sta->addr, authorized);
 	if (authorized)
 		accounting_sta_start(hapd, sta);
 }
 
 
-void ieee802_1x_request_identity(hostapd *hapd, struct sta_info *sta, u8 id)
+static void ieee802_1x_eap_timeout(void *eloop_ctx, void *timeout_ctx)
 {
-	char *buf;
+	struct sta_info *sta = eloop_ctx;
+	struct eapol_state_machine *sm = sta->eapol_sm;
+	if (sm == NULL)
+		return;
+	hostapd_logger(sm->hapd, sta->addr, HOSTAPD_MODULE_IEEE8021X,
+		       HOSTAPD_LEVEL_DEBUG, "EAP timeout");
+	sm->eapTimeout = TRUE;
+	eapol_sm_step(sm);
+}
+
+
+void ieee802_1x_request_identity(struct hostapd_data *hapd,
+				 struct sta_info *sta)
+{
+	u8 *buf;
 	struct eap_hdr *eap;
 	int extra, tlen;
 	u8 *pos;
+	struct eapol_state_machine *sm = sta->eapol_sm;
+
+	if (hapd->conf->eap_authenticator) {
+		HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
+			      "IEEE 802.1X: Integrated EAP Authenticator in "
+			      "use - do not generate EAP-Request/Identity\n");
+		return;
+	}
+
+	if (sm == NULL || !sm->auth_pae.eapRestart)
+		return;
 
 	ieee802_1x_new_auth_session(hapd, sta);
 
@@ -127,7 +156,7 @@ void ieee802_1x_request_identity(hostapd *hapd, struct sta_info *sta, u8 id)
 
 	tlen = sizeof(*eap) + 1 + extra;
 
-	buf = (char *) malloc(tlen);
+	buf = malloc(tlen);
 	if (buf == NULL) {
 		printf("Could not allocate memory for identity request\n");
 		return;
@@ -137,31 +166,45 @@ void ieee802_1x_request_identity(hostapd *hapd, struct sta_info *sta, u8 id)
 
 	eap = (struct eap_hdr *) buf;
 	eap->code = EAP_CODE_REQUEST;
-	eap->identifier = id;
+	eap->identifier = ++sm->currentId;
 	eap->length = htons(tlen);
 	pos = (u8 *) (eap + 1);
 	*pos++ = EAP_TYPE_IDENTITY;
 	if (hapd->conf->eap_req_id_text)
 		memcpy(pos, hapd->conf->eap_req_id_text, extra);
 
+	sm->be_auth.eapReq = TRUE;
+	free(sm->last_eap_radius);
+	sm->last_eap_radius = buf;
+	sm->last_eap_radius_len = tlen;
+
+	eloop_cancel_timeout(ieee802_1x_eap_timeout, sta, NULL);
+	eloop_register_timeout(30, 0, ieee802_1x_eap_timeout, sta, NULL);
+	sm->eapTimeout = FALSE;
+
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-		      "IEEE 802.1X: Sending EAP Request-Identity to " MACSTR
-		      " (identifier %d)\n", MAC2STR(sta->addr),
+		      "IEEE 802.1X: Generated EAP Request-Identity for " MACSTR
+		      " (identifier %d, timeout 30)\n", MAC2STR(sta->addr),
 		      eap->identifier);
-	ieee802_1x_send(hapd, sta, IEEE802_1X_TYPE_EAP_PACKET, buf, tlen);
-	free(buf);
+
+	sm->auth_pae.eapRestart = FALSE;
 }
 
 
-void ieee802_1x_tx_canned_eap(hostapd *hapd, struct sta_info *sta, u8 id,
+void ieee802_1x_tx_canned_eap(struct hostapd_data *hapd, struct sta_info *sta,
 			      int success)
 {
 	struct eap_hdr eap;
+	struct eapol_state_machine *sm = sta->eapol_sm;
 
 	memset(&eap, 0, sizeof(eap));
 
 	eap.code = success ? EAP_CODE_SUCCESS : EAP_CODE_FAILURE;
-	eap.identifier = id;
+	eap.identifier = 1;
+	if (sm && sm->last_eap_radius) {
+		struct eap_hdr *hdr = (struct eap_hdr *) sm->last_eap_radius;
+		eap.identifier = hdr->identifier + 1;
+	}
 	eap.length = htons(sizeof(eap));
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
@@ -170,13 +213,16 @@ void ieee802_1x_tx_canned_eap(hostapd *hapd, struct sta_info *sta, u8 id,
 		      MAC2STR(sta->addr), eap.identifier);
 	ieee802_1x_send(hapd, sta, IEEE802_1X_TYPE_EAP_PACKET, (u8 *) &eap,
 			sizeof(eap));
+	if (sm)
+		sm->dot1xAuthEapolFramesTx++;
 }
 
 
-void ieee802_1x_tx_req(hostapd *hapd, struct sta_info *sta, u8 id)
+void ieee802_1x_tx_req(struct hostapd_data *hapd, struct sta_info *sta)
 {
 	struct eap_hdr *eap;
 	struct eapol_state_machine *sm = sta->eapol_sm;
+	u8 *type;
 	if (sm == NULL)
 		return;
 
@@ -188,17 +234,36 @@ void ieee802_1x_tx_req(hostapd *hapd, struct sta_info *sta, u8 id)
 	}
 
 	eap = (struct eap_hdr *) sm->last_eap_radius;
-	if (eap->identifier != id) {
-		printf("IEEE 802.1X: TxReq(%d) - changing id from %d\n",
-		       id, eap->identifier);
-		eap->identifier = id;
-	}
-
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 		      "IEEE 802.1X: Sending EAP Packet to " MACSTR
-		      " (identifier %d)\n", MAC2STR(sm->addr), id);
+		      " (identifier %d)\n", MAC2STR(sm->addr),
+		      eap->identifier);
+	sm->currentId = eap->identifier;
 	ieee802_1x_send(hapd, sta, IEEE802_1X_TYPE_EAP_PACKET,
 			sm->last_eap_radius, sm->last_eap_radius_len);
+	sm->dot1xAuthEapolFramesTx++;
+	type = (u8 *) (eap + 1);
+	if (sm->last_eap_radius_len > sizeof(*eap) &&
+	    *type == EAP_TYPE_IDENTITY)
+		sm->dot1xAuthEapolReqIdFramesTx++;
+	else
+		sm->dot1xAuthEapolReqFramesTx++;
+}
+
+
+void hostapd_get_ntp_timestamp(u8 *buf)
+{
+	struct timeval now;
+	u32 sec, usec;
+
+	/* 64-bit NTP timestamp (time from 1900-01-01 00:00:00) */
+	gettimeofday(&now, NULL);
+	sec = htonl(now.tv_sec + 2208988800U); /* Epoch to 1900 */
+	/* Estimate 2^32/10^6 = 4295 - 1/32 - 1/512 */
+	usec = now.tv_usec;
+	usec = htonl(4295 * usec - (usec >> 5) - (usec >> 9));
+	memcpy(buf, (u8 *) &sec, 4);
+	memcpy(buf + 4, (u8 *) &usec, 4);
 }
 
 
@@ -209,9 +274,7 @@ static void ieee802_1x_tx_key_one(hostapd *hapd, struct sta_info *sta,
 	u8 *buf, *ekey;
 	struct ieee802_1x_hdr *hdr;
 	struct ieee802_1x_eapol_key *key;
-	struct timeval now;
 	size_t len, ekey_len;
-	u32 ntp_hi, ntp_lo, sec, usec;
 	struct eapol_state_machine *sm = sta->eapol_sm;
 
 	if (sm == NULL)
@@ -227,19 +290,7 @@ static void ieee802_1x_tx_key_one(hostapd *hapd, struct sta_info *sta,
 	key = (struct ieee802_1x_eapol_key *) (hdr + 1);
 	key->type = EAPOL_KEY_TYPE_RC4;
 	key->key_length = htons(key_len);
-
-	/* Set the NTP timestamp as the replay counter */
-	gettimeofday(&now, NULL);
-	sec = now.tv_sec;
-	usec = now.tv_usec;
-
-#define JAN_1970 0x83aa7e80UL /* seconds from 1900 to 1970 */
-	ntp_hi = htonl(sec + JAN_1970);
-	/* approximation of 2^32/1000000 * usec */
-	ntp_lo = htonl(4295 * usec - (usec >> 5) - (usec >> 9));
-
-	memcpy(&key->replay_counter[0], &ntp_hi, sizeof(u32));
-	memcpy(&key->replay_counter[4], &ntp_lo, sizeof(u32));
+	hostapd_get_ntp_timestamp(key->replay_counter);
 
 	if (hostapd_get_rand(key->key_iv, sizeof(key->key_iv))) {
 		printf("Could not get random numbers\n");
@@ -286,11 +337,13 @@ static void ieee802_1x_tx_key_one(hostapd *hapd, struct sta_info *sta,
 		      " (%s index=%d)\n", MAC2STR(sm->addr),
 		      broadcast ? "broadcast" : "unicast", index);
 	ieee802_1x_send(hapd, sta, IEEE802_1X_TYPE_EAPOL_KEY, (u8 *) key, len);
+	if (sta->eapol_sm)
+		sta->eapol_sm->dot1xAuthEapolFramesTx++;
 	free(buf);
 }
 
 
-void ieee802_1x_tx_key(hostapd *hapd, struct sta_info *sta, u8 id)
+void ieee802_1x_tx_key(struct hostapd_data *hapd, struct sta_info *sta)
 {
 	struct eapol_state_machine *sm = sta->eapol_sm;
 
@@ -298,8 +351,8 @@ void ieee802_1x_tx_key(hostapd *hapd, struct sta_info *sta, u8 id)
 		return;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-		      "IEEE 802.1X: Sending EAPOL-Key(s) to " MACSTR
-		      " (identifier %d)\n", MAC2STR(sta->addr), id);
+		      "IEEE 802.1X: Sending EAPOL-Key(s) to " MACSTR "\n",
+		      MAC2STR(sta->addr));
 
 	if (hapd->default_wep_key)
 		ieee802_1x_tx_key_one(hapd, sta, hapd->default_wep_key_idx, 1,
@@ -328,8 +381,7 @@ void ieee802_1x_tx_key(hostapd *hapd, struct sta_info *sta, u8 id)
 
 		/* TODO: set encryption in TX callback, i.e., only after STA
 		 * has ACKed EAPOL-Key frame */
-		if (hostapd_set_encryption(hapd->driver.data, "WEP", 
-					   sta->addr, 0, ikey,
+		if (hostapd_set_encryption(hapd, "WEP", sta->addr, 0, ikey,
 					   hapd->conf->
 					   individual_wep_key_len)) {
 			printf("Could not set individual WEP encryption.\n");
@@ -340,96 +392,6 @@ void ieee802_1x_tx_key(hostapd *hapd, struct sta_info *sta, u8 id)
 }
 
 
-/* This is just a minimal authentication server for testing the authenticator.
- * It accepts anything, i.e., no keys are needed and this should not be used
- * for anything else than testing.
- * Real implementation would forward the packet to an external authentication
- * server in ieee802_1x_send_resp_to_server() */
-
-static void minimal_auth_serv_reply(u8 id, struct sta_info *sta, u8 eap_code)
-{
-	struct eap_hdr *eap;
-	struct eapol_state_machine *sm = sta->eapol_sm;
-	if (sm == NULL)
-		return;
-
-	sm->be_auth.idFromServer = id;
-
-	if (sm->last_eap_radius)
-		free(sm->last_eap_radius);
-
-	sm->last_eap_radius = malloc(sizeof(*eap));
-	if (sm->last_eap_radius == NULL) {
-		printf("minimal_auth_serv_reply: could not add EAP data\n");
-		return;
-	}
-	sm->last_eap_radius_len = sizeof(*eap);
-
-	eap = (struct eap_hdr *) sm->last_eap_radius;
-	memset(eap, 0, sizeof(*eap));
-	eap->code = eap_code;
-	eap->identifier = id;
-	eap->length = sizeof(*eap);
-}
-
-
-static void minimal_auth_serv(hostapd *hapd, struct sta_info *sta,
-			      u8 *data, size_t len)
-{
-	struct eap_hdr *eap;
-	u8 *pos;
-	size_t left;
-	int i;
-	static u8 id = 0;
-	struct eapol_state_machine *sm = sta->eapol_sm;
-
-	if (sm == NULL)
-		return;
-
-	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-		      "Minimal Authentication Server\n");
-
-	if (len < sizeof(*eap)) {
-		printf("Too short packet for minimal auth serv\n");
-		goto fail;
-	}
-
-	eap = (struct eap_hdr *) data;
-	if (eap->code != EAP_CODE_RESPONSE) {
-		printf("Invalid EAP code (%d) in minimal auth serv\n",
-		       eap->code);
-		goto fail;
-	}
-
-	pos = (u8 *) (eap + 1);
-	left = len - sizeof(*eap);
-	if (*pos != EAP_TYPE_IDENTITY) {
-		printf("Invalid EAP Response type (%d) in minimal auth serv\n",
-		       *pos);
-		goto fail;
-	}
-
-	pos++;
-	left--;
-
-	printf("Minimal Authentication Server: station " MACSTR
-	       " EAP Response-Identity: (len=%d) '", MAC2STR(sm->addr), len);
-	for (i = 0; i < left; i++)
-		print_char(pos[i]);
-	printf("'\n");
-
-	minimal_auth_serv_reply(id++, sta, EAP_CODE_SUCCESS);
-	sm->be_auth.aSuccess = TRUE;
-	eapol_sm_step(sm);
-	return;
-
- fail:
-	minimal_auth_serv_reply(id++, sta, EAP_CODE_FAILURE);
-	sm->be_auth.aFail = TRUE;
-	eapol_sm_step(sm);
-}
-
-
 static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 					  u8 *eap, size_t len)
 {
@@ -443,7 +405,7 @@ static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 		      "Encapsulating EAP message into a RADIUS packet\n");
 
-	sm->radius_identifier = radius_client_get_id(hapd);
+	sm->radius_identifier = radius_client_get_id(hapd->radius);
 	msg = radius_msg_new(RADIUS_CODE_ACCESS_REQUEST,
 			     sm->radius_identifier);
 	if (msg == NULL) {
@@ -468,7 +430,7 @@ static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 
 	if (hapd->conf->nas_identifier &&
 	    !radius_msg_add_attr(msg, RADIUS_ATTR_NAS_IDENTIFIER,
-				 hapd->conf->nas_identifier,
+				 (u8 *) hapd->conf->nas_identifier,
 				 strlen(hapd->conf->nas_identifier))) {
 		printf("Could not add NAS-Identifier\n");
 		goto fail;
@@ -482,7 +444,7 @@ static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 	snprintf(buf, sizeof(buf), RADIUS_802_1X_ADDR_FORMAT ":%s",
 		 MAC2STR(hapd->own_addr), hapd->conf->ssid);
 	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CALLED_STATION_ID,
-				 buf, strlen(buf))) {
+				 (u8 *) buf, strlen(buf))) {
 		printf("Could not add Called-Station-Id\n");
 		goto fail;
 	}
@@ -490,7 +452,7 @@ static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 	snprintf(buf, sizeof(buf), RADIUS_802_1X_ADDR_FORMAT,
 		 MAC2STR(sta->addr));
 	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CALLING_STATION_ID,
-				 buf, strlen(buf))) {
+				 (u8 *) buf, strlen(buf))) {
 		printf("Could not add Calling-Station-Id\n");
 		goto fail;
 	}
@@ -511,7 +473,7 @@ static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 
 	snprintf(buf, sizeof(buf), "CONNECT 11Mbps 802.11b");
 	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CONNECT_INFO,
-				 buf, strlen(buf))) {
+				 (u8 *) buf, strlen(buf))) {
 		printf("Could not add Connect-Info\n");
 		goto fail;
 	}
@@ -538,7 +500,7 @@ static void ieee802_1x_encapsulate_radius(hostapd *hapd, struct sta_info *sta,
 		}
 	}
 
-	radius_client_send(hapd, msg, RADIUS_AUTH, sta->addr);
+	radius_client_send(hapd->radius, msg, RADIUS_AUTH, sta->addr);
 	return;
 
  fail:
@@ -553,9 +515,9 @@ static char *eap_type_text(u8 type)
 	case EAP_TYPE_IDENTITY: return "Identity";
 	case EAP_TYPE_NOTIFICATION: return "Notification";
 	case EAP_TYPE_NAK: return "Nak";
-	case EAP_TYPE_MD5_CHALLENGE: return "MD5-Challenge";
-	case EAP_TYPE_ONE_TIME_PASSWORD: return "One-Time Password";
-	case EAP_TYPE_GENERIC_TOKEN_CARD: return "Generic Token Card";
+	case EAP_TYPE_MD5: return "MD5-Challenge";
+	case EAP_TYPE_OTP: return "One-Time Password";
+	case EAP_TYPE_GTC: return "Generic Token Card";
 	case EAP_TYPE_TLS: return "TLS";
 	case EAP_TYPE_TTLS: return "TTLS";
 	case EAP_TYPE_PEAP: return "PEAP";
@@ -573,20 +535,22 @@ static void handle_eap_response(hostapd *hapd, struct sta_info *sta,
 		return;
 
 	if (eap->identifier != sm->currentId) {
-		printf("EAP Identifier of the Response-Identity from " MACSTR
-		       " does not match (was %d, expected %d)\n",
-		       MAC2STR(sm->addr), eap->identifier,
-		       sm->currentId);
+		hostapd_logger(hapd, sm->addr, HOSTAPD_MODULE_IEEE8021X,
+			       HOSTAPD_LEVEL_DEBUG,
+			       "EAP Identifier of the Response-Identity does "
+			       "not match (was %d, expected %d) - ignored",
+			       eap->identifier, sm->currentId);
 		return;
 	}
 
 	if (len < 1) {
-		printf("   too short response data\n");
+		printf("handle_eap_response: too short response data\n");
 		return;
 	}
 
-	if (sm->last_eap_supp != NULL)
-		free(sm->last_eap_supp);
+	eloop_cancel_timeout(ieee802_1x_eap_timeout, sta, NULL);
+
+	free(sm->last_eap_supp);
 	sm->last_eap_supp_len = sizeof(*eap) + len;
 	sm->last_eap_supp = (u8 *) malloc(sm->last_eap_supp_len);
 	if (sm->last_eap_supp == NULL) {
@@ -606,9 +570,6 @@ static void handle_eap_response(hostapd *hapd, struct sta_info *sta,
 		       eap->code, eap->identifier, ntohs(eap->length),
 		       eap_type_text(type), type);
 
-	/* TODO: IEEE 802.1aa/D4: should use auth_pae.initialEAPMsg to check
-	 * which EAP packet is accepted as response; currently, hostapd only
-	 * supports EAP Response-Identity, so this can be hardcoded */
 	if (type == EAP_TYPE_IDENTITY) {
 		char *buf, *pos;
 		int i;
@@ -631,21 +592,21 @@ static void handle_eap_response(hostapd *hapd, struct sta_info *sta,
 			free(buf);
 		}
 
-		sm->auth_pae.rxInitialRsp = TRUE;
+		sm->rx_identity = TRUE;
+		sm->dot1xAuthEapolRespIdFramesRx++;
 
 		/* Save station identity for future RADIUS packets */
-		if (sm->identity)
-			free(sm->identity);
-		sm->identity = (u8 *) malloc(len);
+		free(sm->identity);
+		sm->identity = malloc(len + 1);
 		if (sm->identity) {
 			memcpy(sm->identity, data, len);
+			sm->identity[len] = '\0';
 			sm->identity_len = len;
 		}
-	} else {
-		if (type != EAP_TYPE_NAK)
-			sm->be_auth.backendNonNakResponsesFromSupplicant++;
-		sm->be_auth.rxResp = TRUE;
-	}
+	} else
+		sm->dot1xAuthEapolRespFramesRx++;
+
+	sm->eapolEap = TRUE;
 }
 
 
@@ -674,8 +635,8 @@ static void handle_eap(hostapd *hapd, struct sta_info *sta, u8 *buf,
 		printf("   Too short frame to contain this EAP packet\n");
 		return;
 	} else if (eap_len < len) {
-		printf("   Ignoring %d extra bytes after EAP packet\n",
-		       len - eap_len);
+		printf("   Ignoring %lu extra bytes after EAP packet\n",
+		       (unsigned long) len - eap_len);
 	}
 
 	eap_len -= sizeof(*eap);
@@ -713,8 +674,8 @@ void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
 		return;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-		      "IEEE 802.1X: %d bytes from " MACSTR "\n", len,
-		      MAC2STR(sa));
+		      "IEEE 802.1X: %lu bytes from " MACSTR "\n",
+		      (unsigned long) len, MAC2STR(sa));
 	sta = ap_get_sta(hapd, sa);
 	if (!sta) {
 		printf("   no station information available\n");
@@ -734,12 +695,20 @@ void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
 
 	if (len - sizeof(*hdr) < datalen) {
 		printf("   frame too short for this IEEE 802.1X packet\n");
+		if (sta->eapol_sm)
+			sta->eapol_sm->dot1xAuthEapLengthErrorFramesRx++;
 		return;
 	}
 	if (len - sizeof(*hdr) > datalen) {
 		HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-			      "   ignoring %d extra octets after IEEE 802.1X "
-			      "packet\n", len - sizeof(*hdr) - datalen);
+			      "   ignoring %lu extra octets after IEEE 802.1X "
+			      "packet\n",
+			      (unsigned long) len - sizeof(*hdr) - datalen);
+	}
+
+	if (sta->eapol_sm) {
+		sta->eapol_sm->dot1xAuthLastEapolFrameVersion = hdr->version;
+		sta->eapol_sm->dot1xAuthEapolFramesRx++;
 	}
 
 	key = (struct ieee802_1x_eapol_key *) (hdr + 1);
@@ -762,6 +731,10 @@ void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
 
 	/* since we support version 1, we can ignore version field and proceed
 	 * as specified in version 1 standard [IEEE Std 802.1X-2001, 7.5.5] */
+	/* TODO: actually, we are not version 1 anymore.. However, Version 2
+	 * does not change frame contents, so should be ok to process frames
+	 * more or less identically. Some changes might be needed for
+	 * verification of fields. */
 
 	switch (hdr->type) {
 	case IEEE802_1X_TYPE_EAP_PACKET:
@@ -779,7 +752,9 @@ void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
 				       "STA sent EAPOL-Start");
 			sta->pmksa = NULL;
 		}
-		sta->eapol_sm->auth_pae.eapStart = TRUE;
+		sta->eapol_sm->auth_pae.eapolStart = TRUE;
+		sta->eapol_sm->dot1xAuthEapolStartFramesRx++;
+		wpa_sm_event(hapd, sta, WPA_REAUTH_EAPOL);
 		eapol_sm_step(sta->eapol_sm);
 		break;
 
@@ -789,7 +764,8 @@ void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
 			       "from STA");
 		sta->acct_terminate_cause =
 			RADIUS_ACCT_TERMINATE_CAUSE_USER_REQUEST;
-		sta->eapol_sm->auth_pae.eapLogoff = TRUE;
+		sta->eapol_sm->auth_pae.eapolLogoff = TRUE;
+		sta->eapol_sm->dot1xAuthEapolLogoffFramesRx++;
 		eapol_sm_step(sta->eapol_sm);
 		break;
 
@@ -811,6 +787,7 @@ void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len)
 	default:
 		HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 			      "   unknown IEEE 802.1X packet type\n");
+		sta->eapol_sm->dot1xAuthInvalidEapolFramesRx++;
 		break;
 	}
 
@@ -841,6 +818,8 @@ void ieee802_1x_free_station(struct sta_info *sta)
 {
 	struct eapol_state_machine *sm = sta->eapol_sm;
 
+	eloop_cancel_timeout(ieee802_1x_eap_timeout, sta, NULL);
+
 	if (sm == NULL)
 		return;
 
@@ -863,7 +842,7 @@ void ieee802_1x_free_station(struct sta_info *sta)
 
 static void ieee802_1x_decapsulate_radius(hostapd *hapd, struct sta_info *sta)
 {
-	char *eap;
+	u8 *eap;
 	size_t len;
 	struct eap_hdr *hdr;
 	int eap_type = -1;
@@ -871,8 +850,11 @@ static void ieee802_1x_decapsulate_radius(hostapd *hapd, struct sta_info *sta)
 	struct radius_msg *msg;
 	struct eapol_state_machine *sm = sta->eapol_sm;
 
-	if (sm == NULL || sm->last_recv_radius == NULL)
+	if (sm == NULL || sm->last_recv_radius == NULL) {
+		if (sm)
+			sm->be_auth.eapNoReq = TRUE;
 		return;
+	}
 
 	msg = sm->last_recv_radius;
 
@@ -887,6 +869,7 @@ static void ieee802_1x_decapsulate_radius(hostapd *hapd, struct sta_info *sta)
 		free(sm->last_eap_radius);
 		sm->last_eap_radius = NULL;
 		sm->last_eap_radius_len = 0;
+		sm->be_auth.eapNoReq = TRUE;
 		return;
 	}
 
@@ -895,6 +878,7 @@ static void ieee802_1x_decapsulate_radius(hostapd *hapd, struct sta_info *sta)
 			       HOSTAPD_LEVEL_WARNING, "too short EAP packet "
 			       "received from authentication server");
 		free(eap);
+		sm->be_auth.eapNoReq = TRUE;
 		return;
 	}
 
@@ -927,11 +911,9 @@ static void ieee802_1x_decapsulate_radius(hostapd *hapd, struct sta_info *sta)
 		       HOSTAPD_LEVEL_DEBUG, "decapsulated EAP packet (code=%d "
 		       "id=%d len=%d) from RADIUS server: %s",
 		      hdr->code, hdr->identifier, ntohs(hdr->length), buf);
+	sm->be_auth.eapReq = TRUE;
 
-	sm->be_auth.idFromServer = hdr->identifier;
-
-	if (sm->last_eap_radius)
-		free(sm->last_eap_radius);
+	free(sm->last_eap_radius);
 	sm->last_eap_radius = eap;
 	sm->last_eap_radius_len = len;
 }
@@ -952,14 +934,16 @@ static void ieee802_1x_get_keys(hostapd *hapd, struct sta_info *sta,
 	if (keys) {
 		if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL) && keys->send) {
 			size_t i;
-			printf("MS-MPPE-Send-Key (len=%d):", keys->send_len);
+			printf("MS-MPPE-Send-Key (len=%lu):",
+			       (unsigned long) keys->send_len);
 			for (i = 0; i < keys->send_len; i++)
 				printf(" %02x", keys->send[i]);
 			printf("\n");
 		}
 		if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL) && keys->recv) {
 			size_t i;
-			printf("MS-MPPE-Recv-Key (len=%d):", keys->recv_len);
+			printf("MS-MPPE-Recv-Key (len=%lu):",
+			       (unsigned long) keys->recv_len);
 			for (i = 0; i < keys->recv_len; i++)
 				printf(" %02x", keys->recv[i]);
 			printf("\n");
@@ -1055,16 +1039,17 @@ ieee802_1x_search_radius_identifier(struct hostapd_data *hapd, u8 identifier)
 
 /* Process the RADIUS frames from Authentication Server */
 static RadiusRxResult
-ieee802_1x_receive_auth(hostapd *hapd,
-			struct radius_msg *msg, struct radius_msg *req,
+ieee802_1x_receive_auth(struct radius_msg *msg, struct radius_msg *req,
 			u8 *shared_secret, size_t shared_secret_len,
 			void *data)
 {
+	struct hostapd_data *hapd = data;
 	struct sta_info *sta;
-	u32 session_timeout, termination_action;
+	u32 session_timeout, termination_action, acct_interim_interval;
 	int session_timeout_set;
-	int acct_interim_interval;
+	int eap_timeout;
 	struct eapol_state_machine *sm;
+	int override_eapReq = 0;
 
 	sm = ieee802_1x_search_radius_identifier(hapd, msg->hdr->identifier);
 	if (sm == NULL) {
@@ -1088,7 +1073,7 @@ ieee802_1x_receive_auth(hostapd *hapd,
 				     req)) {
 		printf("Incoming RADIUS packet did not have correct "
 		       "Message-Authenticator - dropped\n");
-		return RADIUS_RX_UNKNOWN;
+		return RADIUS_RX_INVALID_AUTHENTICATOR;
 	}
 
 	if (msg->hdr->code != RADIUS_CODE_ACCESS_ACCEPT &&
@@ -1135,15 +1120,15 @@ ieee802_1x_receive_auth(hostapd *hapd,
 
 	switch (msg->hdr->code) {
 	case RADIUS_CODE_ACCESS_ACCEPT:
-		/* draft-congdon-radius-8021x-22.txt, Ch. 3.17 */
-		if (session_timeout_set &&
-			termination_action ==
+		/* RFC 3580, Ch. 3.17 */
+		if (session_timeout_set && termination_action ==
 		    RADIUS_TERMINATION_ACTION_RADIUS_REQUEST) {
 			sm->reauth_timer.reAuthPeriod = session_timeout;
 		} else if (session_timeout_set)
 			ap_sta_session_timeout(hapd, sta, session_timeout);
 
-		sm->be_auth.aSuccess = TRUE;
+		sm->eapSuccess = TRUE;
+		override_eapReq = 1;
 		ieee802_1x_get_keys(hapd, sta, msg, req, shared_secret,
 				    shared_secret_len);
 		if (sm->keyAvailable) {
@@ -1153,20 +1138,31 @@ ieee802_1x_receive_auth(hostapd *hapd,
 		}
 		break;
 	case RADIUS_CODE_ACCESS_REJECT:
-		sm->be_auth.aFail = TRUE;
+		sm->eapFail = TRUE;
+		override_eapReq = 1;
 		break;
 	case RADIUS_CODE_ACCESS_CHALLENGE:
 		if (session_timeout_set) {
-			/* RFC 2869, Ch. 2.3.2
-			 * draft-congdon-radius-8021x-22.txt, Ch. 3.17 */
-			sm->be_auth.suppTimeout = session_timeout;
-		}
-		sm->be_auth.aReq = TRUE;
+			/* RFC 2869, Ch. 2.3.2; RFC 3580, Ch. 3.17 */
+			eap_timeout = session_timeout;
+		} else
+			eap_timeout = 30;
+		hostapd_logger(hapd, sm->addr, HOSTAPD_MODULE_IEEE8021X,
+			       HOSTAPD_LEVEL_DEBUG,
+			       "using EAP timeout of %d seconds%s",
+			       eap_timeout,
+			       session_timeout_set ? " (from RADIUS)" : "");
+		eloop_cancel_timeout(ieee802_1x_eap_timeout, sta, NULL);
+		eloop_register_timeout(eap_timeout, 0, ieee802_1x_eap_timeout,
+				       sta, NULL);
+		sm->eapTimeout = FALSE;
 		break;
 	}
 
 	ieee802_1x_store_radius_class(hapd, sta, msg);
 	ieee802_1x_decapsulate_radius(hapd, sta);
+	if (override_eapReq)
+		sm->be_auth.eapReq = FALSE;
 
 	eapol_sm_step(sm);
 
@@ -1182,12 +1178,39 @@ void ieee802_1x_send_resp_to_server(hostapd *hapd, struct sta_info *sta)
 	if (sm == NULL)
 		return;
 
-	if (hapd->conf->minimal_eap)
-		minimal_auth_serv(hapd, sta, sm->last_eap_supp,
-				  sm->last_eap_supp_len);
-	else
+	if (hapd->conf->eap_authenticator) {
+		eap_set_eapRespData(sm->eap, sm->last_eap_supp,
+				    sm->last_eap_supp_len);
+	} else {
 		ieee802_1x_encapsulate_radius(hapd, sta, sm->last_eap_supp,
 					      sm->last_eap_supp_len);
+	}
+}
+
+
+void ieee802_1x_abort_auth(struct hostapd_data *hapd, struct sta_info *sta)
+{
+	struct eapol_state_machine *sm = sta->eapol_sm;
+	if (sm == NULL)
+		return;
+
+	hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_IEEE8021X,
+		       HOSTAPD_LEVEL_DEBUG, "aborting authentication");
+
+	if (sm->last_recv_radius) {
+		radius_msg_free(sm->last_recv_radius);
+		free(sm->last_recv_radius);
+		sm->last_recv_radius = NULL;
+	}
+	free(sm->last_eap_supp);
+	sm->last_eap_supp = NULL;
+	sm->last_eap_supp_len = 0;
+	free(sm->last_eap_radius);
+	sm->last_eap_radius = NULL;
+	sm->last_eap_radius_len = 0;
+	free(sm->radius_class);
+	sm->radius_class = NULL;
+	sm->radius_class_len = 0;
 }
 
 
@@ -1232,63 +1255,141 @@ void ieee802_1x_dump_state(FILE *f, const char *prefix, struct sta_info *sta)
 #endif /* HOSTAPD_DUMP_STATE */
 
 
-static int ieee802_1x_init_kernel_1x(hostapd *hapd)
+static int ieee802_1x_rekey_broadcast(hostapd *hapd)
 {
-	/* enable kernel driver support for IEEE 802.1X */
-	if (hostap_ioctl_prism2param(hapd->driver.data,
-				     PRISM2_PARAM_IEEE_802_1X, 1)) {
-		printf("Could not setup IEEE 802.1X support in kernel driver."
-		       "\n");
+	if (hapd->conf->default_wep_key_len < 1)
+		return 0;
+
+	free(hapd->default_wep_key);
+	hapd->default_wep_key = malloc(hapd->conf->default_wep_key_len);
+	if (hapd->default_wep_key == NULL ||
+	    hostapd_get_rand(hapd->default_wep_key,
+			     hapd->conf->default_wep_key_len)) {
+		printf("Could not generate random WEP key.\n");
+		free(hapd->default_wep_key);
+		hapd->default_wep_key = NULL;
 		return -1;
 	}
 
-	/* use host driver implementation of encryption to allow
-	 * individual keys and passing plaintext EAPOL frames */
-	if (hostap_ioctl_prism2param(hapd->driver.data,
-				     PRISM2_PARAM_HOST_DECRYPT, 1) ||
-		hostap_ioctl_prism2param(hapd->driver.data,
-					 PRISM2_PARAM_HOST_ENCRYPT, 1)) {
-		printf("Could not setup host-based encryption in kernel "
-		       "driver.\n");
-		return -1;
+	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MINIMAL)) {
+		hostapd_hexdump("IEEE 802.1X: New default WEP key",
+				hapd->default_wep_key,
+				hapd->conf->default_wep_key_len);
 	}
 
 	return 0;
 }
 
 
+static int ieee802_1x_sta_key_available(struct hostapd_data *hapd,
+					struct sta_info *sta, void *ctx)
+{
+	if (sta->eapol_sm) {
+		sta->eapol_sm->keyAvailable = TRUE;
+		eapol_sm_step(sta->eapol_sm);
+	}
+	return 0;
+}
+
+
+static void ieee802_1x_rekey(void *eloop_ctx, void *timeout_ctx)
+{
+	struct hostapd_data *hapd = eloop_ctx;
+
+	if (hapd->default_wep_key_idx >= 3)
+		hapd->default_wep_key_idx =
+			hapd->conf->individual_wep_key_len > 0 ? 1 : 0;
+	else
+		hapd->default_wep_key_idx++;
+
+	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "IEEE 802.1X: New default WEP "
+		      "key index %d\n", hapd->default_wep_key_idx);
+		      
+	if (ieee802_1x_rekey_broadcast(hapd)) {
+		hostapd_logger(hapd, NULL, HOSTAPD_MODULE_IEEE8021X,
+			       HOSTAPD_LEVEL_WARNING, "failed to generate a "
+			       "new broadcast key");
+		free(hapd->default_wep_key);
+		hapd->default_wep_key = NULL;
+		return;
+	}
+
+	/* TODO: Could setup key for RX here, but change default TX keyid only
+	 * after new broadcast key has been sent to all stations. */
+	if (hostapd_set_encryption(hapd, "WEP", NULL,
+				   hapd->default_wep_key_idx,
+				   hapd->default_wep_key,
+				   hapd->conf->default_wep_key_len)) {
+		hostapd_logger(hapd, NULL, HOSTAPD_MODULE_IEEE8021X,
+			       HOSTAPD_LEVEL_WARNING, "failed to configure a "
+			       "new broadcast key");
+		free(hapd->default_wep_key);
+		hapd->default_wep_key = NULL;
+		return;
+	}
+
+	ap_for_each_sta(hapd, ieee802_1x_sta_key_available, NULL);
+
+	if (hapd->conf->wep_rekeying_period > 0) {
+		eloop_register_timeout(hapd->conf->wep_rekeying_period, 0,
+				       ieee802_1x_rekey, hapd, NULL);
+	}
+}
+
+
 int ieee802_1x_init(hostapd *hapd)
 {
+	int i;
+
 	if ((hapd->conf->ieee802_1x || hapd->conf->wpa) &&
-	    ieee802_1x_init_kernel_1x(hapd))
+	    hostapd_set_ieee8021x(hapd, 1))
 		return -1;
 
-	if (radius_client_register(hapd, RADIUS_AUTH, ieee802_1x_receive_auth,
-				   NULL))
+	if (radius_client_register(hapd->radius, RADIUS_AUTH,
+				   ieee802_1x_receive_auth, hapd))
 		return -1;
 
+	if (hapd->conf->default_wep_key_len) {
+		for (i = 0; i < 4; i++)
+			hostapd_set_encryption(hapd, "none", NULL, i, NULL, 0);
+
+		ieee802_1x_rekey(hapd, NULL);
+
+		if (hapd->default_wep_key == NULL)
+			return -1;
+	}
+
 	return 0;
 }
 
 
 void ieee802_1x_deinit(hostapd *hapd)
 {
+	if (hapd->driver != NULL &&
+	    (hapd->conf->ieee802_1x || hapd->conf->wpa))
+		hostapd_set_ieee8021x(hapd, 0);
 }
 
 
-void ieee802_1x_new_auth_session(hostapd *hapd, struct sta_info *sta)
+static void ieee802_1x_new_auth_session(struct hostapd_data *hapd,
+					struct sta_info *sta)
 {
 	struct eapol_state_machine *sm = sta->eapol_sm;
-	if (sm == NULL || sm->last_recv_radius == NULL)
+	if (sm == NULL)
 		return;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 		      "IEEE 802.1X: station " MACSTR " - new auth session, "
 		      "clearing State\n", MAC2STR(sta->addr));
 
-	radius_msg_free(sm->last_recv_radius);
-	free(sm->last_recv_radius);
-	sm->last_recv_radius = NULL;
+	if (sm->last_recv_radius) {
+		radius_msg_free(sm->last_recv_radius);
+		free(sm->last_recv_radius);
+		sm->last_recv_radius = NULL;
+	}
+
+	sm->eapSuccess = FALSE;
+	sm->eapFail = FALSE;
 }
 
 
@@ -1398,14 +1499,6 @@ void ieee802_1x_notify_port_valid(struct eapol_state_machine *sm,
 }
 
 
-void ieee802_1x_notify_key_available(struct eapol_state_machine *sm,
-				     int available)
-{
-	if (sm == NULL)
-		return;
-	sm->keyAvailable = available ? TRUE : FALSE;
-}
-
 void ieee802_1x_notify_pre_auth(struct eapol_state_machine *sm, int pre_auth)
 {
 	if (sm == NULL)
@@ -1415,3 +1508,142 @@ void ieee802_1x_notify_pre_auth(struct eapol_state_machine *sm, int pre_auth)
 	else
 		sm->flags &= ~EAPOL_SM_PREAUTH;
 }
+
+
+static const char * bool_txt(Boolean bool)
+{
+	return bool ? "TRUE" : "FALSE";
+}
+
+
+int ieee802_1x_get_mib(struct hostapd_data *hapd, char *buf, size_t buflen)
+{
+	/* TODO */
+	return 0;
+}
+
+
+int ieee802_1x_get_mib_sta(struct hostapd_data *hapd, struct sta_info *sta,
+			   char *buf, size_t buflen)
+{
+	int len = 0;
+	struct eapol_state_machine *sm = sta->eapol_sm;
+
+	if (sm == NULL)
+		return 0;
+
+	len += snprintf(buf + len, buflen - len,
+			"dot1xPaePortNumber=%d\n"
+			"dot1xPaePortProtocolVersion=%d\n"
+			"dot1xPaePortCapabilities=1\n"
+			"dot1xPaePortInitialize=%d\n"
+			"dot1xPaePortReauthenticate=FALSE\n",
+			sta->aid,
+			EAPOL_VERSION,
+			sm->initialize);
+
+	/* dot1xAuthConfigTable */
+	len += snprintf(buf + len, buflen - len,
+			"dot1xAuthPaeState=%d\n"
+			"dot1xAuthBackendAuthState=%d\n"
+			"dot1xAuthAdminControlledDirections=%d\n"
+			"dot1xAuthOperControlledDirections=%d\n"
+			"dot1xAuthAuthControlledPortStatus=%d\n"
+			"dot1xAuthAuthControlledPortControl=%d\n"
+			"dot1xAuthQuietPeriod=%u\n"
+			"dot1xAuthServerTimeout=%u\n"
+			"dot1xAuthReAuthPeriod=%u\n"
+			"dot1xAuthReAuthEnabled=%s\n"
+			"dot1xAuthKeyTxEnabled=%s\n",
+			sm->auth_pae.state + 1,
+			sm->be_auth.state + 1,
+			sm->ctrl_dir.adminControlledDirections,
+			sm->ctrl_dir.operControlledDirections,
+			sm->authPortStatus,
+			sm->portControl,
+			sm->auth_pae.quietPeriod,
+			sm->be_auth.serverTimeout,
+			sm->reauth_timer.reAuthPeriod,
+			bool_txt(sm->reauth_timer.reAuthEnabled),
+			bool_txt(sm->keyTxEnabled));
+
+	/* dot1xAuthStatsTable */
+	len += snprintf(buf + len, buflen - len,
+			"dot1xAuthEapolFramesRx=%u\n"
+			"dot1xAuthEapolFramesTx=%u\n"
+			"dot1xAuthEapolStartFramesRx=%u\n"
+			"dot1xAuthEapolLogoffFramesRx=%u\n"
+			"dot1xAuthEapolRespIdFramesRx=%u\n"
+			"dot1xAuthEapolRespFramesRx=%u\n"
+			"dot1xAuthEapolReqIdFramesTx=%u\n"
+			"dot1xAuthEapolReqFramesTx=%u\n"
+			"dot1xAuthInvalidEapolFramesRx=%u\n"
+			"dot1xAuthEapLengthErrorFramesRx=%u\n"
+			"dot1xAuthLastEapolFrameVersion=%u\n"
+			"dot1xAuthLastEapolFrameSource=" MACSTR "\n",
+			sm->dot1xAuthEapolFramesRx,
+			sm->dot1xAuthEapolFramesTx,
+			sm->dot1xAuthEapolStartFramesRx,
+			sm->dot1xAuthEapolLogoffFramesRx,
+			sm->dot1xAuthEapolRespIdFramesRx,
+			sm->dot1xAuthEapolRespFramesRx,
+			sm->dot1xAuthEapolReqIdFramesTx,
+			sm->dot1xAuthEapolReqFramesTx,
+			sm->dot1xAuthInvalidEapolFramesRx,
+			sm->dot1xAuthEapLengthErrorFramesRx,
+			sm->dot1xAuthLastEapolFrameVersion,
+			MAC2STR(sm->addr));
+
+	/* dot1xAuthDiagTable */
+	len += snprintf(buf + len, buflen - len,
+			"dot1xAuthEntersConnecting=%u\n"
+			"dot1xAuthEapLogoffsWhileConnecting=%u\n"
+			"dot1xAuthEntersAuthenticating=%u\n"
+			"dot1xAuthAuthSuccessesWhileAuthenticating=%u\n"
+			"dot1xAuthAuthTimeoutsWhileAuthenticating=%u\n"
+			"dot1xAuthAuthFailWhileAuthenticating=%u\n"
+			"dot1xAuthAuthEapStartsWhileAuthenticating=%u\n"
+			"dot1xAuthAuthEapLogoffWhileAuthenticating=%u\n"
+			"dot1xAuthAuthReauthsWhileAuthenticated=%u\n"
+			"dot1xAuthAuthEapStartsWhileAuthenticated=%u\n"
+			"dot1xAuthAuthEapLogoffWhileAuthenticated=%u\n"
+			"dot1xAuthBackendResponses=%u\n"
+			"dot1xAuthBackendAccessChallenges=%u\n"
+			"dot1xAuthBackendOtherRequestsToSupplicant=%u\n"
+			"dot1xAuthBackendAuthSuccesses=%u\n"
+			"dot1xAuthBackendAuthFails=%u\n",
+			sm->auth_pae.authEntersConnecting,
+			sm->auth_pae.authEapLogoffsWhileConnecting,
+			sm->auth_pae.authEntersAuthenticating,
+			sm->auth_pae.authAuthSuccessesWhileAuthenticating,
+			sm->auth_pae.authAuthTimeoutsWhileAuthenticating,
+			sm->auth_pae.authAuthFailWhileAuthenticating,
+			sm->auth_pae.authAuthEapStartsWhileAuthenticating,
+			sm->auth_pae.authAuthEapLogoffWhileAuthenticating,
+			sm->auth_pae.authAuthReauthsWhileAuthenticated,
+			sm->auth_pae.authAuthEapStartsWhileAuthenticated,
+			sm->auth_pae.authAuthEapLogoffWhileAuthenticated,
+			sm->be_auth.backendResponses,
+			sm->be_auth.backendAccessChallenges,
+			sm->be_auth.backendOtherRequestsToSupplicant,
+			sm->be_auth.backendAuthSuccesses,
+			sm->be_auth.backendAuthFails);
+
+	/* dot1xAuthSessionStatsTable */
+	len += snprintf(buf + len, buflen - len,
+			/* TODO: dot1xAuthSessionOctetsRx */
+			/* TODO: dot1xAuthSessionOctetsTx */
+			/* TODO: dot1xAuthSessionFramesRx */
+			/* TODO: dot1xAuthSessionFramesTx */
+			"dot1xAuthSessionId=%08X-%08X\n"
+			"dot1xAuthSessionAuthenticMethod=%d\n"
+			"dot1xAuthSessionTime=%u\n"
+			"dot1xAuthSessionTerminateCause=999\n"
+			"dot1xAuthSessionUserName=%s\n",
+			sta->acct_session_id_hi, sta->acct_session_id_lo,
+			sta->wpa_key_mgmt == WPA_KEY_MGMT_IEEE8021X ? 1 : 2,
+			(unsigned int) (time(NULL) - sta->acct_session_start),
+			sm->identity);
+
+	return len;
+}
