@@ -1,7 +1,7 @@
 #ifndef IEEE802_1X_H
 #define IEEE802_1X_H
 
-/* IEEE Std 802.1X-2001, 7.2 */
+/* IEEE Std 802.1X-REV-d11, 7.2 */
 
 struct ieee802_1x_hdr {
 	u8 version;
@@ -10,7 +10,7 @@ struct ieee802_1x_hdr {
 	/* followed by length octets of data */
 } __attribute__ ((packed));
 
-#define EAPOL_VERSION 1
+#define EAPOL_VERSION 2
 
 enum { IEEE802_1X_TYPE_EAP_PACKET = 0,
        IEEE802_1X_TYPE_EAPOL_START = 1,
@@ -47,44 +47,18 @@ enum { EAPOL_KEY_TYPE_RC4 = 1, EAPOL_KEY_TYPE_RSN = 2,
        EAPOL_KEY_TYPE_WPA = 254 };
 
 
-/* RFC 2284 - PPP Extensible Authentication Protocol (EAP) */
-
-struct eap_hdr {
-	u8 code;
-	u8 identifier;
-	u16 length; /* including code and identifier */
-	/* followed by length-2 octets of data */
-} __attribute__ ((packed));
-
-enum { EAP_CODE_REQUEST = 1, EAP_CODE_RESPONSE = 2, EAP_CODE_SUCCESS = 3,
-       EAP_CODE_FAILURE = 4 };
-
-/* EAP Request and Response data begins with one octet Type. Success and
- * Failure do not have additional data. */
-
-/* RFC 2284, 3.0 */
-enum { EAP_TYPE_IDENTITY = 1,
-       EAP_TYPE_NOTIFICATION = 2,
-       EAP_TYPE_NAK = 3 /* Response only */,
-       EAP_TYPE_MD5_CHALLENGE = 4,
-       EAP_TYPE_ONE_TIME_PASSWORD = 5 /* RFC 1938 */,
-       EAP_TYPE_GENERIC_TOKEN_CARD = 6,
-       EAP_TYPE_TLS = 13 /* RFC 2716 */,
-       EAP_TYPE_TTLS = 21 /* draft-ietf-pppext-eap-ttls-02.txt */,
-       EAP_TYPE_PEAP = 25 /* draft-josefsson-pppext-eap-tls-eap-06.txt */,
-};
-
-
 void ieee802_1x_receive(hostapd *hapd, u8 *sa, u8 *buf, size_t len);
 void ieee802_1x_new_station(hostapd *hapd, struct sta_info *sta);
 void ieee802_1x_free_station(struct sta_info *sta);
 
-void ieee802_1x_request_identity(hostapd *hapd, struct sta_info *sta, u8 id);
-void ieee802_1x_tx_canned_eap(hostapd *hapd, struct sta_info *sta, u8 id,
+void ieee802_1x_request_identity(struct hostapd_data *hapd,
+				 struct sta_info *sta);
+void ieee802_1x_tx_canned_eap(struct hostapd_data *hapd, struct sta_info *sta,
 			      int success);
-void ieee802_1x_tx_req(hostapd *hapd, struct sta_info *sta, u8 id);
-void ieee802_1x_tx_key(hostapd *hapd, struct sta_info *sta, u8 id);
+void ieee802_1x_tx_req(hostapd *hapd, struct sta_info *sta);
+void ieee802_1x_tx_key(struct hostapd_data *hapd, struct sta_info *sta);
 void ieee802_1x_send_resp_to_server(hostapd *hapd, struct sta_info *sta);
+void ieee802_1x_abort_auth(struct hostapd_data *hapd, struct sta_info *sta);
 void ieee802_1x_set_sta_authorized(hostapd *hapd, struct sta_info *sta,
 				   int authorized);
 void ieee802_1x_set_port_enabled(hostapd *hapd, struct sta_info *sta,
@@ -92,7 +66,6 @@ void ieee802_1x_set_port_enabled(hostapd *hapd, struct sta_info *sta,
 void ieee802_1x_dump_state(FILE *f, const char *prefix, struct sta_info *sta);
 int ieee802_1x_init(hostapd *hapd);
 void ieee802_1x_deinit(hostapd *hapd);
-void ieee802_1x_new_auth_session(hostapd *hapd, struct sta_info *sta);
 int ieee802_1x_tx_status(hostapd *hapd, struct sta_info *sta, u8 *buf,
 			 size_t len, int ack);
 u8 * ieee802_1x_get_identity(struct eapol_state_machine *sm, size_t *len);
@@ -102,8 +75,10 @@ void ieee802_1x_notify_port_enabled(struct eapol_state_machine *sm,
 				    int enabled);
 void ieee802_1x_notify_port_valid(struct eapol_state_machine *sm,
 				  int valid);
-void ieee802_1x_notify_key_available(struct eapol_state_machine *sm,
-				     int available);
 void ieee802_1x_notify_pre_auth(struct eapol_state_machine *sm, int pre_auth);
+int ieee802_1x_get_mib(struct hostapd_data *hapd, char *buf, size_t buflen);
+int ieee802_1x_get_mib_sta(struct hostapd_data *hapd, struct sta_info *sta,
+			   char *buf, size_t buflen);
+void hostapd_get_ntp_timestamp(u8 *buf);
 
 #endif /* IEEE802_1X_H */
