@@ -1,12 +1,16 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver / RADIUS client
- * Copyright (c) 2002-2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2002-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -16,6 +20,7 @@
 #include <string.h>
 #include <time.h>
 #include <sys/types.h>
+#include <sys/time.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <errno.h>
@@ -26,7 +31,7 @@
 #include "eloop.h"
 
 /* Defaults for RADIUS retransmit values (exponential backoff) */
-#define RADIUS_CLIENT_FIRST_WAIT 1 /* seconds */
+#define RADIUS_CLIENT_FIRST_WAIT 3 /* seconds */
 #define RADIUS_CLIENT_MAX_WAIT 120 /* seconds */
 #define RADIUS_CLIENT_MAX_RETRIES 10 /* maximum number of retransmit attempts
 				      * before entry is removed from retransmit
@@ -38,13 +43,61 @@
 				      * many failed retry attempts */
 
 
+struct radius_rx_handler {
+	RadiusRxResult (*handler)(struct radius_msg *msg,
+				  struct radius_msg *req,
+				  u8 *shared_secret, size_t shared_secret_len,
+				  void *data);
+	void *data;
+};
+
+
+/* RADIUS message retransmit list */
+struct radius_msg_list {
+	u8 addr[ETH_ALEN]; /* STA/client address; used to find RADIUS messages
+			    * for the same STA. */
+	struct radius_msg *msg;
+	RadiusType msg_type;
+	time_t first_try;
+	time_t next_try;
+	int attempts;
+	int next_wait;
+	struct timeval last_attempt;
+
+	u8 *shared_secret;
+	size_t shared_secret_len;
+
+	/* TODO: server config with failover to backup server(s) */
+
+	struct radius_msg_list *next;
+};
+
+
+struct radius_client_data {
+	struct hostapd_data *hapd;
+
+	int auth_serv_sock; /* socket for authentication RADIUS messages */
+	int acct_serv_sock; /* socket for accounting RADIUS messages */
+
+	struct radius_rx_handler *auth_handlers;
+	size_t num_auth_handlers;
+	struct radius_rx_handler *acct_handlers;
+	size_t num_acct_handlers;
+
+	struct radius_msg_list *msgs;
+	size_t num_msgs;
+
+	u8 next_radius_identifier;
+};
+
 
 static int
-radius_change_server(hostapd *hapd, struct hostapd_radius_server *nserv,
+radius_change_server(struct radius_client_data *radius,
+		     struct hostapd_radius_server *nserv,
 		     struct hostapd_radius_server *oserv,
 		     int sock, int auth);
-static int radius_client_init_acct(hostapd *hapd);
-static int radius_client_init_auth(hostapd *hapd);
+static int radius_client_init_acct(struct radius_client_data *radius);
+static int radius_client_init_auth(struct radius_client_data *radius);
 
 
 static void radius_client_msg_free(struct radius_msg_list *req)
@@ -55,9 +108,9 @@ static void radius_client_msg_free(struct radius_msg_list *req)
 }
 
 
-int radius_client_register(hostapd *hapd, RadiusType msg_type,
-			   RadiusRxResult (*handler)(hostapd *hapd,
-						     struct radius_msg *msg,
+int radius_client_register(struct radius_client_data *radius,
+			   RadiusType msg_type,
+			   RadiusRxResult (*handler)(struct radius_msg *msg,
 						     struct radius_msg *req,
 						     u8 *shared_secret,
 						     size_t shared_secret_len,
@@ -68,11 +121,11 @@ int radius_client_register(hostapd *hapd, RadiusType msg_type,
 	size_t *num;
 
 	if (msg_type == RADIUS_ACCT) {
-		handlers = &hapd->radius->acct_handlers;
-		num = &hapd->radius->num_acct_handlers;
+		handlers = &radius->acct_handlers;
+		num = &radius->num_acct_handlers;
 	} else {
-		handlers = &hapd->radius->auth_handlers;
-		num = &hapd->radius->num_auth_handlers;
+		handlers = &radius->auth_handlers;
+		num = &radius->num_auth_handlers;
 	}
 
 	newh = (struct radius_rx_handler *)
@@ -90,9 +143,10 @@ int radius_client_register(hostapd *hapd, RadiusType msg_type,
 }
 
 
-static void radius_client_handle_send_error(struct hostapd_data *hapd, int s,
-					    RadiusType msg_type)
+static void radius_client_handle_send_error(struct radius_client_data *radius,
+					    int s, RadiusType msg_type)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	int _errno = errno;
 	perror("send[RADIUS]");
 	if (_errno == ENOTCONN || _errno == EDESTADDRREQ || _errno == EINVAL) {
@@ -103,31 +157,46 @@ static void radius_client_handle_send_error(struct hostapd_data *hapd, int s,
 		eloop_unregister_read_sock(s);
 		close(s);
 		if (msg_type == RADIUS_ACCT || msg_type == RADIUS_ACCT_INTERIM)
-			radius_client_init_acct(hapd);
+			radius_client_init_acct(radius);
 		else
-			radius_client_init_auth(hapd);
+			radius_client_init_auth(radius);
 	}
 }
 
 
-static int radius_client_retransmit(hostapd *hapd,
+static int radius_client_retransmit(struct radius_client_data *radius,
 				    struct radius_msg_list *entry, time_t now)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	int s;
 
 	if (entry->msg_type == RADIUS_ACCT ||
-	    entry->msg_type == RADIUS_ACCT_INTERIM)
-		s = hapd->radius->acct_serv_sock;
-	else
-		s = hapd->radius->auth_serv_sock;
+	    entry->msg_type == RADIUS_ACCT_INTERIM) {
+		s = radius->acct_serv_sock;
+		if (entry->attempts == 0)
+			hapd->conf->acct_server->requests++;
+		else {
+			hapd->conf->acct_server->timeouts++;
+			hapd->conf->acct_server->retransmissions++;
+		}
+	} else {
+		s = radius->auth_serv_sock;
+		if (entry->attempts == 0)
+			hapd->conf->auth_server->requests++;
+		else {
+			hapd->conf->auth_server->timeouts++;
+			hapd->conf->auth_server->retransmissions++;
+		}
+	}
 
 	/* retransmit; remove entry if too many attempts */
 	entry->attempts++;
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "Resending RADIUS message (id=%d)"
 		      "\n", entry->msg->hdr->identifier);
 
+	gettimeofday(&entry->last_attempt, NULL);
 	if (send(s, entry->msg->buf, entry->msg->buf_used, 0) < 0)
-		radius_client_handle_send_error(hapd, s, entry->msg_type);
+		radius_client_handle_send_error(radius, s, entry->msg_type);
 
 	entry->next_try = now + entry->next_wait;
 	entry->next_wait *= 2;
@@ -145,12 +214,13 @@ static int radius_client_retransmit(hostapd *hapd,
 
 static void radius_client_timer(void *eloop_ctx, void *timeout_ctx)
 {
-	hostapd *hapd = eloop_ctx;
+	struct radius_client_data *radius = eloop_ctx;
+	struct hostapd_data *hapd = radius->hapd;
 	time_t now, first;
 	struct radius_msg_list *entry, *prev, *tmp;
 	int auth_failover = 0, acct_failover = 0;
 
-	entry = hapd->radius->msgs;
+	entry = radius->msgs;
 	if (!entry)
 		return;
 
@@ -160,16 +230,16 @@ static void radius_client_timer(void *eloop_ctx, void *timeout_ctx)
 	prev = NULL;
 	while (entry) {
 		if (now >= entry->next_try &&
-		    radius_client_retransmit(hapd, entry, now)) {
+		    radius_client_retransmit(radius, entry, now)) {
 			if (prev)
 				prev->next = entry->next;
 			else
-				hapd->radius->msgs = entry->next;
+				radius->msgs = entry->next;
 
 			tmp = entry;
 			entry = entry->next;
 			radius_client_msg_free(tmp);
-			hapd->radius->num_msgs--;
+			radius->num_msgs--;
 			continue;
 		}
 
@@ -188,11 +258,15 @@ static void radius_client_timer(void *eloop_ctx, void *timeout_ctx)
 		entry = entry->next;
 	}
 
-	if (hapd->radius->msgs) {
+	if (radius->msgs) {
 		if (first < now)
 			first = now;
 		eloop_register_timeout(first - now, 0,
-				       radius_client_timer, hapd, NULL);
+				       radius_client_timer, radius, NULL);
+		HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "Next RADIUS client "
+			      "retransmit in %ld seconds\n",
+			      (long int) (first - now));
+
 	}
 
 	if (auth_failover && hapd->conf->num_auth_servers > 1) {
@@ -204,13 +278,18 @@ static void radius_client_timer(void *eloop_ctx, void *timeout_ctx)
 			       "%s:%d - failover",
 			       inet_ntoa(old->addr), old->port);
 
+		for (entry = radius->msgs; entry; entry = entry->next) {
+			if (entry->msg_type == RADIUS_AUTH)
+				old->timeouts++;
+		}
+
 		next = old + 1;
 		if (next > &(hapd->conf->auth_servers
 			     [hapd->conf->num_auth_servers - 1]))
 			next = hapd->conf->auth_servers;
 		hapd->conf->auth_server = next;
-		radius_change_server(hapd, next, old,
-				     hapd->radius->auth_serv_sock, 1);
+		radius_change_server(radius, next, old,
+				     radius->auth_serv_sock, 1);
 	}
 
 	if (acct_failover && hapd->conf->num_acct_servers > 1) {
@@ -221,18 +300,54 @@ static void radius_client_timer(void *eloop_ctx, void *timeout_ctx)
 			       "No response from Accounting server "
 			       "%s:%d - failover",
 			       inet_ntoa(old->addr), old->port);
+
+		for (entry = radius->msgs; entry; entry = entry->next) {
+			if (entry->msg_type == RADIUS_ACCT ||
+			    entry->msg_type == RADIUS_ACCT_INTERIM)
+				old->timeouts++;
+		}
+
 		next = old + 1;
 		if (next > &hapd->conf->acct_servers
 		    [hapd->conf->num_acct_servers - 1])
 			next = hapd->conf->acct_servers;
 		hapd->conf->acct_server = next;
-		radius_change_server(hapd, next, old,
-				     hapd->radius->acct_serv_sock, 0);
+		radius_change_server(radius, next, old,
+				     radius->acct_serv_sock, 0);
 	}
 }
 
 
-static void radius_client_list_add(hostapd *hapd, struct radius_msg *msg,
+static void radius_client_update_timeout(struct radius_client_data *radius)
+{
+	struct hostapd_data *hapd = radius->hapd;
+	time_t now, first;
+	struct radius_msg_list *entry;
+
+	eloop_cancel_timeout(radius_client_timer, radius, NULL);
+
+	if (radius->msgs == NULL) {
+		return;
+	}
+
+	first = 0;
+	for (entry = radius->msgs; entry; entry = entry->next) {
+		if (first == 0 || entry->next_try < first)
+			first = entry->next_try;
+	}
+
+	time(&now);
+	if (first < now)
+		first = now;
+	eloop_register_timeout(first - now, 0, radius_client_timer, radius,
+			       NULL);
+	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "Next RADIUS client retransmit in"
+		      " %ld seconds\n", (long int) (first - now));
+}
+
+
+static void radius_client_list_add(struct radius_client_data *radius,
+				   struct radius_msg *msg,
 				   RadiusType msg_type, u8 *shared_secret,
 				   size_t shared_secret_len, u8 *addr)
 {
@@ -264,16 +379,13 @@ static void radius_client_list_add(hostapd *hapd, struct radius_msg *msg,
 	time(&entry->first_try);
 	entry->next_try = entry->first_try + RADIUS_CLIENT_FIRST_WAIT;
 	entry->attempts = 1;
+	gettimeofday(&entry->last_attempt, NULL);
 	entry->next_wait = RADIUS_CLIENT_FIRST_WAIT * 2;
+	entry->next = radius->msgs;
+	radius->msgs = entry;
+	radius_client_update_timeout(radius);
 
-	if (!hapd->radius->msgs)
-		eloop_register_timeout(RADIUS_CLIENT_FIRST_WAIT, 0,
-				       radius_client_timer, hapd, NULL);
-
-	entry->next = hapd->radius->msgs;
-	hapd->radius->msgs = entry;
-
-	if (hapd->radius->num_msgs >= RADIUS_CLIENT_MAX_ENTRIES) {
+	if (radius->num_msgs >= RADIUS_CLIENT_MAX_ENTRIES) {
 		printf("Removing the oldest un-ACKed RADIUS packet due to "
 		       "retransmit list limits.\n");
 		prev = NULL;
@@ -286,19 +398,20 @@ static void radius_client_list_add(hostapd *hapd, struct radius_msg *msg,
 			radius_client_msg_free(entry);
 		}
 	} else
-		hapd->radius->num_msgs++;
+		radius->num_msgs++;
 }
 
 
-static void radius_client_list_del(struct hostapd_data *hapd,
+static void radius_client_list_del(struct radius_client_data *radius,
 				   RadiusType msg_type, u8 *addr)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	struct radius_msg_list *entry, *prev, *tmp;
 
 	if (addr == NULL)
 		return;
 
-	entry = hapd->radius->msgs;
+	entry = radius->msgs;
 	prev = NULL;
 	while (entry) {
 		if (entry->msg_type == msg_type &&
@@ -306,14 +419,14 @@ static void radius_client_list_del(struct hostapd_data *hapd,
 			if (prev)
 				prev->next = entry->next;
 			else
-				hapd->radius->msgs = entry->next;
+				radius->msgs = entry->next;
 			tmp = entry;
 			entry = entry->next;
 			HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 				      "Removing matching RADIUS message for "
 				      MACSTR "\n", MAC2STR(addr));
 			radius_client_msg_free(tmp);
-			hapd->radius->num_msgs--;
+			radius->num_msgs--;
 			continue;
 		}
 		prev = entry;
@@ -322,9 +435,10 @@ static void radius_client_list_del(struct hostapd_data *hapd,
 }
 
 
-int radius_client_send(hostapd *hapd, struct radius_msg *msg,
-		       RadiusType msg_type, u8 *addr)
+int radius_client_send(struct radius_client_data *radius,
+		       struct radius_msg *msg, RadiusType msg_type, u8 *addr)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	u8 *shared_secret;
 	size_t shared_secret_len;
 	char *name;
@@ -332,7 +446,7 @@ int radius_client_send(hostapd *hapd, struct radius_msg *msg,
 
 	if (msg_type == RADIUS_ACCT_INTERIM) {
 		/* Remove any pending interim acct update for the same STA. */
-		radius_client_list_del(hapd, msg_type, addr);
+		radius_client_list_del(radius, msg_type, addr);
 	}
 
 	if (msg_type == RADIUS_ACCT || msg_type == RADIUS_ACCT_INTERIM) {
@@ -340,13 +454,15 @@ int radius_client_send(hostapd *hapd, struct radius_msg *msg,
 		shared_secret_len = hapd->conf->acct_server->shared_secret_len;
 		radius_msg_finish_acct(msg, shared_secret, shared_secret_len);
 		name = "accounting";
-		s = hapd->radius->acct_serv_sock;
+		s = radius->acct_serv_sock;
+		hapd->conf->acct_server->requests++;
 	} else {
 		shared_secret = hapd->conf->auth_server->shared_secret;
 		shared_secret_len = hapd->conf->auth_server->shared_secret_len;
 		radius_msg_finish(msg, shared_secret, shared_secret_len);
 		name = "authentication";
-		s = hapd->radius->auth_serv_sock;
+		s = radius->auth_serv_sock;
+		hapd->conf->auth_server->requests++;
 	}
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
@@ -356,9 +472,9 @@ int radius_client_send(hostapd *hapd, struct radius_msg *msg,
 
 	res = send(s, msg->buf, msg->buf_used, 0);
 	if (res < 0)
-		radius_client_handle_send_error(hapd, s, msg_type);
+		radius_client_handle_send_error(radius, s, msg_type);
 
-	radius_client_list_add(hapd, msg, msg_type, shared_secret,
+	radius_client_list_add(radius, msg, msg_type, shared_secret,
 			       shared_secret_len, addr);
 
 	return res;
@@ -367,14 +483,28 @@ int radius_client_send(hostapd *hapd, struct radius_msg *msg,
 
 static void radius_client_receive(int sock, void *eloop_ctx, void *sock_ctx)
 {
-	hostapd *hapd = (hostapd *) eloop_ctx;
+	struct radius_client_data *radius = eloop_ctx;
+	struct hostapd_data *hapd = radius->hapd;
 	RadiusType msg_type = (RadiusType) sock_ctx;
-	int len, i;
+	int len, i, roundtrip;
 	unsigned char buf[3000];
 	struct radius_msg *msg;
 	struct radius_rx_handler *handlers;
 	size_t num_handlers;
 	struct radius_msg_list *req, *prev_req;
+	struct timeval tv;
+	struct hostapd_radius_server *rconf;
+	int invalid_authenticator = 0;
+
+	if (msg_type == RADIUS_ACCT) {
+		handlers = radius->acct_handlers;
+		num_handlers = radius->num_acct_handlers;
+		rconf = hapd->conf->acct_server;
+	} else {
+		handlers = radius->auth_handlers;
+		num_handlers = radius->num_auth_handlers;
+		rconf = hapd->conf->auth_server;
+	}
 
 	len = recv(sock, buf, sizeof(buf), 0);
 	if (len < 0) {
@@ -392,6 +522,7 @@ static void radius_client_receive(int sock, void *eloop_ctx, void *sock_ctx)
 	msg = radius_msg_parse(buf, len);
 	if (msg == NULL) {
 		printf("Parsing incoming RADIUS frame failed\n");
+		rconf->malformed_responses++;
 		return;
 	}
 
@@ -400,16 +531,23 @@ static void radius_client_receive(int sock, void *eloop_ctx, void *sock_ctx)
 	if (HOSTAPD_DEBUG_COND(HOSTAPD_DEBUG_MSGDUMPS))
 		radius_msg_dump(msg);
 
-	if (msg_type == RADIUS_ACCT) {
-		handlers = hapd->radius->acct_handlers;
-		num_handlers = hapd->radius->num_acct_handlers;
-	} else {
-		handlers = hapd->radius->auth_handlers;
-		num_handlers = hapd->radius->num_auth_handlers;
+	switch (msg->hdr->code) {
+	case RADIUS_CODE_ACCESS_ACCEPT:
+		rconf->access_accepts++;
+		break;
+	case RADIUS_CODE_ACCESS_REJECT:
+		rconf->access_rejects++;
+		break;
+	case RADIUS_CODE_ACCESS_CHALLENGE:
+		rconf->access_challenges++;
+		break;
+	case RADIUS_CODE_ACCOUNTING_RESPONSE:
+		rconf->responses++;
+		break;
 	}
 
 	prev_req = NULL;
-	req = hapd->radius->msgs;
+	req = radius->msgs;
 	while (req) {
 		/* TODO: also match by src addr:port of the packet when using
 		 * alternative RADIUS servers (?) */
@@ -431,17 +569,24 @@ static void radius_client_receive(int sock, void *eloop_ctx, void *sock_ctx)
 		goto fail;
 	}
 
+	gettimeofday(&tv, NULL);
+	roundtrip = (tv.tv_sec - req->last_attempt.tv_sec) * 100 +
+		(tv.tv_usec - req->last_attempt.tv_usec) / 10000;
+	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "Received RADIUS packet matched "
+		      "with a pending request, round trip time %d.%02d sec\n",
+		      roundtrip / 100, roundtrip % 100);
+	rconf->round_trip_time = roundtrip;
+
 	/* Remove ACKed RADIUS packet from retransmit list */
 	if (prev_req)
 		prev_req->next = req->next;
 	else
-		hapd->radius->msgs = req->next;
-	hapd->radius->num_msgs--;
+		radius->msgs = req->next;
+	radius->num_msgs--;
 
 	for (i = 0; i < num_handlers; i++) {
 		RadiusRxResult res;
-		res = handlers[i].handler(hapd, msg, req->msg,
-					  req->shared_secret,
+		res = handlers[i].handler(msg, req->msg, req->shared_secret,
 					  req->shared_secret_len,
 					  handlers[i].data);
 		switch (res) {
@@ -452,14 +597,25 @@ static void radius_client_receive(int sock, void *eloop_ctx, void *sock_ctx)
 		case RADIUS_RX_QUEUED:
 			radius_client_msg_free(req);
 			return;
+		case RADIUS_RX_INVALID_AUTHENTICATOR:
+			invalid_authenticator++;
+			/* continue */
 		case RADIUS_RX_UNKNOWN:
 			/* continue with next handler */
 			break;
 		}
 	}
 
-	printf("No RADIUS RX handler found (type=%d code=%d id=%d) - dropping "
-	       "packet\n", msg_type, msg->hdr->code, msg->hdr->identifier);
+	if (invalid_authenticator)
+		rconf->bad_authenticators++;
+	else
+		rconf->unknown_types++;
+	hostapd_logger(hapd, req->addr, HOSTAPD_MODULE_RADIUS,
+		       HOSTAPD_LEVEL_DEBUG, "No RADIUS RX handler found "
+		       "(type=%d code=%d id=%d)%s - dropping packet",
+		       msg_type, msg->hdr->code, msg->hdr->identifier,
+		       invalid_authenticator ? " [INVALID AUTHENTICATOR]" :
+		       "");
 	radius_client_msg_free(req);
 
  fail:
@@ -468,14 +624,15 @@ static void radius_client_receive(int sock, void *eloop_ctx, void *sock_ctx)
 }
 
 
-u8 radius_client_get_id(hostapd *hapd)
+u8 radius_client_get_id(struct radius_client_data *radius)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	struct radius_msg_list *entry, *prev, *remove;
-	u8 id = hapd->radius->next_radius_identifier++;
+	u8 id = radius->next_radius_identifier++;
 
 	/* remove entries with matching id from retransmit list to avoid
 	 * using new reply from the RADIUS server with an old request */
-	entry = hapd->radius->msgs;
+	entry = radius->msgs;
 	prev = NULL;
 	while (entry) {
 		if (entry->msg->hdr->identifier == id) {
@@ -485,7 +642,7 @@ u8 radius_client_get_id(hostapd *hapd)
 			if (prev)
 				prev->next = entry->next;
 			else
-				hapd->radius->msgs = entry->next;
+				radius->msgs = entry->next;
 			remove = entry;
 		} else
 			remove = NULL;
@@ -500,18 +657,18 @@ u8 radius_client_get_id(hostapd *hapd)
 }
 
 
-void radius_client_flush(hostapd *hapd)
+void radius_client_flush(struct radius_client_data *radius)
 {
 	struct radius_msg_list *entry, *prev;
 
-	if (!hapd->radius)
+	if (!radius)
 		return;
 
-	eloop_cancel_timeout(radius_client_timer, hapd, NULL);
+	eloop_cancel_timeout(radius_client_timer, radius, NULL);
 
-	entry = hapd->radius->msgs;
-	hapd->radius->msgs = NULL;
-	hapd->radius->num_msgs = 0;
+	entry = radius->msgs;
+	radius->msgs = NULL;
+	radius->num_msgs = 0;
 	while (entry) {
 		prev = entry;
 		entry = entry->next;
@@ -521,10 +678,12 @@ void radius_client_flush(hostapd *hapd)
 
 
 static int
-radius_change_server(hostapd *hapd, struct hostapd_radius_server *nserv,
+radius_change_server(struct radius_client_data *radius,
+		     struct hostapd_radius_server *nserv,
 		     struct hostapd_radius_server *oserv,
 		     int sock, int auth)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	struct sockaddr_in serv;
 
 	hostapd_logger(hapd, NULL, HOSTAPD_MODULE_RADIUS, HOSTAPD_LEVEL_INFO,
@@ -540,11 +699,11 @@ radius_change_server(hostapd *hapd, struct hostapd_radius_server *nserv,
 		 * update all message authenticators and
 		 * User-Passwords, etc. and retry with new server. For
 		 * now, just drop all pending packets. */
-		radius_client_flush(hapd);
+		radius_client_flush(radius);
 	} else {
 		/* Reset retry counters for the new server */
 		struct radius_msg_list *entry;
-		entry = hapd->radius->msgs;
+		entry = radius->msgs;
 		while (entry) {
 			entry->next_try = entry->first_try +
 				RADIUS_CLIENT_FIRST_WAIT;
@@ -552,10 +711,11 @@ radius_change_server(hostapd *hapd, struct hostapd_radius_server *nserv,
 			entry->next_wait = RADIUS_CLIENT_FIRST_WAIT * 2;
 			entry = entry->next;
 		}
-		if (hapd->radius->msgs) {
-			eloop_cancel_timeout(radius_client_timer, hapd, NULL);
+		if (radius->msgs) {
+			eloop_cancel_timeout(radius_client_timer, radius,
+					     NULL);
 			eloop_register_timeout(RADIUS_CLIENT_FIRST_WAIT, 0,
-					       radius_client_timer, hapd,
+					       radius_client_timer, radius,
 					       NULL);
 		}
 	}
@@ -576,45 +736,48 @@ radius_change_server(hostapd *hapd, struct hostapd_radius_server *nserv,
 
 static void radius_retry_primary_timer(void *eloop_ctx, void *timeout_ctx)
 {
-	hostapd *hapd = eloop_ctx;
+	struct radius_client_data *radius = eloop_ctx;
+	struct hostapd_data *hapd = radius->hapd;
 	struct hostapd_radius_server *oserv;
 
-	if (hapd->radius->auth_serv_sock >= 0 && hapd->conf->auth_servers &&
+	if (radius->auth_serv_sock >= 0 && hapd->conf->auth_servers &&
 	    hapd->conf->auth_server != hapd->conf->auth_servers) {
 		oserv = hapd->conf->auth_server;
 		hapd->conf->auth_server = hapd->conf->auth_servers;
-		radius_change_server(hapd, hapd->conf->auth_server, oserv,
-				     hapd->radius->auth_serv_sock, 1);
+		radius_change_server(radius, hapd->conf->auth_server, oserv,
+				     radius->auth_serv_sock, 1);
 	}
 
-	if (hapd->radius->acct_serv_sock >= 0 && hapd->conf->acct_servers &&
+	if (radius->acct_serv_sock >= 0 && hapd->conf->acct_servers &&
 	    hapd->conf->acct_server != hapd->conf->acct_servers) {
 		oserv = hapd->conf->acct_server;
 		hapd->conf->acct_server = hapd->conf->acct_servers;
-		radius_change_server(hapd, hapd->conf->acct_server, oserv,
-				     hapd->radius->acct_serv_sock, 0);
+		radius_change_server(radius, hapd->conf->acct_server, oserv,
+				     radius->acct_serv_sock, 0);
 	}
 
 	if (hapd->conf->radius_retry_primary_interval)
 		eloop_register_timeout(hapd->conf->
 				       radius_retry_primary_interval, 0,
-				       radius_retry_primary_timer, hapd, NULL);
+				       radius_retry_primary_timer, radius,
+				       NULL);
 }
 
 
-static int radius_client_init_auth(hostapd *hapd)
+static int radius_client_init_auth(struct radius_client_data *radius)
 {
-	hapd->radius->auth_serv_sock = socket(PF_INET, SOCK_DGRAM, 0);
-	if (hapd->radius->auth_serv_sock < 0) {
+	struct hostapd_data *hapd = radius->hapd;
+	radius->auth_serv_sock = socket(PF_INET, SOCK_DGRAM, 0);
+	if (radius->auth_serv_sock < 0) {
 		perror("socket[PF_INET,SOCK_DGRAM]");
 		return -1;
 	}
 
-	radius_change_server(hapd, hapd->conf->auth_server, NULL,
-			     hapd->radius->auth_serv_sock, 1);
+	radius_change_server(radius, hapd->conf->auth_server, NULL,
+			     radius->auth_serv_sock, 1);
 
-	if (eloop_register_read_sock(hapd->radius->auth_serv_sock,
-				     radius_client_receive, hapd,
+	if (eloop_register_read_sock(radius->auth_serv_sock,
+				     radius_client_receive, radius,
 				     (void *) RADIUS_AUTH)) {
 		printf("Could not register read socket for authentication "
 		       "server\n");
@@ -625,19 +788,20 @@ static int radius_client_init_auth(hostapd *hapd)
 }
 
 
-static int radius_client_init_acct(hostapd *hapd)
+static int radius_client_init_acct(struct radius_client_data *radius)
 {
-	hapd->radius->acct_serv_sock = socket(PF_INET, SOCK_DGRAM, 0);
-	if (hapd->radius->acct_serv_sock < 0) {
+	struct hostapd_data *hapd = radius->hapd;
+	radius->acct_serv_sock = socket(PF_INET, SOCK_DGRAM, 0);
+	if (radius->acct_serv_sock < 0) {
 		perror("socket[PF_INET,SOCK_DGRAM]");
 		return -1;
 	}
 
-	radius_change_server(hapd, hapd->conf->acct_server, NULL,
-			     hapd->radius->acct_serv_sock, 0);
+	radius_change_server(radius, hapd->conf->acct_server, NULL,
+			     radius->acct_serv_sock, 0);
 
-	if (eloop_register_read_sock(hapd->radius->acct_serv_sock,
-				     radius_client_receive, hapd,
+	if (eloop_register_read_sock(radius->acct_serv_sock,
+				     radius_client_receive, radius,
 				     (void *) RADIUS_ACCT)) {
 		printf("Could not register read socket for accounting "
 		       "server\n");
@@ -648,51 +812,59 @@ static int radius_client_init_acct(hostapd *hapd)
 }
 
 
-int radius_client_init(hostapd *hapd)
+struct radius_client_data * radius_client_init(struct hostapd_data *hapd)
 {
-	hapd->radius = malloc(sizeof(struct radius_client_data));
-	if (hapd->radius == NULL)
-		return -1;
+	struct radius_client_data *radius;
 
-	memset(hapd->radius, 0, sizeof(struct radius_client_data));
-	hapd->radius->auth_serv_sock = hapd->radius->acct_serv_sock = -1;
+	radius = malloc(sizeof(struct radius_client_data));
+	if (radius == NULL)
+		return NULL;
 
-	if (hapd->conf->auth_server && radius_client_init_auth(hapd))
-		return -1;
+	memset(radius, 0, sizeof(struct radius_client_data));
+	radius->hapd = hapd;
+	radius->auth_serv_sock = radius->acct_serv_sock = -1;
 
-	if (hapd->conf->acct_server && radius_client_init_acct(hapd))
-		return -1;
+	if (hapd->conf->auth_server && radius_client_init_auth(radius)) {
+		radius_client_deinit(radius);
+		return NULL;
+	}
+
+	if (hapd->conf->acct_server && radius_client_init_acct(radius)) {
+		radius_client_deinit(radius);
+		return NULL;
+	}
 
 	if (hapd->conf->radius_retry_primary_interval)
 		eloop_register_timeout(hapd->conf->
 				       radius_retry_primary_interval, 0,
-				       radius_retry_primary_timer, hapd, NULL);
+				       radius_retry_primary_timer, radius,
+				       NULL);
 
-	return 0;
+	return radius;
 }
 
 
-void radius_client_deinit(hostapd *hapd)
+void radius_client_deinit(struct radius_client_data *radius)
 {
-	if (!hapd->radius)
+	if (!radius)
 		return;
 
-	eloop_cancel_timeout(radius_retry_primary_timer, hapd, NULL);
+	eloop_cancel_timeout(radius_retry_primary_timer, radius, NULL);
 
-	radius_client_flush(hapd);
-	free(hapd->radius->auth_handlers);
-	free(hapd->radius->acct_handlers);
-	free(hapd->radius);
-	hapd->radius = NULL;
+	radius_client_flush(radius);
+	free(radius->auth_handlers);
+	free(radius->acct_handlers);
+	free(radius);
 }
 
 
-void radius_client_flush_auth(struct hostapd_data *hapd, u8 *addr)
+void radius_client_flush_auth(struct radius_client_data *radius, u8 *addr)
 {
+	struct hostapd_data *hapd = radius->hapd;
 	struct radius_msg_list *entry, *prev, *tmp;
 
 	prev = NULL;
-	entry = hapd->radius->msgs;
+	entry = radius->msgs;
 	while (entry) {
 		if (entry->msg_type == RADIUS_AUTH &&
 		    memcmp(entry->addr, addr, ETH_ALEN) == 0) {
@@ -704,12 +876,12 @@ void radius_client_flush_auth(struct hostapd_data *hapd, u8 *addr)
 			if (prev)
 				prev->next = entry->next;
 			else
-				hapd->radius->msgs = entry->next;
+				radius->msgs = entry->next;
 
 			tmp = entry;
 			entry = entry->next;
 			radius_client_msg_free(tmp);
-			hapd->radius->num_msgs--;
+			radius->num_msgs--;
 			continue;
 		}
 
@@ -717,3 +889,128 @@ void radius_client_flush_auth(struct hostapd_data *hapd, u8 *addr)
 		entry = entry->next;
 	}
 }
+
+
+static int radius_client_dump_auth_server(char *buf, size_t buflen,
+					  struct hostapd_radius_server *serv,
+					  struct radius_client_data *cli)
+{
+	int pending = 0;
+	struct radius_msg_list *msg;
+
+	if (cli) {
+		for (msg = cli->msgs; msg; msg = msg->next) {
+			if (msg->msg_type == RADIUS_AUTH)
+				pending++;
+		}
+	}
+
+	return snprintf(buf, buflen,
+			"radiusAuthServerIndex=%d\n"
+			"radiusAuthServerAddress=%s\n"
+			"radiusAuthClientServerPortNumber=%d\n"
+			"radiusAuthClientRoundTripTime=%d\n"
+			"radiusAuthClientAccessRequests=%u\n"
+			"radiusAuthClientAccessRetransmissions=%u\n"
+			"radiusAuthClientAccessAccepts=%u\n"
+			"radiusAuthClientAccessRejects=%u\n"
+			"radiusAuthClientAccessChallenges=%u\n"
+			"radiusAuthClientMalformedAccessResponses=%u\n"
+			"radiusAuthClientBadAuthenticators=%u\n"
+			"radiusAuthClientPendingRequests=%u\n"
+			"radiusAuthClientTimeouts=%u\n"
+			"radiusAuthClientUnknownTypes=%u\n"
+			"radiusAuthClientPacketsDropped=%u\n",
+			serv->index,
+			inet_ntoa(serv->addr),
+			serv->port,
+			serv->round_trip_time,
+			serv->requests,
+			serv->retransmissions,
+			serv->access_accepts,
+			serv->access_rejects,
+			serv->access_challenges,
+			serv->malformed_responses,
+			serv->bad_authenticators,
+			pending,
+			serv->timeouts,
+			serv->unknown_types,
+			serv->packets_dropped);
+}
+
+
+static int radius_client_dump_acct_server(char *buf, size_t buflen,
+					  struct hostapd_radius_server *serv,
+					  struct radius_client_data *cli)
+{
+	int pending = 0;
+	struct radius_msg_list *msg;
+
+	if (cli) {
+		for (msg = cli->msgs; msg; msg = msg->next) {
+			if (msg->msg_type == RADIUS_ACCT ||
+			    msg->msg_type == RADIUS_ACCT_INTERIM)
+				pending++;
+		}
+	}
+
+	return snprintf(buf, buflen,
+			"radiusAccServerIndex=%d\n"
+			"radiusAccServerAddress=%s\n"
+			"radiusAccClientServerPortNumber=%d\n"
+			"radiusAccClientRoundTripTime=%d\n"
+			"radiusAccClientRequests=%u\n"
+			"radiusAccClientRetransmissions=%u\n"
+			"radiusAccClientResponses=%u\n"
+			"radiusAccClientMalformedResponses=%u\n"
+			"radiusAccClientBadAuthenticators=%u\n"
+			"radiusAccClientPendingRequests=%u\n"
+			"radiusAccClientTimeouts=%u\n"
+			"radiusAccClientUnknownTypes=%u\n"
+			"radiusAccClientPacketsDropped=%u\n",
+			serv->index,
+			inet_ntoa(serv->addr),
+			serv->port,
+			serv->round_trip_time,
+			serv->requests,
+			serv->retransmissions,
+			serv->responses,
+			serv->malformed_responses,
+			serv->bad_authenticators,
+			pending,
+			serv->timeouts,
+			serv->unknown_types,
+			serv->packets_dropped);
+}
+
+
+int radius_client_get_mib(struct radius_client_data *radius, char *buf,
+			  size_t buflen)
+{
+	struct hostapd_data *hapd = radius->hapd;
+	int i;
+	struct hostapd_radius_server *serv;
+	int count = 0;
+
+	if (hapd->conf->auth_servers) {
+		for (i = 0; i < hapd->conf->num_auth_servers; i++) {
+			serv = &hapd->conf->auth_servers[i];
+			count += radius_client_dump_auth_server(
+				buf + count, buflen - count, serv,
+				serv == hapd->conf->auth_server ?
+				radius : NULL);
+		}
+	}
+
+	if (hapd->conf->acct_servers) {
+		for (i = 0; i < hapd->conf->num_acct_servers; i++) {
+			serv = &hapd->conf->acct_servers[i];
+			count += radius_client_dump_acct_server(
+				buf + count, buflen - count, serv,
+				serv == hapd->conf->acct_server ?
+				radius : NULL);
+		}
+	}
+
+	return count;
+}
