@@ -1,12 +1,16 @@
 /*
  * Host AP (software wireless LAN access point) user space daemon for
  * Host AP kernel driver / WPA Authenticator
- * Copyright (c) 2004, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See README and COPYING for
- * more details.
+ * published by the Free Software Foundation.
+ *
+ * Alternatively, this software may be distributed under the terms of BSD
+ * license.
+ *
+ * See README and COPYING for more details.
  */
 
 #include <stdlib.h>
@@ -16,11 +20,6 @@
 #include <sys/time.h>
 #include <time.h>
 #include <unistd.h>
-#include <sys/ioctl.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <netpacket/packet.h>
-#include <net/if.h>
 
 #include "hostapd.h"
 #include "eapol_sm.h"
@@ -34,6 +33,7 @@
 #include "ieee802_11.h"
 #include "eloop.h"
 #include "sta_info.h"
+#include "l2_packet.h"
 
 
 static void wpa_send_eapol_timeout(void *eloop_ctx, void *timeout_ctx);
@@ -344,10 +344,8 @@ static int wpa_gen_wpa_ie(struct hostapd_data *hapd)
 
 static void wpa_sta_disconnect(struct hostapd_data *hapd, struct sta_info *sta)
 {
-	ieee802_11_send_deauth(hapd, sta->addr,
-			       WLAN_REASON_PREV_AUTH_NOT_VALID);
+	hostapd_sta_deauth(hapd, sta->addr, WLAN_REASON_PREV_AUTH_NOT_VALID);
 	sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC | WLAN_STA_AUTHORIZED);
-	remove_sta(hapd->driver.data, sta->addr);
 	eloop_cancel_timeout(ap_handle_timer, hapd, sta);
 	eloop_register_timeout(0, 0, ap_handle_timer, hapd, sta);
 	sta->timeout_next = STA_REMOVE;
@@ -396,37 +394,22 @@ static void wpa_rekey_gtk(void *eloop_ctx, void *timeout_ctx)
 }
 
 
-struct l2_ethhdr {
-	u8 h_dest[ETH_ALEN];
-	u8 h_source[ETH_ALEN];
-	u16 h_proto;
-} __attribute__ ((packed));
+#ifdef CONFIG_RSN_PREAUTH
 
-
-static void rsn_preauth_receive(int sock, void *eloop_ctx, void *sock_ctx)
+static void rsn_preauth_receive(void *ctx, unsigned char *src_addr,
+				unsigned char *buf, size_t len)
 {
-	struct hostapd_data *hapd = eloop_ctx;
-	struct rsn_preauth_interface *piface = sock_ctx;
+	struct rsn_preauth_interface *piface = ctx;
+	struct hostapd_data *hapd = piface->hapd;
 	struct ieee802_1x_hdr *hdr;
-	char buf[1600];
-	int res;
 	struct sta_info *sta;
 	struct l2_ethhdr *ethhdr;
-	struct sockaddr_ll from;
-	socklen_t fromlen;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "RSN: receive pre-auth packet "
 		      "from interface '%s'\n", piface->ifname);
-	fromlen = sizeof(from);
-	res = recvfrom(sock, buf, sizeof(buf), 0,
-		       (struct sockaddr *) &from, &fromlen);
-	if (res < 0) {
-		perror("recvfrom[preauth]");
-		return;
-	}
-	if (res < sizeof(*ethhdr) + sizeof(*hdr)) {
+	if (len < sizeof(*ethhdr) + sizeof(*hdr)) {
 		HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "RSN: too short pre-auth "
-			      "packet (len=%d)\n", res);
+			      "packet (len=%lu)\n", (unsigned long) len);
 		return;
 	}
 
@@ -474,15 +457,13 @@ static void rsn_preauth_receive(int sock, void *eloop_ctx, void *sock_ctx)
 		return;
 	sta->preauth_iface = piface;
 	ieee802_1x_receive(hapd, ethhdr->h_source, (u8 *) (ethhdr + 1),
-			   res - sizeof(*ethhdr));
+			   len - sizeof(*ethhdr));
 }
 
 
 static int rsn_preauth_iface_add(struct hostapd_data *hapd, const char *ifname)
 {
 	struct rsn_preauth_interface *piface;
-	struct ifreq ifr;
-	struct sockaddr_ll ll;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "RSN pre-auth interface '%s'\n",
 		      ifname);
@@ -491,43 +472,26 @@ static int rsn_preauth_iface_add(struct hostapd_data *hapd, const char *ifname)
 	if (piface == NULL)
 		return -1;
 	memset(piface, 0, sizeof(*piface));
+	piface->hapd = hapd;
 
 	piface->ifname = strdup(ifname);
 	if (piface->ifname == NULL) {
 		goto fail1;
 	}
 
-	piface->fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_PREAUTH));
-	if (piface->fd < 0) {
-		perror("socket(PF_PACKET,SOCK_RAW,ETH_P_PREAUTH)");
+	piface->l2 = l2_packet_init(piface->ifname, NULL, ETH_P_PREAUTH,
+				    rsn_preauth_receive, piface);
+	if (piface->l2 == NULL) {
+		printf("Failed to open register layer 2 access to "
+		       "ETH_P_PREAUTH\n");
 		goto fail2;
 	}
+	l2_packet_set_rx_l2_hdr(piface->l2, 1);
 
-	memset(&ifr, 0, sizeof(ifr));
-	strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
-	if (ioctl(piface->fd, SIOCGIFINDEX, &ifr) < 0) {
-		perror("ioctl[SIOCGIFINDEX]");
-		goto fail3;
-	}
-	piface->ifindex = ifr.ifr_ifindex;
-
-	memset(&ll, 0, sizeof(ll));
-	ll.sll_family = PF_PACKET;
-	ll.sll_ifindex = ifr.ifr_ifindex;
-	ll.sll_protocol = htons(ETH_P_PREAUTH);
-	if (bind(piface->fd, (struct sockaddr *) &ll, sizeof(ll)) < 0) {
-		perror("bind[PF_PACKET,ETH_P_PREAUTH]");
-		goto fail3;
-	}
-
-	eloop_register_read_sock(piface->fd, rsn_preauth_receive, hapd,
-				 piface);
 	piface->next = hapd->preauth_iface;
 	hapd->preauth_iface = piface;
 	return 0;
 
-fail3:
-	close(piface->fd);
 fail2:
 	free(piface->ifname);
 fail1:
@@ -545,14 +509,48 @@ static void rsn_preauth_iface_deinit(struct hostapd_data *hapd)
 	while (piface) {
 		prev = piface;
 		piface = piface->next;
-		eloop_unregister_read_sock(prev->fd);
-		close(prev->fd);
+		l2_packet_deinit(prev->l2);
 		free(prev->ifname);
 		free(prev);
 	}
 }
 
 
+static int rsn_preauth_iface_init(struct hostapd_data *hapd)
+{
+	char *tmp, *start, *end;
+
+	if (hapd->conf->rsn_preauth_interfaces == NULL)
+		return 0;
+
+	tmp = strdup(hapd->conf->rsn_preauth_interfaces);
+	if (tmp == NULL)
+		return -1;
+	start = tmp;
+	for (;;) {
+		while (*start == ' ')
+			start++;
+		if (*start == '\0')
+			break;
+		end = strchr(start, ' ');
+		if (end)
+			*end = '\0';
+
+		if (rsn_preauth_iface_add(hapd, start)) {
+			rsn_preauth_iface_deinit(hapd);
+			return -1;
+		}
+
+		if (end)
+			start = end + 1;
+		else
+			break;
+	}
+	free(tmp);
+	return 0;
+}
+
+
 void rsn_preauth_finished(struct hostapd_data *hapd, struct sta_info *sta,
 			  int success)
 {
@@ -600,50 +598,46 @@ void rsn_preauth_send(struct hostapd_data *hapd, struct sta_info *sta,
 	ethhdr->h_proto = htons(ETH_P_PREAUTH);
 	memcpy(ethhdr + 1, buf, len);
 
-	if (send(piface->fd, ethhdr, sizeof(*ethhdr) + len, 0) < 0) {
-		perror("send[preauth]");
+	if (l2_packet_send(piface->l2, (u8 *) ethhdr, sizeof(*ethhdr) + len) <
+	    0) {
+		printf("Failed to send preauth packet using l2_packet_send\n");
 	}
 	free(ethhdr);
 }
 
+#else /* CONFIG_RSN_PREAUTH */
+
+static inline int rsn_preauth_iface_init(struct hostapd_data *hapd)
+{
+	return 0;
+}
+
+static inline void rsn_preauth_iface_deinit(struct hostapd_data *hapd)
+{
+}
+
+void rsn_preauth_finished(struct hostapd_data *hapd, struct sta_info *sta,
+			  int success)
+{
+}
+
+void rsn_preauth_send(struct hostapd_data *hapd, struct sta_info *sta,
+		      u8 *buf, size_t len)
+{
+}
+
+#endif /* CONFIG_RSN_PREAUTH */
+
 
 int wpa_init(struct hostapd_data *hapd)
 {
 	u8 rkey[32];
 	u8 buf[ETH_ALEN + 8];
-	struct timeval now;
-	u32 ntp_hi, ntp_lo, sec, usec;
 
-	if (hapd->conf->rsn_preauth_interfaces) {
-		char *tmp, *start, *end;
-		tmp = strdup(hapd->conf->rsn_preauth_interfaces);
-		if (tmp == NULL)
-			return -1;
-		start = tmp;
-		for (;;) {
-			while (*start == ' ')
-				start++;
-			if (*start == '\0')
-				break;
-			end = strchr(start, ' ');
-			if (end)
-				*end = '\0';
+	if (rsn_preauth_iface_init(hapd))
+		return -1;
 
-			if (rsn_preauth_iface_add(hapd, start)) {
-				rsn_preauth_iface_deinit(hapd);
-				return -1;
-			}
-
-			if (end)
-				start = end + 1;
-			else
-				break;
-		}
-		free(tmp);
-	}
-
-	if (hostap_ioctl_prism2param(hapd->driver.data,
-				     PRISM2_PARAM_PRIVACY_INVOKED, 1)) {
+	if (hostapd_set_privacy(hapd, 1)) {
 		printf("Could not set PrivacyInvoked for interface %s\n",
 		       hapd->conf->iface);
 		return -1;
@@ -654,8 +648,7 @@ int wpa_init(struct hostapd_data *hapd)
 		return -1;
 	}
 
-	if (hostapd_set_generic_elem(hapd->driver.data,
-				     hapd->wpa_ie, hapd->wpa_ie_len)) {
+	if (hostapd_set_generic_elem(hapd, hapd->wpa_ie, hapd->wpa_ie_len)) {
 		printf("Failed to configure WPA IE for the kernel driver.\n");
 		return -1;
 	}
@@ -683,16 +676,8 @@ int wpa_init(struct hostapd_data *hapd)
 	/* Counter = PRF-256(Random number, "Init Counter",
 	 *                   Local MAC Address || Time)
 	 */
-	gettimeofday(&now, NULL);
-	sec = now.tv_sec;
-	usec = now.tv_usec;
-#define JAN_1970 0x83aa7e80UL /* seconds from 1900 to 1970 */
-	ntp_hi = htonl(sec + JAN_1970);
-	/* approximation of 2^32/1000000 * usec */
-	ntp_lo = htonl(4295 * usec - (usec >> 5) - (usec >> 9));
 	memcpy(buf, hapd->own_addr, ETH_ALEN);
-	memcpy(buf + ETH_ALEN, &ntp_hi, 4);
-	memcpy(buf + ETH_ALEN + 4, &ntp_lo, 4);
+	hostapd_get_ntp_timestamp(buf + ETH_ALEN);
 	if (hostapd_get_rand(rkey, sizeof(rkey)) ||
 	    hostapd_get_rand(hapd->wpa_auth->GMK, WPA_GMK_LEN)) {
 		printf("Failed to get random data for WPA initialization.\n");
@@ -730,13 +715,12 @@ void wpa_deinit(struct hostapd_data *hapd)
 	eloop_cancel_timeout(wpa_rekey_gmk, hapd, NULL);
 	eloop_cancel_timeout(wpa_rekey_gtk, hapd, NULL);
 
-	if (hostap_ioctl_prism2param(hapd->driver.data,
-				     PRISM2_PARAM_PRIVACY_INVOKED, 0)) {
+	if (hostapd_set_privacy(hapd, 0)) {
 		printf("Could not disable PrivacyInvoked for interface %s\n",
 		       hapd->conf->iface);
 	}
 
-	if (hostapd_set_generic_elem(hapd->driver.data, "", 0)) {
+	if (hostapd_set_generic_elem(hapd, (u8 *) "", 0)) {
 		printf("Could not remove generic information element from "
 		       "interface %s\n", hapd->conf->iface);
 	}
@@ -804,13 +788,17 @@ static int rsn_key_mgmt_to_bitfield(u8 *s)
 }
 
 
-static void rsn_pmkid(u8 *pmk, u8 *aa, u8 *spa, u8 *pmkid)
+static void rsn_pmkid(const u8 *pmk, const u8 *aa, const u8 *spa, u8 *pmkid)
 {
 	char *title = "PMK Name";
-	unsigned char *addr[3] = { title, aa, spa };
-	unsigned int len[3] = { 8, ETH_ALEN, ETH_ALEN };
+	const u8 *addr[3];
+	const size_t len[3] = { 8, ETH_ALEN, ETH_ALEN };
 	unsigned char hash[SHA1_MAC_LEN];
 
+	addr[0] = (u8 *) title;
+	addr[1] = aa;
+	addr[2] = spa;
+
 	hmac_sha1_vector(pmk, PMK_LEN, 3, addr, len, hash);
 	memcpy(pmkid, hash, PMKID_LEN);
 }
@@ -1166,8 +1154,8 @@ static int wpa_parse_wpa_ie_rsn(u8 *rsn_ie, size_t rsn_ie_len,
 		left -= 2;
 		if (left < data->num_pmkid * PMKID_LEN) {
 			printf("RSN: too short RSN IE for PMKIDs "
-			       "(num=%d, left=%d)\n",
-			       data->num_pmkid, left);
+			       "(num=%lu, left=%d)\n",
+			       (unsigned long) data->num_pmkid, left);
 			return -9;
 		}
 		data->pmkid = pos;
@@ -1188,11 +1176,85 @@ int wpa_validate_wpa_ie(struct hostapd_data *hapd, struct sta_info *sta,
 {
 	struct wpa_ie_data data;
 	int ciphers, key_mgmt, res, i;
+	const u8 *selector;
 
-	if (version == HOSTAPD_WPA_VERSION_WPA2)
+	if (version == HOSTAPD_WPA_VERSION_WPA2) {
 		res = wpa_parse_wpa_ie_rsn(wpa_ie, wpa_ie_len, &data);
-	else
+
+		selector = RSN_AUTH_KEY_MGMT_UNSPEC_802_1X;
+		if (data.key_mgmt & WPA_KEY_MGMT_IEEE8021X)
+			selector = RSN_AUTH_KEY_MGMT_UNSPEC_802_1X;
+		else if (data.key_mgmt & WPA_KEY_MGMT_PSK)
+			selector = RSN_AUTH_KEY_MGMT_PSK_OVER_802_1X;
+		memcpy(hapd->wpa_auth->dot11RSNAAuthenticationSuiteSelected,
+		       selector, RSN_SELECTOR_LEN);
+
+		selector = RSN_CIPHER_SUITE_CCMP;
+		if (data.pairwise_cipher & WPA_CIPHER_CCMP)
+			selector = RSN_CIPHER_SUITE_CCMP;
+		else if (data.pairwise_cipher & WPA_CIPHER_TKIP)
+			selector = RSN_CIPHER_SUITE_TKIP;
+		else if (data.pairwise_cipher & WPA_CIPHER_WEP104)
+			selector = RSN_CIPHER_SUITE_WEP104;
+		else if (data.pairwise_cipher & WPA_CIPHER_WEP40)
+			selector = RSN_CIPHER_SUITE_WEP40;
+		else if (data.pairwise_cipher & WPA_CIPHER_NONE)
+			selector = RSN_CIPHER_SUITE_NONE;
+		memcpy(hapd->wpa_auth->dot11RSNAPairwiseCipherSelected,
+		       selector, RSN_SELECTOR_LEN);
+
+		selector = RSN_CIPHER_SUITE_CCMP;
+		if (data.group_cipher & WPA_CIPHER_CCMP)
+			selector = RSN_CIPHER_SUITE_CCMP;
+		else if (data.group_cipher & WPA_CIPHER_TKIP)
+			selector = RSN_CIPHER_SUITE_TKIP;
+		else if (data.group_cipher & WPA_CIPHER_WEP104)
+			selector = RSN_CIPHER_SUITE_WEP104;
+		else if (data.group_cipher & WPA_CIPHER_WEP40)
+			selector = RSN_CIPHER_SUITE_WEP40;
+		else if (data.group_cipher & WPA_CIPHER_NONE)
+			selector = RSN_CIPHER_SUITE_NONE;
+		memcpy(hapd->wpa_auth->dot11RSNAGroupCipherSelected,
+		       selector, RSN_SELECTOR_LEN);
+	} else {
 		res = wpa_parse_wpa_ie_wpa(wpa_ie, wpa_ie_len, &data);
+
+		selector = WPA_AUTH_KEY_MGMT_UNSPEC_802_1X;
+		if (data.key_mgmt & WPA_KEY_MGMT_IEEE8021X)
+			selector = WPA_AUTH_KEY_MGMT_UNSPEC_802_1X;
+		else if (data.key_mgmt & WPA_KEY_MGMT_PSK)
+			selector = WPA_AUTH_KEY_MGMT_PSK_OVER_802_1X;
+		memcpy(hapd->wpa_auth->dot11RSNAAuthenticationSuiteSelected,
+		       selector, WPA_SELECTOR_LEN);
+
+		selector = WPA_CIPHER_SUITE_TKIP;
+		if (data.pairwise_cipher & WPA_CIPHER_CCMP)
+			selector = WPA_CIPHER_SUITE_CCMP;
+		else if (data.pairwise_cipher & WPA_CIPHER_TKIP)
+			selector = WPA_CIPHER_SUITE_TKIP;
+		else if (data.pairwise_cipher & WPA_CIPHER_WEP104)
+			selector = WPA_CIPHER_SUITE_WEP104;
+		else if (data.pairwise_cipher & WPA_CIPHER_WEP40)
+			selector = WPA_CIPHER_SUITE_WEP40;
+		else if (data.pairwise_cipher & WPA_CIPHER_NONE)
+			selector = WPA_CIPHER_SUITE_NONE;
+		memcpy(hapd->wpa_auth->dot11RSNAPairwiseCipherSelected,
+		       selector, WPA_SELECTOR_LEN);
+
+		selector = WPA_CIPHER_SUITE_TKIP;
+		if (data.group_cipher & WPA_CIPHER_CCMP)
+			selector = WPA_CIPHER_SUITE_CCMP;
+		else if (data.group_cipher & WPA_CIPHER_TKIP)
+			selector = WPA_CIPHER_SUITE_TKIP;
+		else if (data.group_cipher & WPA_CIPHER_WEP104)
+			selector = WPA_CIPHER_SUITE_WEP104;
+		else if (data.group_cipher & WPA_CIPHER_WEP40)
+			selector = WPA_CIPHER_SUITE_WEP40;
+		else if (data.group_cipher & WPA_CIPHER_NONE)
+			selector = WPA_CIPHER_SUITE_NONE;
+		memcpy(hapd->wpa_auth->dot11RSNAGroupCipherSelected,
+		       selector, WPA_SELECTOR_LEN);
+	}
 	if (res) {
 		printf("Failed to parse WPA/RSN IE from " MACSTR " (res=%d)\n",
 		       MAC2STR(sta->addr), res);
@@ -1246,6 +1308,10 @@ int wpa_validate_wpa_ie(struct hostapd_data *hapd, struct sta_info *sta,
 			hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_WPA,
 				       HOSTAPD_LEVEL_DEBUG,
 				       "PMKID found from PMKSA cache");
+			if (hapd->wpa_auth) {
+				memcpy(hapd->wpa_auth->dot11RSNAPMKIDUsed,
+				       sta->pmksa->pmkid, PMKID_LEN);
+			}
 			break;
 		}
 	}
@@ -1295,6 +1361,15 @@ void wpa_free_station(struct sta_info *sta)
 	if (sm == NULL)
 		return;
 
+	if (sm->hapd->conf->wpa_strict_rekey && sm->has_GTK) {
+		hostapd_logger(sm->hapd, sta->addr, HOSTAPD_MODULE_WPA,
+			       HOSTAPD_LEVEL_DEBUG, "strict rekeying - force "
+			       "GTK rekey since STA is leaving");
+		eloop_cancel_timeout(wpa_rekey_gtk, sm->hapd, NULL);
+		eloop_register_timeout(0, 500000, wpa_rekey_gtk, sm->hapd,
+				       NULL);
+	}
+
 	eloop_cancel_timeout(wpa_send_eapol_timeout, sm->hapd, sta);
 	eloop_cancel_timeout(wpa_sm_call_step, sm->hapd, sta->wpa_sm);
 	free(sm->last_rx_eapol_key);
@@ -1401,7 +1476,7 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 		}
 		if (sta->wpa_ie == NULL ||
 		    sta->wpa_ie_len != key_data_length ||
-		    memcmp(sta->wpa_ie, key->key_data, key_data_length) != 0) {
+		    memcmp(sta->wpa_ie, key + 1, key_data_length) != 0) {
 			hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_WPA,
 				       HOSTAPD_LEVEL_INFO,
 				       "WPA IE from (Re)AssocReq did not match"
@@ -1413,7 +1488,7 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 							sta->wpa_ie_len);
 				}
 				hostapd_hexdump("WPA IE in msg 2/4",
-						key->key_data,
+						(u8 *) (key + 1),
 						key_data_length);
 			}
 			/* MLME-DEAUTHENTICATE.request */
@@ -1474,6 +1549,8 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 			return;
 		}
 		sm->MICVerified = TRUE;
+		eloop_cancel_timeout(wpa_send_eapol_timeout, sta->wpa_sm->hapd,
+				     sta);
 	}
 
 	if (key_info & WPA_KEY_INFO_REQUEST) {
@@ -1496,6 +1573,7 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 				       "received EAPOL-Key Error Request "
 				       "(STA detected Michael MIC failure)");
 			ieee80211_michael_mic_failure(hapd, sta->addr, 0);
+			sta->dot11RSNAStatsTKIPRemoteMICFailures++;
 			hapd->wpa_auth->dot11RSNAStatsTKIPRemoteMICFailures++;
 			/* Error report is not a request for a new key
 			 * handshake, but since Authenticator may do it, let's
@@ -1508,6 +1586,10 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 				       "4-Way Handshake");
 			wpa_request_new_ptk(hapd, sta);
 		} else {
+			/* TODO: this could also be a request for STAKey
+			 * if Key Data fields contains peer MAC address KDE.
+			 * STAKey request should have 0xdd <len> 00-0F-AC:2 in
+			 * the beginning of Key Data */
 			hostapd_logger(hapd, sta->addr, HOSTAPD_MODULE_WPA,
 				       HOSTAPD_LEVEL_INFO,
 				       "received EAPOL-Key Request for GTK "
@@ -1532,14 +1614,14 @@ void wpa_receive(struct hostapd_data *hapd, struct sta_info *sta,
 	sm->EAPOLKeyPairwise = !!(key_info & WPA_KEY_INFO_KEY_TYPE);
 	sm->EAPOLKeyRequest = !!(key_info & WPA_KEY_INFO_REQUEST);
 	memcpy(sm->SNonce, key->key_nonce, WPA_NONCE_LEN);
-	eloop_cancel_timeout(wpa_send_eapol_timeout, sta->wpa_sm->hapd, sta);
 	wpa_sm_step(sm);
 }
 
 
-static void wpa_pmk_to_ptk(struct hostapd_data *hapd, u8 *pmk,
-			   u8 *addr1, u8 *addr2,
-			   u8 *nonce1, u8 *nonce2, u8 *ptk, size_t ptk_len)
+static void wpa_pmk_to_ptk(struct hostapd_data *hapd, const u8 *pmk,
+			   const u8 *addr1, const u8 *addr2,
+			   const u8 *nonce1, const u8 *nonce2,
+			   u8 *ptk, size_t ptk_len)
 {
 	u8 data[2 * ETH_ALEN + 2 * WPA_NONCE_LEN];
 
@@ -1738,7 +1820,7 @@ static void wpa_send_eapol(struct hostapd_data *hapd, struct sta_info *sta,
 		memcpy(key->key_rsc, key_rsc, WPA_KEY_RSC_LEN);
 
 	if (ie && !gtk) {
-		memcpy(key->key_data, ie, ie_len);
+		memcpy(key + 1, ie, ie_len);
 		key->key_data_length = htons(ie_len);
 	} else if (gtk) {
 		buf = malloc(key_data_len);
@@ -1771,7 +1853,7 @@ static void wpa_send_eapol(struct hostapd_data *hapd, struct sta_info *sta,
 		}
 		if (key_info & WPA_KEY_INFO_TYPE_HMAC_SHA1_AES) {
 			aes_wrap(sm->PTK.encr_key, (key_data_len - 8) / 8, buf,
-				 key->key_data);
+				 (u8 *) (key + 1));
 			key->key_data_length = htons(key_data_len);
 		} else {
 			u8 ek[32];
@@ -1781,8 +1863,8 @@ static void wpa_send_eapol(struct hostapd_data *hapd, struct sta_info *sta,
 			inc_byte_array(hapd->wpa_auth->Counter, WPA_NONCE_LEN);
 			memcpy(ek, key->key_iv, 16);
 			memcpy(ek + 16, sm->PTK.encr_key, 16);
-			memcpy(key->key_data, buf, key_data_len);
-			rc4_skip(ek, 32, 256, key->key_data, key_data_len);
+			memcpy(key + 1, buf, key_data_len);
+			rc4_skip(ek, 32, 256, (u8 *) (key + 1), key_data_len);
 			key->key_data_length = htons(key_data_len);
 		}
 		free(buf);
@@ -1801,10 +1883,9 @@ static void wpa_send_eapol(struct hostapd_data *hapd, struct sta_info *sta,
 				       key->key_mic);
 	}
 
-	if (hapd->driver.send_eapol) {
-		hapd->driver.send_eapol(hapd->driver.data, sta->addr,
-					(u8 *) hdr, len, sm->pairwise_set);
-	}
+	if (sta->eapol_sm)
+		sta->eapol_sm->dot1xAuthEapolFramesTx++;
+	hostapd_send_eapol(hapd, sta->addr, (u8 *) hdr, len, sm->pairwise_set);
 	free(hdr);
 
 	timeout_ms = pairwise ? dot11RSNAConfigPairwiseUpdateTimeOut :
@@ -1819,7 +1900,7 @@ static int wpa_verify_key_mic(struct wpa_ptk *PTK, u8 *data, size_t data_len)
 	struct ieee802_1x_hdr *hdr;
 	struct wpa_eapol_key *key;
 	u16 key_info;
-	int type;
+	int type, ret = 0;
 	u8 mic[16];
 
 	if (data_len < sizeof(*hdr) + sizeof(*key))
@@ -1834,8 +1915,9 @@ static int wpa_verify_key_mic(struct wpa_ptk *PTK, u8 *data, size_t data_len)
 	if (wpa_calc_eapol_key_mic(key_info & WPA_KEY_INFO_TYPE_MASK,
 				   PTK->mic_key, data, data_len, key->key_mic)
 	    || memcmp(mic, key->key_mic, 16) != 0)
-		return -1;
-	return 0;
+		ret = -1;
+	memcpy(key->key_mic, mic, 16);
+	return ret;
 }
 
 
@@ -1882,8 +1964,8 @@ void wpa_sm_event(struct hostapd_data *hapd, struct sta_info *sta,
 
 	if (event != WPA_REAUTH_EAPOL) {
 		sm->pairwise_set = FALSE;
-		hostapd_set_encryption(sm->hapd->driver.data, "none",
-				       sm->sta->addr, 0, "", 0);
+		hostapd_set_encryption(sm->hapd, "none", sm->sta->addr, 0,
+				       (u8 *) "", 0);
 	}
 
 	wpa_sm_step(sm);
@@ -1948,8 +2030,8 @@ SM_STATE(WPA_PTK, INITIALIZE)
 		sm->Pair = TRUE;
 	}
 	ieee802_1x_notify_port_enabled(sm->sta->eapol_sm, 0);
-	hostapd_set_encryption(sm->hapd->driver.data, "none", 
-			       sm->sta->addr, 0, "", 0);
+	hostapd_set_encryption(sm->hapd, "none", sm->sta->addr, 0, (u8 *) "",
+			       0);
 	sm->pairwise_set = FALSE;
 	sm->PTK_valid = FALSE;
 	memset(&sm->PTK, 0, sizeof(sm->PTK));
@@ -2010,11 +2092,16 @@ SM_STATE(WPA_PTK, INITPMK)
 	size_t len;
 	SM_ENTRY(WPA_PTK, INITPMK, wpa_ptk);
 	if (sm->sta->pmksa) {
+		wpa_printf(MSG_DEBUG, "WPA: PMK from PMKSA cache");
 		memcpy(sm->PMK, sm->sta->pmksa->pmk, WPA_PMK_LEN);
 	} else if ((key = ieee802_1x_get_key_crypt(sm->sta->eapol_sm, &len))) {
+		wpa_printf(MSG_DEBUG, "WPA: PMK from EAPOL state machine "
+			   "(len=%lu)", (unsigned long) len);
 		if (len > WPA_PMK_LEN)
 			len = WPA_PMK_LEN;
 		memcpy(sm->PMK, key, len);
+	} else {
+		wpa_printf(MSG_DEBUG, "WPA: Could not get PMK");
 	}
 	sm->sta->req_replay_counter_used = 0;
 	/* IEEE 802.11i/D9.0 does not set keyRun to FALSE, but not doing this
@@ -2032,22 +2119,29 @@ SM_STATE(WPA_PTK, INITPMK)
 
 SM_STATE(WPA_PTK, INITPSK)
 {
+	const u8 *psk;
 	SM_ENTRY(WPA_PTK, INITPSK, wpa_ptk);
-	memcpy(sm->PMK, sm->hapd->conf->wpa_psk, WPA_PMK_LEN);
+	psk = hostapd_get_psk(sm->hapd->conf, sm->sta->addr, NULL);
+	if (psk)
+		memcpy(sm->PMK, psk, WPA_PMK_LEN);
 	sm->sta->req_replay_counter_used = 0;
 }
 
 
 SM_STATE(WPA_PTK, PTKSTART)
 {
-	u8 pmkid[2 + RSN_SELECTOR_LEN + PMKID_LEN];
+	u8 *pmkid = NULL;
+	size_t pmkid_len = 0;
+
 	SM_ENTRY(WPA_PTK, PTKSTART, wpa_ptk);
 	sm->PTKRequest = FALSE;
 	sm->TimeoutEvt = FALSE;
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
 		       HOSTAPD_LEVEL_DEBUG,
 		       "sending 1/4 msg of 4-Way Handshake");
-	if (sm->sta->pmksa) {
+	if (sm->sta->pmksa &&
+	    (pmkid = malloc(2 + RSN_SELECTOR_LEN + PMKID_LEN))) {
+		pmkid_len = 2 + RSN_SELECTOR_LEN + PMKID_LEN;
 		pmkid[0] = WLAN_EID_GENERIC;
 		pmkid[1] = RSN_SELECTOR_LEN + PMKID_LEN;
 		memcpy(&pmkid[2], RSN_KEY_DATA_PMKID, RSN_SELECTOR_LEN);
@@ -2055,9 +2149,8 @@ SM_STATE(WPA_PTK, PTKSTART)
 		       PMKID_LEN);
 	}
 	wpa_send_eapol(sm->hapd, sm->sta, 0, 0, 1, 0, 1, NULL, sm->ANonce,
-		       sm->sta->pmksa ? pmkid : NULL,
-		       sm->sta->pmksa ? 2 + RSN_SELECTOR_LEN + PMKID_LEN : 0,
-		       NULL, 0, 0);
+		       pmkid, pmkid_len, NULL, 0, 0);
+	free(pmkid);
 	sm->TimeoutCtr++;
 }
 
@@ -2065,19 +2158,54 @@ SM_STATE(WPA_PTK, PTKSTART)
 SM_STATE(WPA_PTK, PTKCALCNEGOTIATING)
 {
 	struct wpa_ptk PTK;
+	int ok = 0;
+	const u8 *pmk = NULL;
 
 	SM_ENTRY(WPA_PTK, PTKCALCNEGOTIATING, wpa_ptk);
 	sm->EAPOLKeyReceived = FALSE;
-	wpa_pmk_to_ptk(sm->hapd, sm->PMK, sm->hapd->own_addr, sm->sta->addr,
-		       sm->ANonce, sm->SNonce,
-		       (u8 *) &PTK, sizeof(PTK));
-	if (wpa_verify_key_mic(&PTK, sm->last_rx_eapol_key,
-			       sm->last_rx_eapol_key_len)) {
+
+	/* WPA with IEEE 802.1X: use the derived PMK from EAP
+	 * WPA-PSK: iterate through possible PSKs and select the one matching
+	 * the packet */
+	for (;;) {
+		if (sm->sta->wpa_key_mgmt == WPA_KEY_MGMT_PSK) {
+			pmk = hostapd_get_psk(sm->hapd->conf, sm->sta->addr,
+					      pmk);
+			if (pmk == NULL)
+				break;
+		} else
+			pmk = sm->PMK;
+
+		wpa_pmk_to_ptk(sm->hapd, pmk, sm->hapd->own_addr,
+			       sm->sta->addr, sm->ANonce, sm->SNonce,
+			       (u8 *) &PTK, sizeof(PTK));
+
+		if (wpa_verify_key_mic(&PTK, sm->last_rx_eapol_key,
+				       sm->last_rx_eapol_key_len) == 0) {
+			ok = 1;
+			break;
+		}
+
+		if (sm->sta->wpa_key_mgmt != WPA_KEY_MGMT_PSK)
+			break;
+	}
+
+	if (!ok) {
 		hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
 			       HOSTAPD_LEVEL_DEBUG, "invalid MIC in msg 2/4 "
 			       "of 4-Way Handshake");
 		return;
 	}
+
+	eloop_cancel_timeout(wpa_send_eapol_timeout, sm->hapd, sm->sta);
+
+	if (sm->sta->wpa_key_mgmt == WPA_KEY_MGMT_PSK) {
+		/* PSK may have changed from the previous choice, so update
+		 * state machine data based on whatever PSK was selected here.
+		 */
+		memcpy(sm->PMK, pmk, WPA_PMK_LEN);
+	}
+
 	sm->MICVerified = TRUE;
 
 	memcpy(&sm->PTK, &PTK, sizeof(PTK));
@@ -2104,7 +2232,7 @@ SM_STATE(WPA_PTK, PTKINITNEGOTIATING)
 	/* Send EAPOL(1, 1, 1, Pair, P, RSC, ANonce, MIC(PTK), RSNIE, GTK[GN])
 	 */
 	memset(rsc, 0, WPA_KEY_RSC_LEN);
-	hostapd_get_seqnum(sm->hapd->driver.data, NULL, gsm->GN, rsc);
+	hostapd_get_seqnum(sm->hapd, NULL, gsm->GN, rsc);
 	wpa_ie = sm->hapd->wpa_ie;
 	wpa_ie_len = sm->hapd->wpa_ie_len;
 	if (sm->sta->wpa == WPA_VERSION_WPA &&
@@ -2140,8 +2268,7 @@ SM_STATE(WPA_PTK, PTKINITDONE)
 			alg = "CCMP";
 			klen = 16;
 		}
-		if (hostapd_set_encryption(sm->hapd->driver.data, 
-					   alg, sm->sta->addr, 0,
+		if (hostapd_set_encryption(sm->hapd, alg, sm->sta->addr, 0,
 					   sm->PTK.tk1, klen)) {
 			wpa_sta_disconnect(sm->hapd, sm->sta);
 			return;
@@ -2167,6 +2294,8 @@ SM_STATE(WPA_PTK, PTKINITDONE)
 	}
 	if (sm->sta->wpa == WPA_VERSION_WPA)
 		sm->PInitAKeys = TRUE;
+	else
+		sm->has_GTK = TRUE;
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
 		       HOSTAPD_LEVEL_INFO, "pairwise key handshake completed "
 		       "(%s)",
@@ -2176,6 +2305,8 @@ SM_STATE(WPA_PTK, PTKINITDONE)
 
 SM_STEP(WPA_PTK)
 {
+	struct wpa_authenticator *wpa_auth = sm->hapd->wpa_auth;
+
 	if (sm->Init)
 		SM_ENTER(WPA_PTK, INITIALIZE);
 	else if (sm->Disconnect
@@ -2212,20 +2343,31 @@ SM_STEP(WPA_PTK)
 	case WPA_PTK_INITPMK:
 		if (sm->sta->eapol_sm && sm->sta->eapol_sm->keyAvailable)
 			SM_ENTER(WPA_PTK, PTKSTART);
-		else
+		else {
+			wpa_auth->dot11RSNA4WayHandshakeFailures++;
 			SM_ENTER(WPA_PTK, DISCONNECT);
+		}
 		break;
 	case WPA_PTK_INITPSK:
-		if (sm->hapd->conf->wpa_psk /* 802.1X::keyAvailable */)
+		if (hostapd_get_psk(sm->hapd->conf, sm->sta->addr, NULL))
 			SM_ENTER(WPA_PTK, PTKSTART);
+		else {
+			hostapd_logger(sm->hapd, sm->sta->addr,
+				       HOSTAPD_MODULE_WPA,
+				       HOSTAPD_LEVEL_INFO,
+				       "no PSK configured for the STA");
+			wpa_auth->dot11RSNA4WayHandshakeFailures++;
+			SM_ENTER(WPA_PTK, DISCONNECT);
+		}
 		break;
 	case WPA_PTK_PTKSTART:
 		if (sm->EAPOLKeyReceived && !sm->EAPOLKeyRequest &&
 		    sm->EAPOLKeyPairwise)
 			SM_ENTER(WPA_PTK, PTKCALCNEGOTIATING);
-		else if (sm->TimeoutCtr > dot11RSNAConfigPairwiseUpdateCount)
+		else if (sm->TimeoutCtr > dot11RSNAConfigPairwiseUpdateCount) {
+			wpa_auth->dot11RSNA4WayHandshakeFailures++;
 			SM_ENTER(WPA_PTK, DISCONNECT);
-		else if (sm->TimeoutEvt)
+		} else if (sm->TimeoutEvt)
 			SM_ENTER(WPA_PTK, PTKSTART);
 		break;
 	case WPA_PTK_PTKCALCNEGOTIATING:
@@ -2244,9 +2386,10 @@ SM_STEP(WPA_PTK)
 		if (sm->EAPOLKeyReceived && !sm->EAPOLKeyRequest &&
 		    sm->EAPOLKeyPairwise && sm->MICVerified)
 			SM_ENTER(WPA_PTK, PTKINITDONE);
-		else if (sm->TimeoutCtr > dot11RSNAConfigPairwiseUpdateCount)
+		else if (sm->TimeoutCtr > dot11RSNAConfigPairwiseUpdateCount) {
+			wpa_auth->dot11RSNA4WayHandshakeFailures++;
 			SM_ENTER(WPA_PTK, DISCONNECT);
-		else if (sm->TimeoutEvt)
+		} else if (sm->TimeoutEvt)
 			SM_ENTER(WPA_PTK, PTKINITNEGOTIATING);
 		break;
 	case WPA_PTK_PTKINITDONE:
@@ -2279,7 +2422,7 @@ SM_STATE(WPA_PTK_GROUP, REKEYNEGOTIATING)
 	/* Send EAPOL(1, 1, 1, !Pair, G, RSC, GNonce, MIC(PTK), GTK[GN]) */
 	memset(rsc, 0, WPA_KEY_RSC_LEN);
 	if (gsm->wpa_group_state == WPA_GROUP_SETKEYSDONE)
-		hostapd_get_seqnum(sm->hapd->driver.data, NULL, gsm->GN, rsc);
+		hostapd_get_seqnum(sm->hapd, NULL, gsm->GN, rsc);
 	hostapd_logger(sm->hapd, sm->sta->addr, HOSTAPD_MODULE_WPA,
 		       HOSTAPD_LEVEL_DEBUG,
 		       "sending 1/2 msg of Group Key Handshake");
@@ -2302,6 +2445,7 @@ SM_STATE(WPA_PTK_GROUP, REKEYESTABLISHED)
 		       HOSTAPD_LEVEL_INFO, "group key handshake completed "
 		       "(%s)",
 		       sm->sta->wpa == WPA_VERSION_WPA ? "WPA" : "RSN");
+	sm->has_GTK = TRUE;
 }
 
 
@@ -2404,10 +2548,8 @@ static void wpa_group_setkeysdone(struct hostapd_data *hapd)
 		      "entering state SETKEYSDONE\n");
 	sm->changed = TRUE;
 	sm->wpa_group_state = WPA_GROUP_SETKEYSDONE;
-	hostapd_set_encryption(hapd->driver.data,
-			       wpa_alg_txt(hapd->conf->wpa_group),
-			       NULL, sm->GN, sm->GTK[sm->GN - 1],
-			       sm->GTK_len);
+	hostapd_set_encryption(hapd, wpa_alg_txt(hapd->conf->wpa_group),
+			       NULL, sm->GN, sm->GTK[sm->GN - 1], sm->GTK_len);
 }
 
 
@@ -2500,3 +2642,179 @@ void wpa_gtk_rekey(struct hostapd_data *hapd)
 			       sm->GTK[sm->GN - 1], sm->GTK_len);
 	}
 }
+
+
+static const char * wpa_bool_txt(int bool)
+{
+	return bool ? "TRUE" : "FALSE";
+}
+
+
+static int wpa_cipher_bits(int cipher)
+{
+	switch (cipher) {
+	case WPA_CIPHER_CCMP:
+		return 128;
+	case WPA_CIPHER_TKIP:
+		return 256;
+	case WPA_CIPHER_WEP104:
+		return 104;
+	case WPA_CIPHER_WEP40:
+		return 40;
+	default:
+		return 0;
+	}
+}
+
+
+#define RSN_SUITE "%02x-%02x-%02x-%d"
+#define RSN_SUITE_ARG(s) (s)[0], (s)[1], (s)[2], (s)[3]
+
+int wpa_get_mib(struct hostapd_data *hapd, char *buf, size_t buflen)
+{
+	int len = 0, i;
+	char pmkid_txt[PMKID_LEN * 2 + 1], *pos;
+
+	len += snprintf(buf + len, buflen - len,
+			"dot11RSNAOptionImplemented=TRUE\n"
+			"dot11RSNAPreauthenticationImplemented=TRUE\n"
+			"dot11RSNAEnabled=%s\n"
+			"dot11RSNAPreauthenticationEnabled=%s\n",
+			wpa_bool_txt(hapd->conf->wpa &
+				     HOSTAPD_WPA_VERSION_WPA2),
+			wpa_bool_txt(hapd->conf->rsn_preauth));
+
+	if (hapd->wpa_auth == NULL)
+		return len;
+
+	pos = pmkid_txt;
+	for (i = 0; i < PMKID_LEN; i++) {
+		pos += sprintf(pos, "%02x",
+			       hapd->wpa_auth->dot11RSNAPMKIDUsed[i]);
+	}
+
+	len += snprintf(buf + len, buflen - len,
+			"dot11RSNAConfigVersion=%u\n"
+			"dot11RSNAConfigPairwiseKeysSupported=9999\n"
+			/* FIX: dot11RSNAConfigGroupCipher */
+			/* FIX: dot11RSNAConfigGroupRekeyMethod */
+			/* FIX: dot11RSNAConfigGroupRekeyTime */
+			/* FIX: dot11RSNAConfigGroupRekeyPackets */
+			"dot11RSNAConfigGroupRekeyStrict=%u\n"
+			"dot11RSNAConfigGroupUpdateCount=%u\n"
+			"dot11RSNAConfigPairwiseUpdateCount=%u\n"
+			"dot11RSNAConfigGroupCipherSize=%u\n"
+			"dot11RSNAConfigPMKLifetime=%u\n"
+			"dot11RSNAConfigPMKReauthThreshold=%u\n"
+			"dot11RSNAConfigNumberOfPTKSAReplayCounters=0\n"
+			"dot11RSNAConfigSATimeout=%u\n"
+			"dot11RSNAAuthenticationSuiteSelected=" RSN_SUITE "\n"
+			"dot11RSNAPairwiseCipherSelected=" RSN_SUITE "\n"
+			"dot11RSNAGroupCipherSelected=" RSN_SUITE "\n"
+			"dot11RSNAPMKIDUsed=%s\n"
+			"dot11RSNAAuthenticationSuiteRequested=" RSN_SUITE "\n"
+			"dot11RSNAPairwiseCipherRequested=" RSN_SUITE "\n"
+			"dot11RSNAGroupCipherRequested=" RSN_SUITE "\n"
+			"dot11RSNATKIPCounterMeasuresInvoked=%u\n"
+			"dot11RSNA4WayHandshakeFailures=%u\n"
+			"dot11RSNAConfigNumberOfGTKSAReplayCounters=0\n",
+			RSN_VERSION,
+			!!hapd->conf->wpa_strict_rekey,
+			dot11RSNAConfigGroupUpdateCount,
+			dot11RSNAConfigPairwiseUpdateCount,
+			wpa_cipher_bits(hapd->conf->wpa_group),
+			dot11RSNAConfigPMKLifetime,
+			dot11RSNAConfigPMKReauthThreshold,
+			dot11RSNAConfigSATimeout,
+			RSN_SUITE_ARG(hapd->wpa_auth->
+				      dot11RSNAAuthenticationSuiteSelected),
+			RSN_SUITE_ARG(hapd->wpa_auth->
+				      dot11RSNAPairwiseCipherSelected),
+			RSN_SUITE_ARG(hapd->wpa_auth->
+				      dot11RSNAGroupCipherSelected),
+			pmkid_txt,
+			RSN_SUITE_ARG(hapd->wpa_auth->
+				      dot11RSNAAuthenticationSuiteRequested),
+			RSN_SUITE_ARG(hapd->wpa_auth->
+				      dot11RSNAPairwiseCipherRequested),
+			RSN_SUITE_ARG(hapd->wpa_auth->
+				      dot11RSNAGroupCipherRequested),
+			hapd->wpa_auth->dot11RSNATKIPCounterMeasuresInvoked,
+			hapd->wpa_auth->dot11RSNA4WayHandshakeFailures);
+
+	/* TODO: dot11RSNAConfigPairwiseCiphersTable */
+	/* TODO: dot11RSNAConfigAuthenticationSuitesTable */
+
+	/* Private MIB */
+	len += snprintf(buf + len, buflen - len,
+			"hostapdWPAGroupState=%d\n",
+			hapd->wpa_auth->wpa_group_state);
+
+	return len;
+}
+
+
+int wpa_get_mib_sta(struct hostapd_data *hapd, struct sta_info *sta,
+		    char *buf, size_t buflen)
+{
+	int len = 0;
+	u8 not_used[4] = { 0, 0, 0, 0 };
+	const u8 *pairwise = not_used;
+
+	/* TODO: FF-FF-FF-FF-FF-FF entry for broadcast/multicast stats */
+
+	/* dot11RSNAStatsEntry */
+
+	if (sta->wpa == WPA_VERSION_WPA) {
+		if (sta->pairwise == WPA_CIPHER_CCMP)
+			pairwise = WPA_CIPHER_SUITE_CCMP;
+		else if (sta->pairwise == WPA_CIPHER_TKIP)
+			pairwise = WPA_CIPHER_SUITE_TKIP;
+		else if (sta->pairwise == WPA_CIPHER_WEP104)
+			pairwise = WPA_CIPHER_SUITE_WEP104;
+		else if (sta->pairwise == WPA_CIPHER_WEP40)
+			pairwise = WPA_CIPHER_SUITE_WEP40;
+		else if (sta->pairwise == WPA_CIPHER_NONE)
+			pairwise = WPA_CIPHER_SUITE_NONE;
+	} else if (sta->wpa == WPA_VERSION_WPA2) {
+		if (sta->pairwise == WPA_CIPHER_CCMP)
+			pairwise = RSN_CIPHER_SUITE_CCMP;
+		else if (sta->pairwise == WPA_CIPHER_TKIP)
+			pairwise = RSN_CIPHER_SUITE_TKIP;
+		else if (sta->pairwise == WPA_CIPHER_WEP104)
+			pairwise = RSN_CIPHER_SUITE_WEP104;
+		else if (sta->pairwise == WPA_CIPHER_WEP40)
+			pairwise = RSN_CIPHER_SUITE_WEP40;
+		else if (sta->pairwise == WPA_CIPHER_NONE)
+			pairwise = RSN_CIPHER_SUITE_NONE;
+	} else
+		return 0;
+
+	len += snprintf(buf + len, buflen - len,
+			/* TODO: dot11RSNAStatsIndex */
+			"dot11RSNAStatsSTAAddress=" MACSTR "\n"
+			"dot11RSNAStatsVersion=1\n"
+			"dot11RSNAStatsSelectedPairwiseCipher=" RSN_SUITE "\n"
+			/* TODO: dot11RSNAStatsTKIPICVErrors */
+			"dot11RSNAStatsTKIPLocalMICFailures=%u\n"
+			"dot11RSNAStatsTKIPRemoveMICFailures=%u\n"
+			/* TODO: dot11RSNAStatsCCMPReplays */
+			/* TODO: dot11RSNAStatsCCMPDecryptErrors */
+			/* TODO: dot11RSNAStatsTKIPReplays */,
+			MAC2STR(sta->addr),
+			RSN_SUITE_ARG(pairwise),
+			sta->dot11RSNAStatsTKIPLocalMICFailures,
+			sta->dot11RSNAStatsTKIPRemoteMICFailures);
+
+	if (sta->wpa_sm == NULL)
+		return len;
+
+	/* Private MIB */
+	len += snprintf(buf + len, buflen - len,
+			"hostapdWPAPTKState=%d\n"
+			"hostapdWPAPTKGroupState=%d\n",
+			sta->wpa_sm->wpa_ptk_state,
+			sta->wpa_sm->wpa_ptk_group_state);
+
+	return len;
+}
