@@ -20,7 +20,8 @@ struct rsn_pmksa_cache {
 
 struct rsn_preauth_interface {
 	struct rsn_preauth_interface *next;
-	int fd;
+	struct hostapd_data *hapd;
+	struct l2_packet_data *l2;
 	char *ifname;
 	int ifindex;
 };
@@ -36,7 +37,7 @@ struct wpa_eapol_key {
 	u8 key_id[8]; /* Reserved */
 	u8 key_mic[16];
 	u16 key_data_length;
-	u8 key_data[0];
+	/* followed by key_data_length bytes of key_data */
 } __attribute__ ((packed));
 
 #define WPA_KEY_INFO_TYPE_MASK (BIT(0) | BIT(1) | BIT(2))
@@ -115,6 +116,7 @@ struct wpa_state_machine {
 	Boolean key_replay_counter_valid;
 	Boolean PInitAKeys; /* WPA only, not in IEEE 802.11i/D8 */
 	Boolean PTKRequest; /* not in IEEE 802.11i state machine */
+	Boolean has_GTK;
 
 	u8 *last_rx_eapol_key; /* starting from IEEE 802.1X header */
 	size_t last_rx_eapol_key_len;
@@ -144,6 +146,15 @@ struct wpa_authenticator {
 	Boolean changed;
 
 	unsigned int dot11RSNAStatsTKIPRemoteMICFailures;
+	u8 dot11RSNAAuthenticationSuiteSelected[4];
+	u8 dot11RSNAPairwiseCipherSelected[4];
+	u8 dot11RSNAGroupCipherSelected[4];
+	u8 dot11RSNAPMKIDUsed[PMKID_LEN];
+	u8 dot11RSNAAuthenticationSuiteRequested[4]; /* FIX: update */
+	u8 dot11RSNAPairwiseCipherRequested[4]; /* FIX: update */
+	u8 dot11RSNAGroupCipherRequested[4]; /* FIX: update */
+	unsigned int dot11RSNATKIPCounterMeasuresInvoked;
+	unsigned int dot11RSNA4WayHandshakeFailures;
 };
 
 
@@ -175,5 +186,8 @@ void rsn_preauth_finished(struct hostapd_data *hapd, struct sta_info *sta,
 void rsn_preauth_send(struct hostapd_data *hapd, struct sta_info *sta,
 		      u8 *buf, size_t len);
 void wpa_gtk_rekey(struct hostapd_data *hapd);
+int wpa_get_mib(struct hostapd_data *hapd, char *buf, size_t buflen);
+int wpa_get_mib_sta(struct hostapd_data *hapd, struct sta_info *sta,
+		    char *buf, size_t buflen);
 
 #endif /* WPA_H */
